     /********************************************************************
      ***                      Gestion SD Card                         ***
      ***               Fonctions de gestion de la FAT                 ***
      ***                      dsPIC30F4011                            ***
      ***                                                              ***
      ***  Fichier "source" : FAT.c                                    ***
      ***  Auteur : CREMMEL Marcel                                     ***
      ***  Date   : 30/09/2007                                         ***
      ********************************************************************/

#include <p30f4011.h>
#include <string.h>
#include "Hardware.h"
#include "Typedefs.h"
#include "Typedefs_SD.h"
#include "Typedefs_FAT.h"
#include "SD_Card.h"
#include "UART2.h"

/*------------------------------------------------------------------------------
                                  Variables                              
------------------------------------------------------------------------------*/

/* Couche physique
******************/
extern SDC_Error SDC_Status; // Etat de la carte

/* Dclaration multitypes du buffer pour un secteur
***************************************************/
extern Secteur *msd_buffer;  // Buffer de lecture et d'criture d'un secteur

/* Informations sur la BPB
**************************/
uint32 BPB_StartSector;         // Adresse du secteur BPB
uint16 BPB_ReservedSectorCount; // Nb de secteurs rservs derrire BPB 
byte   BPB_NumFATs;             // Nombre de blocs FAT (en principe 2)
byte   BPB_SectorsPerCluster;   // Nombre de secteurs par cluster
uint16 BPB_BytesPerSector;      // Nombre d'octets par secteur (souvent 512)
uint32 BPB_FATSize;             // Taille du bloc FAT en octets
uint32 BPB_TotSec;              // Nombre total de secteurs
uint32 BPB_DataSec;             // Nombre de secteurs du bloc de donnes
uint16 BPB_CountofClusters;     // Nombre total de clusters
uint16 BPB_RootDirEntryCount;   // Nombre de structures de rpertoires ou fichiers

/* Informations sur la FAT
**************************/
FAT_type FATType;               // Type de FAT de la SD Card
uint32 FAT_StartSector;         // Adresse du 1 secteur du bloc FAT
uint32 FAT_FirstDataSector;     // Adresse du 1 secteur du bloc de donnes
uint32 FAT_RootDirStartSector;  // Adresse du 1 secteur du bloc des structures de fichier
uint16 FAT_RootDirNumSectors;   // Nb de secteur du bloc des structures de fichier 

/* Informations sur le fichier ouvert
*************************************/
char  File_Name[11];       // Nom court du fichier (ou rpertoire)
char  File_Attr;           // Attributs du fichier ou rpertoire  
byte  File_CurrSectorNumInCluster;// Indice du secteur courant ds un cluster
uint16 File_StartCluster;  // N du 1 cluster du fichier (32 bits en FAT32)
uint16 File_CurrCluster;   // N du cluster courant (32 bits en FAT32)
uint32 File_Size;          // Taille du fichier en octets
uint32 File_StartSector;   // Premier secteur de donnes du 1 cluster
uint32 File_CurrSector;    // Adresse du secteur courant ds le bloc de donnes
uint32 File_CurrByte;      // Compteur d'octets lus ou crits
uint32 File_DirSectorNum;  // Adresse du 1 secteur de la structure de fichier
uint16 File_DirEntryNum;   // Indice ds le secteur de la structure d'entre de fichier
                    
/* Variables "courantes"
************************/
uint32 SD_CurrSector; // Adresse du secteur courant
uint32 SecteurTampon; // Adresse du secteur utilis comme tampon d'criture

/* Constantes spcifiques
*************************/
// Masques des attributs de fichier
#define FAT_IS_DIR        0x10
#define FAT_END_OF_DIR    0x02
#define FAT_FILE_DELETED  0xE5
#define FAT_LONG_FILENAME 0x03

/*------------------------------------------------------------------------------
             Dclarations des fonctions en mmoire flash
------------------------------------------------------------------------------*/

/*---------------------------------------------------------------
  Function   : void Reset_FAT(void)
  Description: Initialisation de la structure BPB pour les tests     
               (infos prises sur SD Card 1G en FAT16)
-----------------------------------------------------------------*/
void Reset_FAT(void)
{
 BPB_StartSector=133; 
 msd_buffer->BPB_FAT.BS_jmpBoot[0]=0xEB; 
 msd_buffer->BPB_FAT.BS_jmpBoot[1]=0; 
 msd_buffer->BPB_FAT.BS_jmpBoot[2]=0x90;
 memcpy(msd_buffer->BPB_FAT.BS_OEMName,"CREMMEL.",8);
 msd_buffer->BPB_FAT.BPB_BytsPerSec[0]=0; 
 msd_buffer->BPB_FAT.BPB_BytsPerSec[1]=0x02;
 msd_buffer->BPB_FAT.BPB_SecPerClus=32;
 msd_buffer->BPB_FAT.BPB_RsvdSecCnt=6; 
 msd_buffer->BPB_FAT.BPB_NumFATs=2;
 msd_buffer->BPB_FAT.BPB_RootEntCnt[0]=0; 
 msd_buffer->BPB_FAT.BPB_RootEntCnt[1]=2;
 msd_buffer->BPB_FAT.BPB_TotSec16[0]=0; 
 msd_buffer->BPB_FAT.BPB_TotSec16[1]=0;
 msd_buffer->BPB_FAT.BPB_Media=0xF8;
 msd_buffer->BPB_FAT.BPB_FATSz16=237;
 msd_buffer->BPB_FAT.BPB_SecPerTrk=63;
 msd_buffer->BPB_FAT.BPB_NumHeads=32;
 msd_buffer->BPB_FAT.BPB_HiddSec=133;
 msd_buffer->BPB_FAT.BPB_TotSec32=1935227;
 msd_buffer->BPB_FAT.BS_DrvNum=0x80;
 msd_buffer->BPB_FAT.BS_Reserved1=0;
 msd_buffer->BPB_FAT.BS_BootSig=0x29;
 msd_buffer->BPB_FAT.BS_VolID[0]=0; 
 msd_buffer->BPB_FAT.BS_VolID[1]=1;
 msd_buffer->BPB_FAT.BS_VolID[2]=2; 
 msd_buffer->BPB_FAT.BS_VolID[3]=3;
 memcpy(msd_buffer->BPB_FAT.BS_VolLab,"SD_CARD    ",11);
 memcpy(msd_buffer->BPB_FAT.BS_FilSysType,"FAT16   ",8);
 msd_buffer->BPB_FAT.SignSector0=0xAA55;
}

/*---------------------------------------------------------------
  Function   : void SampleFileData(Secteur *buffer)
  Description: Affectation de qq strucures de fichiers ds la FAT   
               Utilis pour des tests             
  Paramtre  : Pointeur vers un secteur structur "Secteur" de 
               512 octets
-----------------------------------------------------------------*/
void SampleFileData(Secteur *buffer)
{
 int i;
 strcpy(buffer->FATFilesInfos[0].DIR_Name,"SD CARD    ");
 buffer->FATFilesInfos[0].DIR_Attr=0x08;
 buffer->FATFilesInfos[0].DIR_NTRes=0;
 buffer->FATFilesInfos[0].DIR_CrtTimeTenth=0;
 buffer->FATFilesInfos[0].DIR_CrtTime[0]=0;
 buffer->FATFilesInfos[0].DIR_CrtTime[1]=0;
 buffer->FATFilesInfos[0].DIR_CrtDate[0]=0;
 buffer->FATFilesInfos[0].DIR_CrtDate[1]=0;
 buffer->FATFilesInfos[0].DIR_LstAccDate[0]=0;
 buffer->FATFilesInfos[0].DIR_LstAccDate[1]=0;
 buffer->FATFilesInfos[0].DIR_FstClusHI=0;
 buffer->FATFilesInfos[0].DIR_WrtTime[0]=0;
 buffer->FATFilesInfos[0].DIR_WrtTime[1]=0;
 buffer->FATFilesInfos[0].DIR_WrtDate[0]=0;
 buffer->FATFilesInfos[0].DIR_WrtDate[1]=0;
 buffer->FATFilesInfos[0].DIR_FstClusLO=0;
 buffer->FATFilesInfos[0].DIR_FileSize=0;

 strcpy(buffer->FATFilesInfos[1].DIR_Name,"TEST1   TXT");
 buffer->FATFilesInfos[1].DIR_Attr=0x20;
 buffer->FATFilesInfos[1].DIR_NTRes=0;
 buffer->FATFilesInfos[1].DIR_CrtTimeTenth=0;
 buffer->FATFilesInfos[1].DIR_CrtTime[0]=0;
 buffer->FATFilesInfos[1].DIR_CrtTime[1]=0;
 buffer->FATFilesInfos[1].DIR_CrtDate[0]=0;
 buffer->FATFilesInfos[1].DIR_CrtDate[1]=0;
 buffer->FATFilesInfos[1].DIR_LstAccDate[0]=0;
 buffer->FATFilesInfos[1].DIR_LstAccDate[1]=0;
 buffer->FATFilesInfos[1].DIR_FstClusHI=0;
 buffer->FATFilesInfos[1].DIR_WrtTime[0]=0;
 buffer->FATFilesInfos[1].DIR_WrtTime[1]=0;
 buffer->FATFilesInfos[1].DIR_WrtDate[0]=0;
 buffer->FATFilesInfos[1].DIR_WrtDate[1]=0;
 buffer->FATFilesInfos[1].DIR_FstClusLO=2;
 buffer->FATFilesInfos[1].DIR_FileSize=0x00000046;

 buffer->FATFilesInfos[2].DIR_Name[0]=FAT_FILE_DELETED;
 buffer->FATFilesInfos[2].DIR_Attr=0x20;
 buffer->FATFilesInfos[2].DIR_NTRes=0;
 buffer->FATFilesInfos[2].DIR_CrtTimeTenth=0;
 buffer->FATFilesInfos[2].DIR_CrtTime[0]=0;
 buffer->FATFilesInfos[2].DIR_CrtTime[1]=0;
 buffer->FATFilesInfos[2].DIR_CrtDate[0]=0;
 buffer->FATFilesInfos[2].DIR_CrtDate[1]=0;
 buffer->FATFilesInfos[2].DIR_LstAccDate[0]=0;
 buffer->FATFilesInfos[2].DIR_LstAccDate[1]=0;
 buffer->FATFilesInfos[2].DIR_FstClusHI=0;
 buffer->FATFilesInfos[2].DIR_WrtTime[0]=0;
 buffer->FATFilesInfos[2].DIR_WrtTime[1]=0;
 buffer->FATFilesInfos[2].DIR_WrtDate[0]=0;
 buffer->FATFilesInfos[2].DIR_WrtDate[1]=0;
 buffer->FATFilesInfos[2].DIR_FstClusLO=2;
 buffer->FATFilesInfos[2].DIR_FileSize=0;

 strcpy(buffer->FATFilesInfos[3].DIR_Name,"TEST3   C  ");
 buffer->FATFilesInfos[3].DIR_Attr=0x20;
 buffer->FATFilesInfos[3].DIR_NTRes=0;
 buffer->FATFilesInfos[3].DIR_CrtTimeTenth=0;
 buffer->FATFilesInfos[3].DIR_CrtTime[0]=0;
 buffer->FATFilesInfos[3].DIR_CrtTime[1]=0;
 buffer->FATFilesInfos[3].DIR_CrtDate[0]=0;
 buffer->FATFilesInfos[3].DIR_CrtDate[1]=0;
 buffer->FATFilesInfos[3].DIR_LstAccDate[0]=0;
 buffer->FATFilesInfos[3].DIR_LstAccDate[1]=0;
 buffer->FATFilesInfos[3].DIR_FstClusHI=0;
 buffer->FATFilesInfos[3].DIR_WrtTime[0]=0;
 buffer->FATFilesInfos[3].DIR_WrtTime[1]=0;
 buffer->FATFilesInfos[3].DIR_WrtDate[0]=0;
 buffer->FATFilesInfos[3].DIR_WrtDate[1]=0;
 buffer->FATFilesInfos[3].DIR_FstClusLO=5;
 buffer->FATFilesInfos[3].DIR_FileSize=0x0001658A;

 for (i=4; i<16; i++) buffer->FATFilesInfos[i].DIR_Name[0]=FAT_FILE_DELETED;
}

/*----------------------------------------------------------
  Function   : uint32 FindFreeClusters(void)
  Description: Calcul et renvoi du nombre de clusters libres
  Retour     : Nombre de clusters libres
------------------------------------------------------------*/
uint32 FindFreeClusters(void)
{
 uint32 freeclusters=0;
 int i,j;
 uint fatentry;
 for (i=0; i< BPB_FATSize; i++) 
 {// Pour chaque secteur de la FAT ...
  SD_CurrSector=FAT_StartSector+i;
  SectorRead(SD_CurrSector,msd_buffer);
  for (j=0; j<256; j++) 
  {// ... compter le nombre d'entres de cluster  0
   fatentry=msd_buffer->mots_16b[j]; // FAT16 seulement,  modifier
   if (fatentry==0) freeclusters++;
  }
 }
 return freeclusters;
}

/*--------------------------------------------------------------------
  Function   : uint32 ConvertClusterToSector(uint16 cluster)
  Description: Calcul de l'adresse du 1er secteur de donnes du cluster
  Paramtre  : Numro du cluster
  Retour     : Adresse de son premier secteur
----------------------------------------------------------------------*/
uint32 ConvertClusterToSector(uint16 cluster) 
{
 uint32 sector;
 sector=(uint32)cluster-2;    	// Les 2 premires entres ne sont pas valides
 sector*=BPB_SectorsPerCluster; 
 // Calcul de l'adresse du 1 secteur de donnes
 sector+=FAT_FirstDataSector;   
 return sector;
}

/*------------------------------------------------
  Function   : GetNextCluster(uint16 cluster)
  Description: Lecture du N du cluster suivant
  Paramtre  : Numro du cluster actuel
  Retour     : Numro du prochain cluster
---------------------------------------------------*/
uint16 GetNextCluster(uint16 cluster) 
{
 uint32 AdrFATbyte;   // Adresse "octet" relative de l'entre dans le bloc FAT
 uint32 AdrFATsector; // Adresse "secteur" relative de l'entre dans le bloc FAT
 uint32 OffsetFATcluster; // Offset de l'entre dans le secteur
 uint16 data;
 // Recherche du secteur ds le bloc FAT o se trouve l'entre
 if (FATType==FAT16) AdrFATbyte=cluster<<1;     // 2 octets par lment en FAT16
 else if (FATType==FAT32) AdrFATbyte=cluster<<2;// 4 octets par lment en FAT32
 // Adresse relative "secteur" ds le bloc FAT
 AdrFATsector=AdrFATbyte/BPB_BytesPerSector;
 // Adresse absolue du secteur contenant l'entre de cluster
 SD_CurrSector=FAT_StartSector+AdrFATsector;
 // Offset du 1 octet de l'lment ds ce secteur
 OffsetFATcluster=AdrFATbyte-(uint32)(AdrFATsector*BPB_BytesPerSector);
 // Lecture du secteur FAT concern ds le buffer
 SectorRead(SD_CurrSector,msd_buffer);
 // Lecture du N du cluster suivant
 if (FATType==FAT16) data=msd_buffer->mots_16b[(uint16)OffsetFATcluster>>1];
 else if (FATType==FAT32) data=msd_buffer->mots_16b[(uint16)OffsetFATcluster>>2];
 //Attention: data devrait tre code sur 32 bits en FAT32 :data=msd_buffer.mots_32b[]
 return data;
}

/*------------------------------------------------------------------------------
  Function   : uint16 FindNextUnusedCluster(uint16 cluster)
  Description: Recherche du n de cluster de la prochaine entre de cluster libre
  Paramtre  : N du cluster initial
  Retour     : N du prochain cluster libre
--------------------------------------------------------------------------------*/
uint16 FindNextUnusedCluster(uint16 cluster) 
{
 uint16 i; // Compteur de secteurs ds le bloc FAT
 uint16 j; // Compteur d'lments ds un secteur
 // On commence avec le N de cluster suivant
 cluster+=1;
 i=cluster>>8; // Adresse du secteur ds la FAT (256 entres par secteur)
 for (; i<BPB_FATSize; i++) 
 {// Pour tous les secteurs restant jusqu' la fin de la FAT :
  SD_CurrSector=FAT_StartSector+i; // Adresse absolue du secteur ds le volume
  SectorRead(SD_CurrSector,msd_buffer);// Transfert du secteur ds le buffer
  j=cluster&0x00FF; // Indice initial de l'lment ds le secteur courant
                    // (tjrs 0 pour les secteurs suivants)
  for (; j<256; j++) 
  {// Rechercher la prochaine entre libre ds le secteur "SD_CurrSector" :
   if (msd_buffer->mots_16b[j]==0) return (i*256+j);// FAT16
  }
  cluster=0; // Pour que les boucles suivantes commencent avec j=0
 }
 return 0;
}

/*------------------------------------------
  Function:    void QuickFormat()
  Description: formatage rapide du volume
               A n'utiliser qu'en urgence !
---------------------------------------------*/
void QuickFormat(void) 
{
 uint32 last;
 uint16 j;
 // Dernier secteur des structures de fichier
 last=FAT_RootDirStartSector+FAT_RootDirNumSectors;
 for (j=0; j< 256; j++) msd_buffer->mots_16b[j]=0; // Tous les octets du buffer  0
 for (SD_CurrSector=FAT_StartSector; SD_CurrSector<last; SD_CurrSector++) 
 {// Pour tous les secteurs du tableau des structures de fichiers :
  WTX2_Msg_Nb32bits("Initialisation du secteur ...:", SD_CurrSector);
  // Ecriture du buffer ds le secteur
  SectorWrite(SD_CurrSector,msd_buffer); 
 }
}

/*-------------------------------------------------------------------------------
  Function   : int OpenFileInRoot(uint16 filenumber)
  Description: ouverture d'un fichier ds le rpertoire racine
  Paramtre  : indice de la structure de fichier ds le bloc des structures
  Retour     : TRUE              : le fichier existe et est "ouvert"
               FAT_END_OF_DIR    : le fichier n'existe pas et le n est le dernier
               FAT_FILE_DELETED  : le fichier n'existe pas (il a t effac)
               FAT_LONG_FILENAME : le fichier existe mais a un nom long (non gr)
--------------------------------------------------------------------------------*/
int OpenFileInRoot(uint filenumber) 
{
 uint OffsetSector;
 char c;
 // 16 structures de fichiers par secteur (une structure = 32 octets)
 OffsetSector=filenumber>>4;         // Partie entire de (filenumber/16)
 // Indice de la structure d'entre de fichier ds le secteur
 File_DirEntryNum=filenumber&0x000F; // Reste de (filenumber/16)=indice ds le secteur
 SD_CurrSector=FAT_RootDirStartSector+OffsetSector; // Adresse absolue du secteur
 // Adresse courante du secteur de la structure de fichier
 File_DirSectorNum=SD_CurrSector; 
 // Lecture du secteur de la structure de fichier recherche
 WTX2_Msg_Nb32bits(" Lecture du secteur N ",SD_CurrSector);
 SectorRead(SD_CurrSector,msd_buffer);
 c=msd_buffer->FATFilesInfos[File_DirEntryNum].DIR_Name[0];
 if (c==0x0) 
 {// Structure de fichier libre et c'est la dernire
  WriteLnTXD2_String(" Structure de fichier libre et c'est la derniere.");
  WriteLnTXD2();
  return FAT_END_OF_DIR;
 } 
 if (c==(char)FAT_FILE_DELETED)
 {// Structure de fichier libre
  WriteLnTXD2_String(" Structure de fichier libre.");
  WriteLnTXD2();
  return c;
 }
 // Le fichier existe : lecture des champs de la structure de fichier
 memcpy(File_Name,msd_buffer->FATFilesInfos[File_DirEntryNum].DIR_Name,11); // Nom court
 File_Attr=msd_buffer->FATFilesInfos[File_DirEntryNum].DIR_Attr; // Aucun attribut
 if (File_Attr == 0x0F) // Nom de fichier long, non pris en charge
       return FAT_LONG_FILENAME;
 File_StartCluster=msd_buffer->FATFilesInfos[File_DirEntryNum].DIR_FstClusLO;
 File_CurrCluster=File_StartCluster; // N du 1 cluster du fichier
 // Premier secteur de donne du 1 cluster
 File_StartSector=ConvertClusterToSector(File_StartCluster);
 File_CurrSector=File_StartSector;   // Secteur de donnes courant
 // Taille du fichier en octets
 File_Size=msd_buffer->FATFilesInfos[File_DirEntryNum].DIR_FileSize;
 File_CurrSectorNumInCluster=0;      // N du secteur ds le cluster courant
 File_CurrByte=0;                    // Compteur d'octets lus
 return TRUE;
}

/*---------------------------------------------------------------------
  Function:    int OpenFileWithName(char *filename)
  Description: Ouvrir un fichier ds le rpertoire racine avec son nom
  Paramtre  : Nom court du fichier avec son extension, mais sans le "."
  Retour     : Renvoi 
                 - du n du fichier en cas de succs
                 - de -1 en cas d'erreur
-----------------------------------------------------------------------*/
int OpenFileWithName(char *filename) 
{
 uint16 filenum;
 byte ret;
 filenum=0; // N initial de la structure de fichier
 WriteTXD2_String("Recherche du n de fichier correspondant  : ");
 WriteLnTXD2_String(filename);
 WriteLnTXD2_String("********************************************");
 while (1) 
 {// Recherche du N de fichier correspondant au nom recherch
  WTX2_Msg_Nb8bits(" Test avec le fichier n : ",filenum);
  // Affectations de certaines variables "File_...", notamment "File_Name"
  ret=OpenFileInRoot(filenum); 
  if (ret==FAT_END_OF_DIR) 
  {
   WriteLnTXD2_String(" <EOD>");
   WriteLnTXD2();
   break;
  }
  if ((ret!=FAT_FILE_DELETED)&&(ret!=FAT_LONG_FILENAME)) 
  {
   if (memcmp(File_Name,filename,11)==0) // Comparaison du nom de fichier court 
   {
    WriteLnTXD2_String(" Fichier trouve !");
    WriteLnTXD2();
    return filenum;
   }
   else 
   {
    WriteTXD2_String(" Le nom ne correspond pas. ");
    WriteLnTXD2_String(File_Name);
    WriteLnTXD2();
   }
  }
  filenum++;
 }
 return -1;
}

/*----------------------------------------------------------------------
  Function:    int ReadNextSectorIntoBuff(void)
  Description: Lecture du secteur suivant du cluster courant ds le buffer
  Retour     : Renvoi de 
                  - TRUE si l'opration s'est bien passe
                  - FALSE ds le cas contraire
  Le fichier doit tre "ouvert" : affectations des variables "File_..."
------------------------------------------------------------------------*/
int ReadNextSectorIntoBuff(void) 
{
 // Secteur suivant ds le cluster suivant ?
 if (File_CurrSectorNumInCluster==BPB_SectorsPerCluster)
 {// Oui : recherche du N du cluster suivant :
  WriteLnTXD2(); WTX2_Msg_Nb16bits("Ancien cluster N :",File_CurrCluster);
  File_CurrCluster=GetNextCluster(File_CurrCluster);
  WTX2_Msg_Nb16bits("Nouveau cluster N :",File_CurrCluster);
  if (File_CurrCluster==0||File_CurrCluster>0xFFF7) return FALSE;
  File_CurrSectorNumInCluster=0; // Modulo "BPB_SectorsPerCluster"
 }
 // Premier secteur du cluster courant
 File_CurrSector=(uint32)ConvertClusterToSector(File_CurrCluster);
 // Ajot de l'offset du secteur ds le cluster
 File_CurrSector+=File_CurrSectorNumInCluster;
 SD_CurrSector=File_CurrSector; // Nouveau secteur courant
 WriteLnTXD2();
 WTX2_Msg_Nb32bits("Lecture du secteur N : ",SD_CurrSector);
 WTX2_Msg_Nb8bits("Offset en nb de secteurs ds le cluster: ",File_CurrSectorNumInCluster);
 SectorRead(SD_CurrSector,msd_buffer); // Lecture de ce secteur ds le buffer
 File_CurrSectorNumInCluster++; // Secteur suivant ds le cluster courant
 return TRUE;
}

/*-------------------------------------------------------------------------
  Function:    int WriteBuffIntoNextSector(void)
  Description: Ecriture du secteur tampon dans le secteur suivant du fichier 
               ouvert (soit dans le secteur suivant du cluster courant)
               Mise  jour des 2 clusters concerns
               Mise  jour de la taille du fichier
  Retour     : TRUE ou FALSE suivant l'opration effectue
  Le fichier doit tre "ouvert" : affectations des variables "File_..."
---------------------------------------------------------------------------*/
int WriteBuffIntoNextSector(void) 
{
 uint16 oldcluster;
 oldcluster=File_CurrCluster; // Sauvegarde du N du cluster courant
 // Ecriture ds le premier secteur du premier cluster d'un fichier vide ?
 if (File_StartCluster==0)
 {// Oui : recherche du prochain N de cluster libre :
  File_StartCluster=FindNextUnusedCluster(2); // Premier cluster
  File_CurrCluster=File_StartCluster; // Cluster courant
  WriteLnTXD2(); WTX2_Msg_Nb16bits(" Premier N de cluster du fichier : ",File_StartCluster);
  File_CurrSector=(ConvertClusterToSector(File_CurrCluster))+File_CurrSectorNumInCluster;
  WTX2_Msg_Nb32bits(" Premier secteur de donnees : ",File_CurrSector);
  // Lecture du secteur de la structure de fichier du fichier courant
  SD_CurrSector=File_DirSectorNum;
  WTX2_Msg_Nb32bits(" Lecture du secteur N (structures de fichier) : ",SD_CurrSector);
  // Lecture du secteur courant ds le buffer
  SectorRead(SD_CurrSector,msd_buffer); 
  // Affectation du N du 1 cluster
  msd_buffer->FATFilesInfos[File_DirEntryNum].DIR_FstClusLO=File_StartCluster; 
  WTX2_Msg_Nb32bits(" Ecriture du cluster initial ds le secteur N (structures de fichier) : ",SD_CurrSector);
  SectorWrite(SD_CurrSector,msd_buffer); // Ecriture du buffer modifi ds le secteur
  // Prochaine entre de cluster  EOC
  SD_CurrSector=FAT_StartSector+(File_CurrCluster/256); // Adresse ds la FAT16
  WTX2_Msg_Nb32bits(" Lecture du secteur N (FAT) : ",SD_CurrSector);
  SectorRead(SD_CurrSector,msd_buffer); // Lecture du secteur de la FAT
  msd_buffer->mots_16b[File_CurrCluster&0xff]=0xFFFF; // N cluster suivant = EOC
  WTX2_Msg_Nb32bits(" Ecriture ds ce secteur de EOC ds le cluster N : ",File_CurrCluster);
  SectorWrite(SD_CurrSector,msd_buffer); // Ecriture du buffer modifi ds le secteur
  File_Size=0; // Taille initiale du fichier
 }

 // Ecriture ds le premier secteur du cluster suivant ?
 if (File_CurrSectorNumInCluster==BPB_SectorsPerCluster) 
 {// Oui : recherche du N du cluster suivant :
  File_CurrSectorNumInCluster=0; // Modulo "BPB_SectorsPerCluster"
  File_CurrCluster=FindNextUnusedCluster(File_CurrCluster); // Cluster libre suivant
  File_CurrSector=(ConvertClusterToSector(File_CurrCluster));
  WriteLnTXD2(); WTX2_Msg_Nb16bits(" Nouveau N de cluster aloue : ",File_CurrCluster);
  if (File_CurrCluster==0||File_CurrCluster>0xFFF7) return FALSE;
  // Affectation du N du cluster suivant ds la FAT
  SD_CurrSector=FAT_StartSector+(oldcluster/256); // Adresse ds la FAT16
  WTX2_Msg_Nb32bits(" Lecture du secteur N (FAT) : ",SD_CurrSector);
  SectorRead(SD_CurrSector,msd_buffer); // Lecture du secteur de la FAT
  msd_buffer->mots_16b[oldcluster&0xff]=File_CurrCluster; // N cluster suivant
  WTX2_Msg_Nb32bits(" Ecriture du N du cluster suivant ds le secteur N : ",SD_CurrSector);
  SectorWrite(SD_CurrSector,msd_buffer); // Ecriture du buffer modifi ds le secteur
  // Prochaine entre de cluster  EOC
  SD_CurrSector=FAT_StartSector+(File_CurrCluster/256); // Adresse ds la FAT16
  WTX2_Msg_Nb32bits(" Lecture du secteur N (FAT) : ",SD_CurrSector);
  SectorRead(SD_CurrSector,msd_buffer); // Lecture du secteur de la FAT
  msd_buffer->mots_16b[File_CurrCluster&0xff]=0xFFFF; // N cluster suivant = EOC
  WTX2_Msg_Nb32bits(" Ecriture ds ce secteur de EOC ds le cluster N : ",File_CurrCluster);
  SectorWrite(SD_CurrSector,msd_buffer); // Ecriture du buffer modifi ds le secteur
 }

 // Ecriture des 512 octets du buffer d'criture ds le secteur courant
 WTX2_Msg_Nb32bits(" Lecture des 512 octets du secteur tampon : ",SecteurTampon);
 SectorRead(SecteurTampon,msd_buffer); // Lecture du tampon d'criture
 SD_CurrSector=File_CurrSector; // Nouveau secteur courant
 WTX2_Msg_Nb32bits(" Ecriture des 512 octets du tampon ds le secteur N : ",SD_CurrSector);
 SectorWrite(SD_CurrSector,msd_buffer); // Ecriture du buffer ds le secteur courant

 // Mise  jour de la taille du fichier :
 // Lecture du secteur de la structure de fichier du fichier courant
 SD_CurrSector=File_DirSectorNum;
 WTX2_Msg_Nb32bits(" Lecture du secteur N (structures de fichier) : ",SD_CurrSector);
 // Lecture du secteur courant ds le buffer
 SectorRead(SD_CurrSector,msd_buffer); 
 // Nouvelle taille du fichier
 File_Size+=512; // 512 octets par secteur
 msd_buffer->FATFilesInfos[File_DirEntryNum].DIR_FileSize=File_Size; 
 WTX2_Msg_Nb32bits(" Mise a jour de la taille du fichier : ",File_Size);
 SectorWrite(SD_CurrSector,msd_buffer); // Ecriture du buffer modifi ds le secteur

 File_CurrSectorNumInCluster++; // Secteur suivant ds le cluster courant
 File_CurrSector++;
 return TRUE;
}

/*-----------------------------------------------------
  Function:    void ReadFile(char *name)
  Description: Lecture d'un fichier du rpertoire racine, 
               sortie sur UART2
  Paramtre  : Nom complet du fichier (8+3 caractres)
-------------------------------------------------------*/
void ReadFile(char *name) 
{
 uint16 idx;
 if (OpenFileWithName(name)==0xFFFF) return; // "Ouvrir" le fichier
 WriteLnTXD2();
 WriteTXD2_String("Lecture du fichier :");
 WriteLnTXD2_String(name);
 WriteLnTXD2_String("********************");
 while (File_CurrByte<File_Size) // Dernier octet du fichier lu ? 
 {// Non : lecture du secteur suivant 
  if (ReadNextSectorIntoBuff())
  {
   idx=0; // Indice ds le secteur de l'octet de donne
   while ((File_CurrByte<File_Size)&&(idx<512))
   {// Pour chaque octet du buffer ou jusqu'au dernier octet du fichier :
    WriteTXD2(msd_buffer->octets[idx++]); // Transmission sur UART2
    File_CurrByte++; // Compteur d'octets
   }
  }
 }
}

/*---------------------------------------------
  Function:    void PrintCurrFileInfo(void)
  Description: utilis par ListDir
-----------------------------------------------*/
void PrintCurrFileInfo(void) 
{
 WriteTXD2_String(" Nom et taille :");
 WriteTXD2_String(File_Name);
 WTX2_Msg_Nb32bits(",",File_Size);
 WTX2_Msg_Nb8bits(" Attributs : ",File_Attr);
 WTX2_Msg_Nb16bits(" N du 1 cluster (File_StartCluster):",File_StartCluster);
 WTX2_Msg_Nb32bits(" Premier secteur de donnees :",ConvertClusterToSector(File_StartCluster));
 if (File_Attr & FAT_IS_DIR) 
 {
  WriteLnTXD2_String(" <DIR>");
  WriteLnTXD2();
 }
 else WriteLnTXD2();
}

/*-------------------------------------------------------------
  Function:    void ListDir(void)
  Description: Transmission sur TX2 de la liste des fichiers du 
               rpertoire racine
---------------------------------------------------------------*/
void ListDir(void) 
{
 int filenum;
 char ret;
 uint32 freespace;
 freespace=FindFreeClusters()*BPB_SectorsPerCluster*BPB_BytesPerSector;
 WriteLnTXD2();
 WTX2_Msg_Nb32bits("Espace libre en octets : ",freespace);
 WriteLnTXD2();
 filenum=0;
 while (1) 
 {
  WTX2_Msg_Nb16bits("Fichier N : ",filenum);
  ret=OpenFileInRoot(filenum); // "Ouvrir" le fichier
  if (ret==FAT_END_OF_DIR) 
  {
   WriteLnTXD2_String("<EOD>");
   return;
  }
  if ((ret!=(char)FAT_FILE_DELETED)&&(ret!=(char)FAT_LONG_FILENAME)) 
     PrintCurrFileInfo();
  if (ret==(char)FAT_LONG_FILENAME) 
  {
   WriteLnTXD2_String(" Nom de fichier long non pris en charge");
   WriteLnTXD2();
  }
  filenum++;
 }
}

/*---------------------------------------------------------
  Function:    int CreateFile(char *name, uint32 len)
  Description: Cration d'un fichier dans le rpertoire racine
  Paramtre  : Nom complet du fichier (8+3 caractres)
  Retour     : TRUE ou FALSE suivant la russite l'opration
------------------------------------------------------------*/
int CreateFile(char *name)
{
 char c;
 int i;
 // Recherche d'une entre de fichier libre ds la FAT
 WriteLnTXD2();
 WriteLnTXD2_String("Creation d'un fichier vide :");
 WriteLnTXD2_String("**************************");
 for (i=0;i< BPB_RootDirEntryCount;i++) 
 {// Pour chaque entre de fichier ou rpertoire de 32 octets :
  // Calcul du secteur de la structure courante
  SD_CurrSector=(uint32)FAT_RootDirStartSector+i/16; // 16 entres par secteur
  File_DirEntryNum=i%16; // Numro de l'entre ds le secteur (0  15) 
  File_DirSectorNum=SD_CurrSector;               // Secteur courant
  WTX2_Msg_Nb32bits("Lecture du secteur N : ",SD_CurrSector);
  SectorRead(SD_CurrSector,msd_buffer); // Lecture de ce secteur ds le buffer
  WTX2_Msg_Nb16bits("Test de l'entree de fichier N : ",i);
  c=msd_buffer->FATFilesInfos[File_DirEntryNum].DIR_Name[0]; // 1 octet de la structure
  if (c==0x0||c==(char)FAT_FILE_DELETED) // Entre de fichier libre ?
  {// Oui :
   if (c == 0x0) 
   {// Libre et c'est la dernire dans l'tat actuel des fichiers
    WriteLnTXD2_String("EOD");
    if (File_DirEntryNum!=15) // Prochaine entre ds le secteur suivant ? 
    {// Non : prochaine entre = EOD
     msd_buffer->FATFilesInfos[File_DirEntryNum+1].DIR_Name[0]=0;
    } 
    else if ((i+1)!=BPB_RootDirEntryCount) 
    {// Oui et il y a encore de la place :
     SD_CurrSector++; // Secteur suivant
     WTX2_Msg_Nb32bits("Lecture du secteur N : ",SD_CurrSector);
     SectorRead(SD_CurrSector,msd_buffer); // Lecture de ce secteur ds le buffer
     // Entre de fichier suivante libre et c'est la dernire
     msd_buffer->FATFilesInfos[File_DirEntryNum].DIR_Name[0]=0; 
     WTX2_Msg_Nb32bits("Ecriture ds le secteur N : ",SD_CurrSector);
     SectorWrite(SD_CurrSector,msd_buffer); // Ecriture du buffer modifi ds le secteur
     SD_CurrSector--; // Retour au secteur de l'entre de fichier en cours
     SectorRead(SD_CurrSector,msd_buffer); // Lecture de ce secteur ds le buffer
    } 
    else 
    {// Plus de place pour une nouvelle structure de fichier !
     WriteLnTXD2_String("La derniere entree !!!!");
     return FALSE;
    }
   } 
   else 
   {// c==FAT_FILE_DELETED : la structure est libre et ce n'est pas la dernire
    WriteLnTXD2_String("Creation de fichier ds une entre effacee.");
   }
   // Affectation du champ "DIR_Name"
   memcpy(msd_buffer->FATFilesInfos[File_DirEntryNum].DIR_Name,name,11);
   // Affectation du champ "attribut"
   msd_buffer->FATFilesInfos[File_DirEntryNum].DIR_Attr=0;
   // Affectation du champ "DIR_FstClusLO" (16 bits) avec 0 (longueur 0)
   msd_buffer->FATFilesInfos[File_DirEntryNum].DIR_FstClusHI=0;
   msd_buffer->FATFilesInfos[File_DirEntryNum].DIR_FstClusLO=0;
   // Affectation du champ "DIR_FileSize" (32 bits) : 0
   msd_buffer->FATFilesInfos[File_DirEntryNum].DIR_FileSize=0;
   // Autres informations :
   msd_buffer->FATFilesInfos[File_DirEntryNum].DIR_NTRes=0;
   msd_buffer->FATFilesInfos[File_DirEntryNum].DIR_CrtTimeTenth=0;
   msd_buffer->FATFilesInfos[File_DirEntryNum].DIR_CrtTime[0]=0;
   msd_buffer->FATFilesInfos[File_DirEntryNum].DIR_CrtTime[1]=0;
   msd_buffer->FATFilesInfos[File_DirEntryNum].DIR_CrtDate[0]=0;
   msd_buffer->FATFilesInfos[File_DirEntryNum].DIR_CrtDate[1]=0;
   msd_buffer->FATFilesInfos[File_DirEntryNum].DIR_LstAccDate[0]=0;
   msd_buffer->FATFilesInfos[File_DirEntryNum].DIR_LstAccDate[1]=0;
   msd_buffer->FATFilesInfos[File_DirEntryNum].DIR_WrtTime[0]=0;
   msd_buffer->FATFilesInfos[File_DirEntryNum].DIR_WrtTime[1]=0;
   msd_buffer->FATFilesInfos[File_DirEntryNum].DIR_WrtDate[0]=0;
   msd_buffer->FATFilesInfos[File_DirEntryNum].DIR_WrtDate[1]=0;
   WTX2_Msg_Nb32bits("Ecriture dans le secteur : ",SD_CurrSector);

   File_Attr=0; // Aucun attribut
   File_StartCluster=0;
   File_CurrCluster=File_StartCluster;
   File_Size=0;
   File_CurrSectorNumInCluster=0;

   TEST0=1; // Pour dclencher l'oscillo ou l'analyseur logique
   SectorWrite(SD_CurrSector,msd_buffer); // Ecriture du buffer modifi ds le secteur
   TEST0=0; // Pour dtecter la fin de l'criture d'un secteur
   return TRUE;
  }// Fin du bloc "if (c==0x0||c==(char)FAT_FILE_DELETED)"
 }// Fin du bloc "for (i=0; i< BPB_RootDirEntryCount;  i++)"
 return FALSE;
}

/*--------------------------------------------------------------
  Function   : int DeleteFile(char *name)
  Description: Effacer un fichier ds le rpertoire racine
  Paramtre  : Nom complet du fichier  effacer (8+3 caractres)
  Retour     : TRUE ou FALSE suivant la russite l'opration
----------------------------------------------------------------*/
int DeleteFile(char *name) 
{
 uint16 cluster, newcluster;
 WriteLnTXD2();
 WriteTXD2_String("Effacer le fichier ");
 WriteLnTXD2_String(name);
 WriteLnTXD2_String("******************");
 // Ouverture du fichier : affectation des variables "File_..."
 if (OpenFileWithName(name)==0xFFFF) return FALSE;
 PrintCurrFileInfo(); // Transmettre les principales infos sur TXD2
 WriteLnTXD2_String("Effacer la chaine de cluster :");
 newcluster=File_StartCluster; // Premier cluster du fichier
 WTX2_Msg_Nb16bits(" On commence par le cluster N : ",newcluster);
 cluster=newcluster;
 SD_CurrSector=FAT_StartSector+(cluster/256); // Adresse du secteur (FAT16)
 SectorRead(SD_CurrSector,msd_buffer);// Transfert du secteur ds le buffer
 while ((newcluster<=0xFFF8)&&(newcluster!=0x0000)) 
 {// Pour chaque cluster valide du fichier :
  cluster=newcluster;
  newcluster=msd_buffer->mots_16b[cluster&0xFF];// N du cluster suivant (FAT16)
  WTX2_Msg_Nb16bits(" On poursuit avec le cluster N : ",newcluster);
  msd_buffer->mots_16b[cluster&0xFF]=0x0; // "Effacer" le cluster courant
  if ((newcluster>=0xFFF8)||(newcluster==0x0000)) {break;}
  if ((cluster/256)!=(newcluster/256)) 
  {// Cluster suivant dans un autre secteur (FAT16) :
   WTX2_Msg_Nb32bits(" Ecriture ds le secteur N : ",SD_CurrSector);
   SectorWrite(SD_CurrSector,msd_buffer);// Ecriture du buffer ds le secteur courant
   SD_CurrSector=FAT_StartSector+(newcluster/256);//Premier secteur du cluster suivant
   // Transfert du secteur ds le buffer pour la boucle suivante
   SectorRead(SD_CurrSector,msd_buffer);
  }
 }
 WTX2_Msg_Nb32bits(" Ecriture ds le secteur N : ",SD_CurrSector);
 SectorWrite(SD_CurrSector,msd_buffer);// Ecriture du buffer ds le dernier secteur
 // "Effacer" l'entre de la structure de fichier
 SD_CurrSector=File_DirSectorNum; // Secteur de la structure de fichier  effacer
 WTX2_Msg_Nb32bits(" Effacer l'entree de la structure de fichier ds le secteur : ",SD_CurrSector);
 SectorRead(SD_CurrSector,msd_buffer);// Transfert du secteur ds le buffer
 // Affectation du 1 octet : structure libre
 WriteLnTXD2_String("Liberer la structure d'entree de fichier.");
 msd_buffer->FATFilesInfos[File_DirEntryNum].DIR_Name[0]=FAT_FILE_DELETED;
 WTX2_Msg_Nb32bits(" Ecriture ds le secteur N : ",SD_CurrSector);
 SectorWrite(SD_CurrSector,msd_buffer);// Ecriture du buffer ds le dernier secteur
 return TRUE;
}

/*------------------------------------------------------------------
  Function   : void FAT_Init(void)
  Description: Affectation des variables lie  la FAT de la SD Card  
               La fonction MediaInitialize() a t excute avec succs   
  Retour     : TRUE si tout s'est bien pass
               FALSE en cas d'anomalie : la SD Card n'est pas accessible
--------------------------------------------------------------------*/
int FAT_Init(void)
{
 char Test;
 WriteLnTXD2();
 WriteLnTXD2_String("Nouvelle SD Card :");
 WriteLnTXD2_String("******************");
 SD_CurrSector=0;
 if (SDC_Status!=sdcValid) return FALSE;
 // Lecture du secteur MBR (secteur 0)
 if (SectorRead(0,msd_buffer)!=sdcValid) return FALSE;

 // Test de validit
 Test=msd_buffer->octets[0x1C2]; // Code du type de partition
 if ((Test!=0x04)&&(Test!=0x06)&&(Test!=0x0E)) return FALSE;

 // Lecture de l'adresse du 1 secteur de la 1 partition
 BPB_StartSector=(uint32)msd_buffer->mots_16b[454/2]
                 +((uint32)(msd_buffer->mots_16b[1+454/2])<<16);
 WTX2_Msg_Nb32bits("Secteur de boot (BPB_StartSector)= ",BPB_StartSector);

 // Lecture du secteur de boot (infos BPB de la FAT)
 if (SectorRead(BPB_StartSector,msd_buffer)!=sdcValid) return FALSE;
 // Affectations des paramtres et variables de la FAT
 BPB_BytesPerSector=     msd_buffer->BPB_FAT.BPB_BytsPerSec[0]
                    +256*msd_buffer->BPB_FAT.BPB_BytsPerSec[1];
 WTX2_Msg_Nb16bits("Nombre d'octets par secteur (BPB_BytesPerSector)= "
                   ,BPB_BytesPerSector);

 BPB_SectorsPerCluster=msd_buffer->BPB_FAT.BPB_SecPerClus;
 WTX2_Msg_Nb16bits("Nombre de secteurs par cluster (BPB_SectorsPerCluster)= "
                   ,BPB_SectorsPerCluster);

 BPB_ReservedSectorCount=msd_buffer->BPB_FAT.BPB_RsvdSecCnt;
 WTX2_Msg_Nb16bits("Nombre de secteurs reserves (BPB_ReservedSectorCount)= "
                   ,BPB_ReservedSectorCount);

 // On utilise un secteur rserv comme tampon d'criture
 SecteurTampon=BPB_StartSector+1; 
 WTX2_Msg_Nb32bits("Secteur reserve pour le buffer d'criture (SecteurTampon):"
                   ,SecteurTampon);

 BPB_NumFATs=msd_buffer->BPB_FAT.BPB_NumFATs;
 WTX2_Msg_Nb16bits("Nombre de FATs (BPB_NumFATs)= ",BPB_NumFATs);

 BPB_RootDirEntryCount=     msd_buffer->BPB_FAT.BPB_RootEntCnt[0]
                       +256*msd_buffer->BPB_FAT.BPB_RootEntCnt[1];
 WTX2_Msg_Nb16bits("Nombre d'entrees de repertoire (BPB_RootDirEntryCount)= "
                   ,BPB_RootDirEntryCount);

 BPB_TotSec=(uint32)msd_buffer->BPB_FAT.BPB_TotSec16[0]
            +256*(uint32)msd_buffer->BPB_FAT.BPB_TotSec16[1];
 if (BPB_TotSec==0) BPB_TotSec=msd_buffer->BPB_FAT.BPB_TotSec32;
 WTX2_Msg_Nb32bits("Nombre total de secteurs (BPB_TotSec)= ",BPB_TotSec);

 if (msd_buffer->BPB_FAT.BPB_FATSz16!=0) BPB_FATSize=msd_buffer->BPB_FAT.BPB_FATSz16;
                         else BPB_FATSize=msd_buffer->BPB_FAT.BPB_FATSz32;
 WTX2_Msg_Nb32bits("Taille de la FAT en secteurs (BPB_FATSize)= ",BPB_FATSize);

 FAT_StartSector=BPB_StartSector+BPB_ReservedSectorCount;
 WTX2_Msg_Nb16bits("1 secteur du bloc FAT (FAT_StartSector) = ",FAT_StartSector);

 FAT_RootDirNumSectors=((BPB_RootDirEntryCount*32)+(BPB_BytesPerSector-1))/BPB_BytesPerSector;
 WTX2_Msg_Nb16bits("Nombre de secteurs du repertoire racine (FAT_RootDirNumSectors)= "
                   ,FAT_RootDirNumSectors);

 BPB_DataSec=BPB_TotSec-(BPB_ReservedSectorCount+(BPB_NumFATs*BPB_FATSize)+FAT_RootDirNumSectors);
 WTX2_Msg_Nb32bits("Nombre de secteurs dans la zone de donnees (BPB_DataSec)= "
                   ,BPB_DataSec);

 BPB_CountofClusters=BPB_DataSec/BPB_SectorsPerCluster;
 WTX2_Msg_Nb16bits("Nombre de cluster (BPB_CountofClusters)= ",BPB_CountofClusters);

 // Type de FAT
 if (BPB_CountofClusters<4085) 
  {
   FATType=FAT12;
   WriteLnTXD2_String("FAT12");
  } 
  else if (BPB_CountofClusters<65525) 
   {
    FATType=FAT16;
    WriteLnTXD2_String("FAT16");
   }
   else 
    {
     FATType=FAT32;
     WriteLnTXD2_String("FAT32");
    }

 FAT_RootDirStartSector=FAT_StartSector+BPB_NumFATs*BPB_FATSize;
 WTX2_Msg_Nb16bits("Premier secteur du repertoire racine (FAT_RootDirStartSector)= ",
                    FAT_RootDirStartSector);

 FAT_FirstDataSector=BPB_StartSector+BPB_ReservedSectorCount+(BPB_NumFATs*BPB_FATSize)+FAT_RootDirNumSectors;
 //FAT_FirstDataSector=FAT_RootDirStartSector+FAT_RootDirNumSectors;
 WTX2_Msg_Nb16bits("Premier secteur de donnee (FAT_FirstDataSector)= "
                   ,FAT_FirstDataSector);

 return TRUE;
}
