     /********************************************************************
      ***                      Gestion SD Card                         ***
      ***               Fonctions de gestion de la FAT                 ***
      ***                      dsPIC30F4011                            ***
      ***                                                              ***
      ***  Fichier "entte" : FAT.h                                    ***
      ***  Auteur : CREMMEL Marcel                                     ***
      ***  Date   : 30/09/2007                                         ***
      ********************************************************************/

#ifndef FAT_Defs
#define FAT_Defs

/*------------------------------------------------------------------------------
                Prototypes des fonctions en mmoire flash
------------------------------------------------------------------------------*/

/*---------------------------------------------------------------
  Function   : void Reset_FAT(void)
  Description: Initialisation de la structure BPB pour les tests     
               (infos prises sur SD Card 1G en FAT16)
-----------------------------------------------------------------*/
void Reset_FAT(void);

/*---------------------------------------------------------------
  Function   : void SampleFileData(Secteur *buffer)
  Description: Affectation de qq strucures de fichiers ds la FAT   
               Utilis pour des tests             
  Paramtre  : Pointeur vers un secteur structur "Secteur" de 
               512 octets
-----------------------------------------------------------------*/
void SampleFileData(Secteur *buffer);

/*------------------------------------------
  Function:    void QuickFormat()
  Description: formatage rapide du volume
               A n'utiliser qu'en urgence !
---------------------------------------------*/
void QuickFormat(void);

/*---------------------------------------------------------------------
  Function:    int OpenFileWithName(char *filename)
  Description: Ouvrir un fichier ds le rpertoire racine avec son nom
  Paramtre  : Nom court du fichier avec son extension, mais sans le "."
  Retour     : Renvoi 
                 - du n du fichier en cas de succs
                 - de -1 en cas d'erreur
-----------------------------------------------------------------------*/
int OpenFileWithName(char *filename);

/*----------------------------------------------------------------------
  Function:    int ReadNextSectorIntoBuff(void)
  Description: Lecture du secteur suivant du cluster courant ds le buffer
  Retour     : Renvoi de 
                  - TRUE si l'opration s'est bien passe
                  - FALSE ds le cas contraire
  Le fichier doit tre "ouvert" : affectations des variables "File_..."
------------------------------------------------------------------------*/
int ReadNextSectorIntoBuff(void);

/*-------------------------------------------------------------------------
  Function:    int WriteBuffIntoNextSector(void)
  Description: Ecriture du secteur tampon dans le secteur suivant du fichier 
               ouvert (soit dans le secteur suivant du cluster courant)
               Mise  jour des 2 clusters concerns
               Mise  jour de la taille du fichier
  Retour     : TRUE ou FALSE suivant l'opration effectue
  Le fichier doit tre "ouvert" : affectations des variables "File_..."
---------------------------------------------------------------------------*/
int WriteBuffIntoNextSector(void);

/*-----------------------------------------------------
  Function:    void ReadFile(char *name)
  Description: Lecture d'un fichier du rpertoire racine, 
               sortie sur UART2
  Paramtre  : Nom complet du fichier (8+3 caractres)
-------------------------------------------------------*/
void ReadFile(char *name);

/*-------------------------------------------------------------
  Function:    void ListDir(void)
  Description: Transmission sur TX2 de la liste des fichiers du 
               rpertoire racine
---------------------------------------------------------------*/
void ListDir(void);

/*---------------------------------------------------------
  Function:    int CreateFile(char *name, uint32 len)
  Description: Cration d'un fichier dans le rpertoire racine
  Paramtre  : Nom complet du fichier (8+3 caractres)
  Retour     : TRUE ou FALSE suivant la russite l'opration
------------------------------------------------------------*/
int CreateFile(char *name);

/*--------------------------------------------------------------
  Function:    int DeleteFile(char *name)
  Description: Effacer un fichier ds le rpertoire racine
  Paramtre  : Nom complet du fichier  effacer (8+3 caractres)
  Retour     : TRUE ou FALSE suivant la russite l'opration
----------------------------------------------------------------*/
int DeleteFile(char *name);

/*------------------------------------------------------------------
  Function:    void FAT_Init(void)
  Description: Affectation des variables lie  la FAT de la SD Card     
--------------------------------------------------------------------*/
int FAT_Init(void);

#endif // FAT_Defs
