     /********************************************************************
      ***                      Projet VAE                              ***
      ***             Programme de test d'une carte SD                 ***
      ***                      dsPIC30F4011                            ***
      ***                                                              ***
      ***  Auteur : CREMMEL Marcel                                     ***
      ***  Version pour maquette de tests                              ***
      ***  Date : 08/09/2007                                           ***
      ***  Ressources utilises : SPI, Timer1, Timer 3, UART1, UART2   ***
      ***  Note : pas de gestion des rpertoires : tous les fichiers   ***
      ***         doivent se trouver ds le rpertoire "racine"         ***
      ********************************************************************/

#include <p30f4011.h>
#include <string.h>
#include <stdio.h>
#include "Typedefs.h"    // De Microchip : BYTE, WORD, DWORD, POINTER
#include "Typedefs_SD.h" // SDC_Error, CSD, CID, SDCSTATE, sdmmc_cmd, RESP, CMD_PACKET
                         // REPONSE_1, REPONSE_2, SDC_REPONSE, typSDC_CMD
#include "Typedefs_FAT.h"// uint16, uint, uint32, BPB_Sector; FSInfo
#include "Typedefs_NMEA.h"  // tGGA

//--------------------------Device Configuration------------------------       
_FOSC(CSW_FSCM_OFF & XT_PLL16);//Turn off clock switching and
                               //fail-safe clock monitoring and
                               //use the XT Clock x 16 as the system clock
_FWDT(WDT_OFF);                //Turn off Watchdog Timer
_FBORPOR(PBOR_ON & BORV_27 & PWRT_16 & MCLR_EN);
                               //Set Brown-out Reset voltage and
                               //and set Power-up Timer to 16msecs
_FGS(CODE_PROT_OFF);           //Set Code Protection Off for the
                               //General Segment
//----------------------------------------------------------------------


                        /*****************************
                        *          Cblage           *
                        *****************************/

/*****************************
*          SD Card           *
******************************/
#define SDC_CS        LATDbits.LATD3   // RD3=CS_SD: entre CS de la SD Card
#define SDC_CS_DIR    TRISDbits.TRISD3
#define SDC_CD        PORTDbits.RD0    // RD0=CD_SD: contact de prsence SD Card
#define SDC_CD_DIR    TRISDbits.TRISD0
#define SDC_WP		  PORTDbits.RD2    // RD2=WP_SD: switch de protection en criture
#define SDC_WP_DIR	  TRISDbits.TRISD2
#define SDC_POWER     PORTBbits.RB8    // RB8=POWER_SD: commande alim. carte SD
#define SDC_POWER_DIR TRISBbits.TRISB8
/* Broches ddies :
  - 26 : SDI1=RF2 connect  SDO_SD (sortie donnes SD Card)
  - 25 : SDO1=RF3 connect  SDI_SD (entre donnes SD Card)
  - 24 : SCK1=RF6 connect  CLK_SD (horloge donnes SD Card) */

/*****************************
*       I2C "software"       *
******************************/
#define Set_SDA  TRISFbits.TRISF1=1 // SDA=HiZ: mise  "1" par rsistance "pull-up" 
#define Clr_SDA  TRISFbits.TRISF1=0 // SDA="0": mise  "0" car LATFbits.RF1="0"
#define Init_SDA LATFbits.RF1=0     // Pour forcer l'tat "0" si RF1=SDA en sortie 
#define Set_SCL  TRISFbits.TRISF0=1 // SDA=HiZ: mise  "1" par rsistance "pull-up" 
#define Clr_SCL  TRISFbits.TRISF0=0 // SDA="0": mise  "0" car LATFbits.RF0="0"
#define Init_SCL LATFbits.RF0=0     // Pour forcer l'tat "0" si RF0=SCL en sortie 

/*****************************
*      Leds indicatrices     *
******************************/
#define STATUSLED
#define STRTRIS TRISEbits.TRISE2
#define STRLED  PORTEbits.RE2     // Indicateur de lecture : STatus Read LED sur RE2
#define STWTRIS TRISEbits.TRISE1
#define STWLED  PORTEbits.RE1     // Indicateur d'criture : STatus Write LED sur RE1

/**********************************
* Communications srie asynchrone *
***********************************
Avec le PC via UM232R : UART2 : U2RX (broche 28) & U2TX (broche 27) 
Avec le GPS           : UART1 : U1RX (broche 16) & U1TX (broche 15) */

/*****************************
*       Signaux de test
******************************
TEST0 sur port RE3 (broche 35) 
TEST1 sur port RE4 (broche 34) */
#define TEST0 LATEbits.LATE3 
#define TEST1 LATEbits.LATE4
#define TEST0_Dir TRISEbits.TRISE3
#define TEST1_Dir TRISEbits.TRISE4

                        /*****************************
                        * Constantes non mmorises  *
                        ******************************/
                        
#define FCY     (dword)4000000*16/4 // xtal = 4Mhz; PLLx16 -> 16 MIPS
#define FALSE	0
#define TRUE	!FALSE
#define INPUT	1
#define OUTPUT	0

// SD-Card
#define MOREDATA        !0
#define NODATA          0
#define SDC_SECTOR_SIZE 512
#define CSD_SIZE        16   // 16 octets du registre CSD

/* Command Operands */
#define BLOCKLEN_64                 0x0040
#define BLOCKLEN_128                0x0080
#define BLOCKLEN_256                0x0100
#define BLOCKLEN_512                0x0200

#define SDC_FLOATING_BUS    		0xFF
#define SDC_BAD_RESPONSE    		SDC_FLOATING_BUS
#define SDC_ILLEGAL_CMD     		0x04
#define SDC_GOOD_CMD        		0x00

/* Data Token */
#define DATA_START_TOKEN            0xFE
#define DATA_MULT_WRT_START_TOK     0xFC
#define DATA_MULT_WRT_STOP_TOK      0xFD

/* Data Response */
#define DATA_ACCEPTED               0b00000101
#define DATA_CRC_ERR                0b00001011
#define DATA_WRT_ERR                0b00001101

// Commandes "SD Card" (SPI Bus Mode)
// Bit 7 = 0, Bit 6 = 1, Bits 5-0=Command Code
#define     cmdGO_IDLE_STATE        0x40		//0 : code de la cmd sur 6 bits LSB
#define     cmdSEND_OP_COND         0x41		//1        
#define     cmdSEND_CSD             0x49		//9
#define     cmdSEND_CID             0x4a		//10
#define     cmdSTOP_TRANSMISSION    0x4c		//12
#define     cmdSEND_STATUS          0x4d		//13
#define     cmdSET_BLOCKLEN         0x50		//16
#define     cmdREAD_SINGLE_BLOCK    0x51		//17
#define     cmdREAD_MULTI_BLOCK     0x52		//18
#define     cmdWRITE_SINGLE_BLOCK   0x58		//24    
#define     cmdWRITE_MULTI_BLOCK    0x59		//25
#define     cmdTAG_SECTOR_START     0x60		//32
#define     cmdTAG_SECTOR_END       0x61		//33
#define     cmdUNTAG_SECTOR         0x62		//34
#define     cmdTAG_ERASE_GRP_START  0x63		//35 
#define     cmdTAG_ERASE_GRP_END    0x64		//36
#define     cmdUNTAG_ERASE_GRP      0x65		//37
#define     cmdERASE                0x66		//38
#define     cmdSD_APP_OP_COND       0x69		//41
#define     cmdLOCK_UNLOCK          0x71		//49
#define     cmdAPP_CMD              0x77		//55
#define     cmdREAD_OCR             0x7a		//58
#define     cmdCRC_ON_OFF           0x7b		//59

// Liaison via UART
#define NbDtaTxd 6               // Nombre de paramtres transmis sur TXD2
#define Synchro  'Z'             // Caractre de dtection des trames UART2
#define CR       13              // Retour chariot
#define LF       10              // Saut de ligne
                      
                        /*****************************
                        *           Macros           *
                        ******************************/

//#define mSendMediaCmd_NoData() SendMediaCmd(); SDC_CS=1;

                        /*****************************
                        *          Variables         *
                        ******************************/

/* Indicateurs de fonctionnement
********************************/
struct 
{
 unsigned SDC_CD_OK : 1; // Dtection prsence carte (avec anti-rebonds)
 unsigned SDC_CD_old: 1; // Etat prcdent du contact CD
 unsigned SDC_CD_New: 1; // Nouvel tat stable du contact CD
 unsigned isSDMMC   : 1; // Indicateur de carte SDMMC 
 unsigned isWP      : 1; // Indicateur de protection en criture
 unsigned unused	: 11;
}Flags;

/* Buffers RX et TX pour l'UART1 (communication GPS)
****************************************************/
#define RX1_BufSize 96          // Taille du Buffer_RX1
char Buffer_RX1[RX1_BufSize];   // Buffer circulaire de rception
char *ptrRX1_WRdata=Buffer_RX1; // Pointeur d'criture ds Buffer_RX1
char *ptrRX1_RDdata=Buffer_RX1; // Pointeur de lecture ds Buffer_RX1
char *ptrRX1_Start_NMEA;        // Pointeur du dernier caractre $ reu
char *ptrRX1_Trame_NMEA;        // Pointeur pour lire la trame 
#define TX1_BufSize 16          // Taille du Buffer_TX1
char Buffer_TX1[TX1_BufSize];   // Buffer circulaire de transmission
char *ptrTX1_WRdata=Buffer_TX1; // Pointeur d'criture ds Buffer_TX1
char *ptrTX1_RDdata=Buffer_TX1; // Pointeur de lecture ds Buffer_TX1

/* Buffers RX et TX pour l'UART2 (communication PC)
***************************************************/
#define RX2_BufSize 16          // Taille du Buffer_RX2
char Buffer_RX2[RX2_BufSize];   // Buffer circulaire de rception
char *ptrRX2_WRdata=Buffer_RX2; // Pointeur d'criture ds Buffer_RX2
char *ptrRX2_RDdata=Buffer_RX2; // Pointeur de lecture ds Buffer_RX2
#define TX2_BufSize 512         // Taille du Buffer_TX2
char Buffer_TX2[TX2_BufSize];   // Buffer circulaire de transmission
char *ptrTX2_WRdata=Buffer_TX2; // Pointeur d'criture ds Buffer_TX2
char *ptrTX2_RDdata=Buffer_TX2; // Pointeur de lecture ds Buffer_TX2
unsigned char Buffer2_Cmd[5];   // Buffer pour la rception des commandes

/* GPS
******/
tGGA GGA; // Global Positioning System Fixed Data
tGSA GSA; // GNSS DOP and Active Satellites
tGSV GSV; // GNSS Satellites in View
tRMC RMC; // Recommended Minimum Specific GNSS Data
tVTG VTG; // Course Over Ground and Ground Speed

/* Divers
*********/
uint TMR3_MSB;     // Poids fort de TMR3
uint Compt_T1_TXD; // Compteur d'interruption Timer 1 pour transmission datas

/**********
* SD-Card *
***********/

/* Couche physique
*****************/
int Cpt_AR_SDC_CD;   // Compteur anti-rebonds dtection carte
SDC_Error SDC_Status;// Etat de la carte

/* Dclaration multitypes du buffer pour un secteur
***************************************************/
typedef union
{
 struct
 {// 1 secteur = 512 octets
  byte octets[512];
 };
 struct
 {// 1 secteur = 256 mots de 16 bits
  uint16 mots_16b[256];
 };
 struct
 {// 1 secteur = 128 mots de 32 bits
  uint32 mots_32b[128];
 };
 BPB_Sector BPB_FAT;              // Structure BPB du secteur "boot"
 tFATFileInfos FATFilesInfos[16]; // 16 structures d'entres de fichier
} Secteur;
Secteur msd_buffer;   // Buffer de lecture et d'criture d'un secteur

/* Registre CSD
**************/
CSD gblCSDReg;        // Registre CSD : Card Specific Data (128 bits ou 16 octets)

/* Variables "courantes"
************************/
uint32 SD_CurrSector; // Adresse du secteur courant
uint32 SecteurTampon; // Adresse du secteur utilis comme tampon d'criture

/* BPB
******/
uint16 BPB_StartSector;         // Adresse du secteur BPB
uint16 BPB_ReservedSectorCount; // Nb de secteurs rservs derrire BPB 
byte  BPB_NumFATs;              // Nombre de blocs FAT (en principe 2)
byte  BPB_SectorsPerCluster;    // Nombre de secteurs par cluster
uint16 BPB_BytesPerSector;      // Nombre d'octets par secteur (souvent 512)
uint32 BPB_FATSize;             // Taille du bloc FAT en octets
uint32 BPB_TotSec;              // Nombre total de secteurs
uint32 BPB_DataSec;             // Nombre de secteurs du bloc de donnes
uint16 BPB_CountofClusters;     // Nombre total de clusters
uint16 BPB_RootDirEntryCount;   // Nombre de structures de rpertoires ou fichiers

/* FAT
******/
FAT_type FATType;               // Type de FAT de la SD Card
uint16 FAT_StartSector;         // Adresse du 1 secteur du bloc FAT (passer  32 bits ?)
uint32 FAT_FirstDataSector;     // Adresse du 1 secteur du bloc de donnes
uint32 FAT_RootDirStartSector;  // Adresse du 1 secteur du bloc des structures de fichier
uint16 FAT_RootDirNumSectors;   // Nb de secteur du bloc des structures de fichier

/* Variables lies aux fichiers et aux clusters
***********************************************/
char  File_Name[11];            // Nom court du fichier (ou rpertoire)
char  End_File_Name=0;          // Par prudence !
char  File_Attr;                // Attributs du fichier ou rpertoire  
byte  File_CurrSectorNumInCluster;// Indice du secteur courant ds un cluster
uint16 File_StartCluster;       // N du 1 cluster du fichier (32 bits en FAT32)
uint16 File_CurrCluster;        // N du cluster courant (32 bits en FAT32)
uint32 File_Size;               // Taille du fichier en octets
uint32 File_StartSector;		// Premier secteur de donnes du 1 cluster
uint32 File_CurrSector;         // Adresse du secteur courant ds le bloc de donnes
uint32 File_CurrByte;           // Compteur d'octets lus ou crits
uint32 File_DirSectorNum;       // Adresse du 1 secteur de la structure de fichier
uint16 File_DirEntryNum;        // Indice ds le secteur de la structure d'entre de fichier
                       
/* Constantes spcifiques
*************************/
// Masques des attributs de fichier
#define FAT_IS_DIR        0x10
#define FAT_END_OF_DIR    0x02
#define FAT_FILE_DELETED  0xE5
#define FAT_LONG_FILENAME 0x03

                        /*****************************
                        *    Constantes mmorises   *
                        ******************************/

/*********************************************************************
 Table "sdmmc_cmdtable"
  - Informations sur chaque commande "sdmmc" :
     - code de la commande
     - valeur de l'ventuel CRC
     - nature de la rponse (R1, R1b, R2 ou R3)
     - si la commande est suivie de donnes
 Notes: la fonction CRC est inhibe ds que possible ou viter son calcul
        (valeur 0xFF dans ce cas) 
**********************************************************************/
const typSDC_CMD sdmmc_cmdtable[] =
{
 //     cmd               crc   response  Donnes
 {cmdGO_IDLE_STATE,       0x95,   R1,      NODATA},   // CMD0 : Reset to Idle State
 {cmdSEND_OP_COND,        0xF9,   R1,      NODATA},   // CMD1 : Dclenche le processus d'init.
 {cmdSEND_CSD,            0xAF,   R1,      MOREDATA}, // CMD9 : Lecture du registre CSD
 {cmdSEND_CID,            0x1B,   R1,      MOREDATA}, // CMD10: Lecture du registre CID
 {cmdSTOP_TRANSMISSION,   0xC3,   R1,      NODATA},   // CMD12: Termine une op. R/W blk mult
 {cmdSEND_STATUS,         0xAF,   R2,      NODATA},   // CMD13: Lecture du registre de status
 {cmdSET_BLOCKLEN,        0xFF,   R1,      NODATA},   // CMD16: Affecte la longueur des blocs
 {cmdREAD_SINGLE_BLOCK,   0xFF,   R1,      MOREDATA}, // CMD17: Lecture d'un bloc
 {cmdREAD_MULTI_BLOCK,    0xFF,   R1,      MOREDATA}, // CMD18: Lecture continuelle de blocs
 {cmdWRITE_SINGLE_BLOCK,  0xFF,   R1,      MOREDATA}, // CMD24: Ecriture d'un bloc
 {cmdWRITE_MULTI_BLOCK,   0xFF,   R1,      MOREDATA}, // CMD25: Ecriture continuelle de blocs
 {cmdTAG_SECTOR_START,    0xFF,   R1,      NODATA},   // CMD32: Adresse dbut  effacer
 {cmdTAG_SECTOR_END,      0xFF,   R1,      NODATA},   // CMD33: Adresse fin  effacer
 {cmdUNTAG_SECTOR,        0xFF,   R1,      NODATA},   // CMD34
 {cmdTAG_ERASE_GRP_START, 0xFF,   R1,      NODATA},   // CMD35
 {cmdTAG_ERASE_GRP_END,   0xFF,   R1,      NODATA},   // CMD36
 {cmdUNTAG_ERASE_GRP,     0xFF,   R1,      NODATA},   // CMD37
 {cmdERASE,               0xDF,   R1b,     NODATA},   // CMD38: Effacer les blocs slectionns
 {cmdLOCK_UNLOCK,         0x89,   R1b,     NODATA},   // CMD49
 {cmdSD_APP_OP_COND,      0xE5,   R1,      NODATA},   // ACMD41: Dclenche l'initialisation
 {cmdAPP_CMD,             0x73,   R1,      NODATA},   // CMD55: La commande suivante est de type
                                                      //        "Application specific"
 {cmdREAD_OCR,            0x25,   R3,      NODATA},   // CMD58: Lecture registre OCR
 {cmdCRC_ON_OFF,          0x25,   R1,      NODATA}    // CMD59: CRC ON ou OFF
};

                        /*****************************
                        *         Fonctions          *
                        ******************************/
void WriteTXD2(char c);
SDC_Error SectorRead(uint32 sector_addr, Secteur* buffer);
void SDC_CD_AR(void);

/******************************************************************************
*                            Fonctions diverses                               *
*******************************************************************************/
#define Delai_100ms 6250 // 6250/62500
#define Delai_40ms  2500 // 2500/62500
#define Delai_10ms  625  // 625/62500
#define Delai_1ms   63   // 63/62500
#define Delai_500us 31   // 31/62500
#define Delai_48uS  3    // 3/62500
#define Delai_mini  1    // 1/62500 = 16S
void Delay(uint16 Delai)
{
 uint32 Temps_debut,Temps_actuel;
 IEC0bits.T3IE=0; // Inhibition interruption Timer 3 pdt lecture TMR3
 Temps_debut=TMR3+((unsigned long)TMR3_MSB<<16);//Lecture de TMR3 sur 32 bits
 IEC0bits.T3IE=1; // Validation interruption Timer 3
 do
 {
  IEC0bits.T3IE=0; // Inhibition interruption Timer 3 pdt lecture TMR3
  Temps_actuel=TMR3+((unsigned long)TMR3_MSB<<16);//Lecture de TMR3 sur 32 bits
  IEC0bits.T3IE=1; // Validation interruption Timer 3
 } while ((Temps_actuel-Temps_debut)<=Delai);
}

/******************************************************************************
*                                   UART1                                     *
*******************************************************************************/

/*********************************************************************
  Function:        void InitUART1(void)
  Description:     Initialisation de l'UART1
                     - Sur ports "auxiliaires" U1ATX et U1ARX
                     - 4800 bauds, 8 bits, pas de parit, 1 bit Stop
                     - interruption sur TX si buffer UART vide
                     - interruption sur RX  chaque caractre
********************************************************************/
void InitUART1(void)
{
 U1MODE=0x8400;     // UART valid, 8 bits, pas de parit, 1 bit stop
                    // Sur ports "auxiliaires" U1ATX et U1ARX 
 //U1MODE=0x8000;     // UART valid, 8 bits, pas de parit, 1 bit stop
 U1STA =0x8400;     // Interruption TX si buffer UART vide
                    // UART TX valid
                    // Interruption RX  chaque caractre
 U1BRG =((FCY/16)/4800)-1; // 4800 bauds
}

/*********************************************************************
  Function:        int ReadRXD1(char *c)
  Description:     Lecture d'un octet dans le buffer de rception
                     - renvoi de 0 si aucun caractre reu
                     - renvoi de 1 si caractre reu :
                        - si c=='$' : ptrRX1_Start_NMEA affect avec indice
                          du premier caractre de la trame NMEA
                        - caractre renvoy dans "c"
                        - pointeur circulaire ptrRX1_RDdata incrment
********************************************************************/
int ReadRXD1(char *c)
{
 char c_bis;
 if (ptrRX1_RDdata==ptrRX1_WRdata) return(0); // Pas de caractre reu
 else
 {
  *c=*ptrRX1_RDdata++; // Affecter "c" avec le caractre reu puis
                       // incrmenter le pointeur ptrRX1_RDdata
  c_bis=*c;
  // Pointeur circulaire
  if (ptrRX1_RDdata==Buffer_RX1+RX1_BufSize) ptrRX1_RDdata=Buffer_RX1;
  if (*c=='$') ptrRX1_Start_NMEA=ptrRX1_RDdata; // Indice dbut trame NMEA
  return(1);
 }
}

/*********************************************************************
  Function:        void WriteTXD1(char c)
  Description:     Ecriture d'un octet dans le buffer d'mission
                     - attendre libration du buffer
                     - envoi du caractre reu
                     - pointeur circulaire ptrTX1_WRdata incrment
********************************************************************/
void WriteTXD1(char c)
{
 int Nb_char;
 // Attendre libration d'une partie du buffer
 do
  {// Calcul du nb de caractres restant  transmettre
   Nb_char=ptrTX1_WRdata-ptrTX1_RDdata;
   if (Nb_char < 0) Nb_char=Nb_char+TX1_BufSize;
  } 
 while (Nb_char > (TX1_BufSize-8));
 // Caractre plac dans le buffer circulaire
 *ptrTX1_WRdata++=c; 
 if (ptrTX1_WRdata==Buffer_TX1+TX1_BufSize) ptrTX1_WRdata=Buffer_TX1;
 // Provoquer la 1 interruption si aucune transmission en cours
 // Le buffer de l'UART fait 4 octets (Nb_char=3)
 if ((U1STAbits.TRMT)&&(Nb_char>=3)) IFS0bits.U1TXIF=1; 
}

/*********************************************************************
  Function:        void WriteIntTXD1(int i)
  Description:     Ecriture d'un entier dans le buffer d'mission
********************************************************************/
void WriteIntTXD1(int i)
{
 WriteTXD1(i>>8); // Octet MSB d'abord
 WriteTXD1(i);    // Puis octet LSB
}

/*********************************************************************
  Function:        void WriteTXD1_String(char *String)
  Description:     Ecriture d'une chane de caractres dans le buffer TX
********************************************************************/
void WriteTXD1_String(char *String)
{
 while (*String) // Attendre caractre final (0)
 {
  WriteTXD1(*String++); // Transmission du caractre suivant
 }
}

/*********************************************************************
  Function:        void WriteLnTXD1_String(char *String)
  Description:     Ecriture d'une chane de caractres dans le buffer TX
                   "Retour chariot" (CR+LF) ajout  la fin
********************************************************************/
void WriteLnTXD1_String(char *String)
{
 WriteTXD1_String(String);   // Transmission de la chane
 WriteTXD1(CR); WriteTXD1(LF);// Transmission du "retour chariot"
}

/*********************************************************************
  Function   : void Receive_NMEA_RXD1(void)
  Description: Rception des trames NMEA : GGA, GSA, GSV, RMC et VTG
********************************************************************/
void Receive_NMEA_RXD1(void)
{
 int i,j,k;
 unsigned char c, Checksum;
 if (!ReadRXD1(&c)) return; // Retour immdiat si aucun nouveau car. reu
 WriteTXD2(c); // **** TEST **** Renvoi sur TX2 (USB)
 if (c==LF)
 {// Une trame NMEA complte se trouve ds le buffer Buffer_RX1  partir
  // de l'indice ptrRX1_Start_NMEA
  Checksum=0;
  if (ptrRX1_Start_NMEA==0) return;    // Fausse trame ?
  ptrRX1_Trame_NMEA=ptrRX1_Start_NMEA; // Dbut de la trame
  ptrRX1_Start_NMEA=0; // Pour dtecter une fausse trame
  do
  {// Les premiers caractres (virgule comprise) identifient la trame
   c=*ptrRX1_Trame_NMEA++;
   Checksum^=c;
   // Pointeur circulaire
   if (ptrRX1_Trame_NMEA==Buffer_RX1+RX1_BufSize) ptrRX1_Trame_NMEA=Buffer_RX1;
  }
  while (c!=',');
  i=j=k=0;
  switch(Checksum) // La somme de contrle sur les 6 premiers caractres
                   // de la trame l'identifie
  {
   case 0x7A : // Trame GGA
   {
    do
    {
     c=*ptrRX1_Trame_NMEA++;
     if (ptrRX1_Trame_NMEA==Buffer_RX1+RX1_BufSize) ptrRX1_Trame_NMEA=Buffer_RX1;
     if (c!='*') Checksum^=c;
     if ((c!=',') && (c!='*')) GGA.octets[i++]=c;
     else
     {
      k+=Tab_Size_GGA[j++];
      while (i!=k) GGA.octets[i++]=0; 
     }
    }
    while (c!='*');
    break;
   }
   case 0x6E : // Trame GSA
   {
    do
    {
     c=*ptrRX1_Trame_NMEA++;
     if (ptrRX1_Trame_NMEA==Buffer_RX1+RX1_BufSize) ptrRX1_Trame_NMEA=Buffer_RX1;
     if (c!='*') Checksum^=c;
     if ((c!=',') && (c!='*')) GSA.octets[i++]=c;
     else
     {
      k+=Tab_Size_GSA[j++];
      while (i!=k) GSA.octets[i++]=0; 
     }
    }
    while (c!='*');
    break;
   }
   case 0x79 : // Trame GSV
   {
    do
    {
     c=*ptrRX1_Trame_NMEA++;
     if (ptrRX1_Trame_NMEA==Buffer_RX1+RX1_BufSize) ptrRX1_Trame_NMEA=Buffer_RX1;
     if (c!='*') Checksum^=c;
     if ((c!=',') && (c!='*')) GSV.octets[i++]=c;
     else
     {
      k+=Tab_Size_GSV[j++];
      while (i!=k) GSV.octets[i++]=0; 
     }
    }
    while (c!='*');
    break;
   }
   case 0x67 : // Trame RMC
   {
    do
    {
     c=*ptrRX1_Trame_NMEA++;
     if (ptrRX1_Trame_NMEA==Buffer_RX1+RX1_BufSize) ptrRX1_Trame_NMEA=Buffer_RX1;
     if (c!='*') Checksum^=c;
     if ((c!=',') && (c!='*')) RMC.octets[i++]=c;
     else
     {
      k+=Tab_Size_RMC[j++];
      while (i!=k) RMC.octets[i++]=0; 
     }
    } 
    while (c!='*');
    break;
   }
   case 0x7E : // Trame VTG
   {
    do
    {
     c=*ptrRX1_Trame_NMEA++;
     if (ptrRX1_Trame_NMEA==Buffer_RX1+RX1_BufSize) ptrRX1_Trame_NMEA=Buffer_RX1;
     if (c!='*') Checksum^=c;
     if ((c!=',') && (c!='*')) VTG.octets[i++]=c;
     else
     {
      k+=Tab_Size_VTG[j++];
      while (i!=k) VTG.octets[i++]=0; 
     }
    } 
    while (c!='*');
    break;
   }
   default : break;
  }
 }
}

/*********************************************************************
  Function:        void _ISR _U1TXInterrupt (void)
  Description :    Programme d'interruption dclench quand le buffer
                   de transmission de l'UART (4 octets) est vide
********************************************************************/
void _ISR _U1TXInterrupt (void)
{
 if (ptrTX1_RDdata!=ptrTX1_WRdata) // Y a-t-il un caractre  transmettre ?
 {
  do
  {
   U1TXREG=*ptrTX1_RDdata++;//Transm. du caractre et incrmentation du pointeur
   // Buffer circulaire
   if (ptrTX1_RDdata==Buffer_TX1+TX1_BufSize) ptrTX1_RDdata=Buffer_TX1;
  }
  while ((ptrTX1_RDdata!=ptrTX1_WRdata)&&(!U1STAbits.UTXBF));
 }
 IFS0bits.U1TXIF=0; // Acquitement de l'interruption
}

/*********************************************************************
  Function:        void _ISR _U1RXInterrupt (void)
  Description :    Programme d'interruption dclench  la rception
                   d'un caractre
********************************************************************/
void _ISR _U1RXInterrupt (void)
{
 char c;
 IFS0bits.U1RXIF=0;      // Clear interrupt flag
 while (U1STAbits.URXDA) // Boucle pour lire ts les car. ds le buf. de l'UART
 {
  c=U1RXREG;
  *ptrRX1_WRdata++=c; // Lecture du caractre suivant et incrmentation
                      // du pointeur circulaire de lecture
  if (ptrRX1_WRdata==Buffer_RX1+RX1_BufSize) ptrRX1_WRdata=Buffer_RX1;
 }
}

/******************************************************************************
*                                   UART2                                     *
*******************************************************************************/

/*********************************************************************
  Function:        void InitUART2(void)
  Description:     Initialisation de l'UART
                     - 38400 bauds, 8 bits, pas de parit, 1 bit Stop
                     - interruption sur TX si buffer UART vide
                     - interruption sur RX  chaque caractre
********************************************************************/
void InitUART2(void)
{
 U2MODE=0x8000; // UART valid, 8 bits, pas de parit, 1 bit stop
 U2STA =0x8400; // Interruption TX si buffer UART vide
                // UART TX valid
                // Interruption RX  chaque caractre
 U2BRG =(FCY/(16*38400))-1; // 38400 bauds (erreur +0,16%)
}

/*********************************************************************
  Function:        int ReadRXD2(char *c)
  Description:     Lecture d'un octet dans le buffer de rception
                     - renvoi de 0 si aucun caractre reu
                     - renvoi de 1 si caractre reu :
                        - caractre renvoy dans "c"
                        - pointeur circulaire ptrRX2_RDdata incrment
********************************************************************/
int ReadRXD2(char *c)
{
 if (ptrRX2_RDdata==ptrRX2_WRdata) return(0); // Pas de caractre reu
 else
 {
  *c=*ptrRX2_RDdata++;// Affecter "c" avec le caractre reu puis incrmenter
                     // le pointeur ptrRX2_RDdata
  // Pointeur circulaire
  if (ptrRX2_RDdata==Buffer_RX2+RX2_BufSize) ptrRX2_RDdata=Buffer_RX2;
  return(1);
 }
}

/*********************************************************************
  Function:        void WriteTXD2(char c)
  Description:     Ecriture d'un octet dans le buffer d'mission
                     - attendre libration du buffer
                     - envoi du caractre reu
                     - pointeur circulaire ptrTX2_WRdata incrment
********************************************************************/
void WriteTXD2(char c)
{
 int Nb_char;
 // Attendre libration d'une partie du buffer
 do
  {// Calcul du nb de caractres restant  transmettre
   Nb_char=ptrTX2_WRdata-ptrTX2_RDdata;
   if (Nb_char < 0) Nb_char=Nb_char+TX2_BufSize;
  } 
 while (Nb_char > (TX2_BufSize-8));
 // Caractre plac dans le buffer circulaire
 *ptrTX2_WRdata++=c; 
 if (ptrTX2_WRdata==Buffer_TX2+TX2_BufSize) ptrTX2_WRdata=Buffer_TX2;
 // Provoquer la 1 interruption si aucune transmission en cours
 // Le buffer de l'UART fait 4 octets (Nb_char=3)
 if ((U2STAbits.TRMT)&&(Nb_char>=3)) IFS1bits.U2TXIF=1; 
}

/*********************************************************************
  Function:        void WriteIntTXD2(int i)
  Description:     Ecriture d'un entier dans le buffer d'mission
********************************************************************/
void WriteIntTXD2(int i)
{
 WriteTXD2(i>>8); // Octet MSB d'abord
 WriteTXD2(i);    // Puis octet LSB
}

/*********************************************************************
  Function:        void WriteTXD2_String(char *String)
  Description:     Ecriture d'une chane de caractres dans le buffer TX
********************************************************************/
void WriteTXD2_String(char *String)
{
 while (*String) // Attendre caractre final (0)
 {
  WriteTXD2(*String++); // Transmission du caractre suivant
 }
}

/*********************************************************************
  Function:        void WriteLnTXD2_String(char *String)
  Description:     Ecriture d'une chane de caractres dans le buffer TX
                   "Retour chariot" (CR+LF) ajout  la fin
********************************************************************/
void WriteLnTXD2_String(char *String)
{
 WriteTXD2_String(String);   // Transmission de la chane
 WriteTXD2(CR); WriteTXD2(LF);// Transmission du "retour chariot"
}

/*********************************************************************
  Function:        void WriteLnTXD2(void
  Description:     "Retour chariot" (CR+LF) sur TX2
********************************************************************/
void WriteLnTXD2(void)
{
 WriteTXD2(CR); WriteTXD2(LF);// Transmission du "retour chariot"
}

/*********************************************************************
  Nom             :  Conv_Bin8_String
  Description     :  Conversion d'une valeur code 8 bits en binaire 
                     -> chane de caractres
  Arguments       :  x : nombre binaire  convertir
                     *String : pointeur vers la chane de destination
  Valeur renvoye : aucune
********************************************************************/
void Conv_Bin8_String(char *String, byte x)
{
 String[0]=x/100;                              // Centaines
 String[1]=(x-(100*String[0]))/10;             // Dizaines
 String[2]=x-(100*String[0])-10*String[1]+'0'; // Units en Ascii
 String[0]+='0'; // Conversion BCD -> Ascii des centaines
 String[1]+='0'; // Conversion BCD -> Ascii des dizaines
 String[3]=0;    // Terminateur de chane
}

/*********************************************************************
  Nom             :  Conv_Bin16_String
  Description     :  Conversion d'une valeur code 16 bits en binaire 
                     -> chane de caractres
  Arguments       :  x : nombre binaire  convertir
                     *String : pointeur vers la chane de destination
  Valeur renvoye : aucune
********************************************************************/
void Conv_Bin16_String(char *String, uint x)
{ 
 uint Reste;
 String[0]=x/10000;    Reste=x-(10000*String[0]); // Dizaines de milliers
 String[1]=Reste/1000; Reste=Reste-(1000*String[1]); // Milliers
 String[2]=Reste/100;  Reste=Reste-(100*String[2]);  // Centaines
 String[3]=Reste/10;   Reste=Reste-(10*String[3]);   // Dizaines
 String[4]=Reste+'0'; // Units en Ascii
 String[0]+='0'; // Conversion BCD -> Ascii des centaines
 String[1]+='0'; // Conversion BCD -> Ascii des dizaines
 String[2]+='0'; // Conversion BCD -> Ascii des dizaines
 String[3]+='0'; // Conversion BCD -> Ascii des dizaines
 String[5]=0;    // Terminateur de chane
}

/*********************************************************************
  Nom             :  Conv_Bin32_String
  Description     :  Conversion d'une valeur code 32 bits en binaire 
                     -> chane de caractres
  Arguments       :  x : nombre binaire  convertir
                     *String : pointeur vers la chane de destination
  Valeur renvoye : aucune
********************************************************************/
void Conv_Bin32_String(char *String, uint32 x)
{ 
 uint32 Reste;
 int i;
 String[0]=(uint32)x/1000000000;    
 Reste=x-(1000000000*String[0]); // 
 String[1]=Reste/100000000; Reste=Reste-(100000000*String[1]); // 
 String[2]=Reste/10000000;  Reste=Reste-(10000000*String[2]); // 
 String[3]=Reste/1000000;   Reste=Reste-(1000000*String[3]); // 
 String[4]=Reste/100000;    Reste=Reste-(100000*String[4]); // 
 String[5]=Reste/10000;     Reste=Reste-((uint32)10000*String[5]); // 
 String[6]=Reste/1000;      Reste=Reste-((uint32)1000*String[6]); // Milliers
 String[7]=Reste/100;       Reste=Reste-((uint32)100*String[7]);  // Centaines
 String[8]=Reste/10;        Reste=Reste-((uint32)10*String[8]);   // Dizaines
 String[9]=Reste+'0'; // Units en Ascii
 for (i=0; i<9; i++) String[i]+='0'; // Conversion BCD -> Ascii
 String[10]=0;    // Terminateur de chane
}

/*********************************************************************
  Nom             :  WTX2_Msg_Nb8bits
  Description     :  Transmission d'un message Ascii suivi d'un  
                     nombre cod sur 8 bits       
  Arguments       :  *String : pointeur vers la chane de destination
                     x : nombre binaire  convertir et transmettre 
  Valeur renvoye : aucune
********************************************************************/
void WTX2_Msg_Nb8bits(char *String,byte x)
{
 char s[4];
 WriteTXD2_String(String); // Transmission de la chane
 Conv_Bin8_String(s,x);    // Conversion BN -> Ascii
 WriteTXD2_String(s);      // Transmission de la chane
 WriteTXD2(CR); WriteTXD2(LF);// Transmission du "retour chariot"
}

/*********************************************************************
  Nom             :  WTX2_Msg_Nb16bits
  Description     :  Transmission d'un message Ascii suivi d'un  
                     nombre cod sur 16 bits       
  Arguments       :  *String : pointeur vers la chane de destination
                     x : nombre binaire  convertir et transmettre 
  Valeur renvoye : aucune
********************************************************************/
void WTX2_Msg_Nb16bits(char *String,uint x)
{
 char s[8];
 WriteTXD2_String(String); // Transmission de la chane
 Conv_Bin16_String(s,x);   // Conversion BN -> Ascii
 WriteTXD2_String(s);      // Transmission de la chane
 WriteTXD2(CR); WriteTXD2(LF);// Transmission du "retour chariot"
}

/*********************************************************************
  Nom             :  WTX2_Msg_Nb32bits
  Description     :  Transmission d'un message Ascii suivi d'un  
                     nombre cod sur 32 bits       
  Arguments       :  *String : pointeur vers la chane de destination
                     x : nombre binaire  convertir et transmettre
  Valeur renvoye : aucune
********************************************************************/
void WTX2_Msg_Nb32bits(char *String,uint32 x)
{
 char s[11];
 WriteTXD2_String(String); // Transmission de la chane
 Conv_Bin32_String(s,x);   // Conversion BN -> Ascii
 WriteTXD2_String(s);      // Transmission de la chane
 WriteTXD2(CR); WriteTXD2(LF);// Transmission du "retour chariot"
}

/*********************************************************************
  Function:        void Detect_Cmd_RXD2(void)
  Description:     Dtection et excution d'une ventuelle commande reue
                   Les commandes ont toujours une longueur de 5 octets :
                             Synchro, code, Dh, Dl, Synchro
                      - code : identifie la commande
                      - Dh, Dl : paramtre de la commande
********************************************************************/
void Detect_Cmd_RXD2(void)
{
 int i;
 char c;
 if (!ReadRXD2(&c)) return;
 WriteTXD1(c); // **** TEST **** Renvoi sur TX1 (GPS)
 for (i=1; i<5; i++) Buffer2_Cmd[i-1]=Buffer2_Cmd[i];
 Buffer2_Cmd[4]=c;
 if ((Buffer2_Cmd[0]==Synchro) && (Buffer2_Cmd[4]==Synchro))
 {
  switch(Buffer2_Cmd[1]) // Code de la commande
  {
   case 'R' : // Lecture d'un secteur SD Card
   {
    if (SectorRead(Buffer2_Cmd[2]*256+Buffer2_Cmd[3],&msd_buffer)==sdcValid)
       for(i=0; i<SDC_SECTOR_SIZE; i++) WriteTXD2(msd_buffer.octets[i]);
    break;
   }
   case 'W' : // Ecriture d'un secteur SD Card
   {
    break;
   }
   case 'A' : // Renvoi de C_SIZE
   {
	WriteTXD2('Z'); WriteTXD2('A');
	WriteIntTXD2(gblCSDReg.C_SIZE_H*4+gblCSDReg.C_SIZE_L);
	WriteTXD2('Z'); 
    break;
   }
   case 'B' : // Renvoi de C_SIZE_MULT et READ_BL_LEN
   {
	WriteTXD2('Z'); WriteTXD2('B');
	WriteTXD2(gblCSDReg.C_SIZE_MULT_H*2+gblCSDReg.C_SIZE_MULT_L);
	WriteTXD2(gblCSDReg.READ_BL_LEN);
	WriteTXD2('Z'); 
    break;
   }
   case 'C' : // Renvoi de TAAC et TRAN_SPEED
   {
	WriteTXD2('Z'); WriteTXD2('C');
	WriteTXD2(gblCSDReg.TAAC);
	WriteTXD2(gblCSDReg.TRAN_SPEED);
	WriteTXD2('Z'); 
    break;
   } 
   case 'D' : // Renvoi de R2W_FACTOR
   {
	WriteTXD2('Z'); WriteTXD2('D');
	WriteTXD2(gblCSDReg.R2W_FACTOR);
	WriteTXD2(0);
	WriteTXD2('Z'); 
    break;
   } 
   default : break;
  }
 }
}

/*********************************************************************
  Nom         :  Transmit_Datas_GPS
  Description :  Trannsmission des donnes GPS sur UART2
  Arguments   :  aucun
  Valeur renvoye : aucune
********************************************************************/
void Transmit_Datas_GPS(void)
{
 int i;
 if ((GGA.Fix_Quality=="")|(GGA.Fix_Quality=="0")) return;
 if ((GSA.Fix_type=="")|(GSA.Fix_type=="1")) return;
 if (GSV.Nb_Sat_view=="") return;
 i=(GSV.Nb_Sat_view[1]-'0');
 if (GSV.Nb_Sat_view[0]!=0) i=i+(GSV.Nb_Sat_view[1]-'0')*10;
 if (i<4) return;
 if (RMC.Status!="A") return;
 // Transmission UTC_Time
 WriteTXD2_String("#D1");
 for (i=0;i<6;i++) if (GGA.UTC_Time[i]!=0) WriteTXD2(GGA.UTC_Time[i]);
 //A SUIVRE ..............
}

/*********************************************************************
  Function:        void _ISR _U2TXInterrupt (void)
  Description :    Programme d'interruption dclench quand le buffer
                   de transmission de l'UART (4 octets) est vide
********************************************************************/
void _ISR _U2TXInterrupt (void)
{
 if (ptrTX2_RDdata!=ptrTX2_WRdata) // Y a-t-il un caractre  transmettre ?
 {
  do
  {
   U2TXREG=*ptrTX2_RDdata++;//Transm. du caractre et incrmentation du pointeur
   // Buffer circulaire
   if (ptrTX2_RDdata==Buffer_TX2+TX2_BufSize) ptrTX2_RDdata=Buffer_TX2;
  }
  while ((ptrTX2_RDdata!=ptrTX2_WRdata)&&(!U2STAbits.UTXBF));
 }
 IFS1bits.U2TXIF=0; // Acquitement de l'interruption
}

/*********************************************************************
  Function:        void _ISR _U2RXInterrupt (void)
  Description :    Programme d'interruption dclench  la rception
                   d'un caractre
********************************************************************/
void _ISR _U2RXInterrupt (void)
{
 char c;
 IFS1bits.U2RXIF=0;      // Clear interrupt flag
 while (U2STAbits.URXDA) // Boucle pour lire ts les car. ds le buf. de l'UART
 {
  c=U2RXREG;
  *ptrRX2_WRdata++=c; // Lecture du caractre suivant et incrmentation
                     // du pointeur circulaire de lecture
  if (ptrRX2_WRdata==Buffer_RX2+RX2_BufSize) ptrRX2_WRdata=Buffer_RX2;
 }
}

/******************************************************************************
*                                   TIMER1                                    *
*******************************************************************************/

/*********************************************************************
  Function:        void InitTMR1(void)
  Description:     Initialisation du Timer 1 (Timer A) pour provoquer
                   une interruption  rythme rgulier.
  Note : le rythme est choisi pour obtenir une frquence de transmission
         de l'ensemble des paramtres sur TXD de 10Hz
********************************************************************/
void InitTMR1(void)
{
 T1CON=0x0010;	 // Fonctionnement continu, horloge=Fcy/8=2MHz
 TMR1=0;         // Raz du compteur pour dcaler la 1 interruption
 PR1=(FCY/8)/(10*NbDtaTxd);	// Frquence = 10Hz x NbDtaTxd
 T1CONbits.TON=1;// Validation Timer 1 
 IFS0bits.T1IF=0;// Raz indicateur interruption Timer 1 
 IEC0bits.T1IE=1;// Validation interruption Timer 1 (RTI)
}

/*********************************************************************
  Function   : void _ISR _T1Interrupt (void)
  Descrition : Frquence = 10Hz x NbDtaTxd
  Traitement : 
    - Anti-rebonds sur SDC_SD (contact de dtection carte SD) pour 
      affecter l'indicateur SDC_CD_OK
********************************************************************/
void _ISR _T1Interrupt(void)
{
 //TEST0=1;
 IFS0bits.T1IF=0; // Acquitement interruption
 Compt_T1_TXD++;
 if (Compt_T1_TXD==NbDtaTxd) 
  {// Rythme d'excution = 10Hz
   TEST1=!TEST1;   // Pour contrle avec un oscilloscope (TEST1=port RE4)
   Compt_T1_TXD=0; // Compt_T1_TXD modulo NbDtaTxd
   SDC_CD_AR();    // Affectation des indicateurs SDC_CD_OK et SDC_CD_new
  }
 //TEST0=0; 
}

/******************************************************************************
*                                   TIMER3                                    *
*******************************************************************************/

/*********************************************************************
  Function:        void InitTMR3(void)
  Description:     Initialisation du Timer 3 (Timer C)                       
                   Horloge = Fcy/256 = 62,5kHz
                   Utilis pour : - mesurer la vitesse du moteur        
                                  - produire des dlais logiciels
********************************************************************/
void InitTMR3(void)
{
 T3CON=0x0030; // Horloge interne Fcy/256 = 62,5kHz
 TMR3=0;       // Raz du compteur pour retarder la 1 demande d'interruption
 PR3=0xFFFF;   // Comptage de 0x0000  0xFFFF au reset
 T3CONbits.TON=1; // Timer 3 "On"
 IEC0bits.T3IE=1; // Validation interruption Timer 3
}

/*********************************************************************
  Function   : void _ISR _T3Interrupt (void)
  Descrition : Activ au dpassement de TMR3
               Traitement : 
                 - Incrmentation TMR3_MSB
********************************************************************/
void _ISR _T3Interrupt (void)
{
 IFS0bits.T3IF=0; // Acquitement interruption
 // Incrmentation du compteur 32 bits
 TMR3_MSB++;     
 TMR3=0;
}

/******************************************************************************
*                            Gestion SD Card                                  *
*******************************************************************************/

/*******************************************************************
  Function:    void WriteSPI(unsigned char data)
  Description: Transmission d'un octet sur le coupleur SPI
    - Affectation du registre de transmission
    - Attente de fin de transmission
  Paramtre : data = octet  transmettre
********************************************************************/
void WriteSPI(unsigned char data)
{
 SPI1BUF=data; // Affectation du buffer SPI : dbut d'une transmission/rception
 // Attendre la rception d'un nouveau caractre, donc la fin de transmission
 while (!SPI1STATbits.SPIRBF) {};
 data=SPI1BUF;  // Lecture octet reu pour Raz SPIRBF
 SPI1STATbits.SPIROV=0;
}

/*******************************************************************
  Function:    unsigned char ReadSPI(void)
  Description: Rception d'un octet sur le coupleur SPI
    - Affectation du registre de transmission avec l'octet neutre (0xFF)
      pour dclencher une nouvelle rception
    - Attente la fin de rception pour lire l'octet reu
  Paramtre : data = octet  transmettre
********************************************************************/
unsigned char ReadSPI(void)
{
 unsigned char data;
 //TEST1=1;
 // Dclencher un cycle de lecture (transmission 0xFF)
 WriteSPI(0xFF);// Transmission du caractre neutre pour recevoir 
                // un nouveau caractre
 data=SPI1BUF;  // Lecture octet reu
 SPI1STATbits.SPIROV=0;
 return data;
}

/*******************************************************************
  Function:    void InitSPI(int Div)
  Description: Initialisation coupleur SPI pour cartes SD
    - Master 8 bits sans gestion de trame
    - CKE=1, CKP=0 et SMP=0
    - SS non utili
  Paramtre : Div = 0 : prdivision Fcy par 64
                  = 1 : prdivision Fcy par 16
                  = 2 : prdivision Fcy par 4
                  = 3 : prdivision Fcy par 1
********************************************************************/
void InitSPI(int Div)
{
 SPI1CON=0x013C|Div; // FRMEN=SPIFSD=0 : pas de gestion des trames
                     // DISSD0=0       : SDO1 utilis
                     // MOD16=0        : mode 8 bits
                     // SMP=0          : SDI1 chantillonn au milieu
                     // CKE=1          : SDO1 change qd SCK1 passe de l'tat 
                     //                  actif  l'tat de repos ("Idle")
                     // SSEN=0         : SS non utilis
                     // CKP=0          : tat de repos SCK1=0
                     // MSTEN=1        : mode Master
                     // SPRE<2:0>=0    : prdivision secondaire=1
                     // PPRE<1:0>=Div  : voir ci-dessus
 SPI1STAT=0x8000;    // SPIEN=1        : coupleur SDI actif
                     // SPISIDL=0      : module oprationnel en mode Idle
                     // SPIROV,SPITBF,SPIRBF : indicateurs
}

/*******************************************************************
  Function   : void CloseSPI(void)
  Description: Arrt du coupleur SPI
********************************************************************/
void CloseSPI()
{  
  _SPI1IE = 0;            /* Disable the Interrupt bit in the 
                             Interrupt Enable Control Register */
  SPI1STATbits.SPIEN = 0; /* Disable the module. All pins controlled
                             by PORT Functions */
  _SPI1IF = 0;            /* Disable the Interrupt flag bit in the
                             Interrupt Flag Control Register */    
}

/*******************************************************************
  Function   : SDC_RESPONSE SendSDCCmd(BYTE cmd, DWORD address)
  Description: 
    - Prpation de la trame de la commande
    - Rponse de la carte (R1 ou R2) dans la valeur retourne
  Paramtres : 
    - cmd : code de la commande (type "sdmmc_cmd")
    - adress : argument de la commande
  Retour     : rponse de la carte SD (un octet)
    - 0x00 ou 0x01 : commande bien reue, sinon erreur : 
      - Bit 0 IN_IDLE_STATE:1; // La carte est en  mode "Idle" 
                               // et excute la phase d'initialisation
      - Bit 1 ERASE_RESET  :1; // Acquitement de l'excution d'une commande 
                               // d'arrt d'effacement
      - Bit 2 ILLEGAL_CMD  :1; // Commande illgale
      - Bit 3 CRC_ERR      :1; // Le CRC de la dernire commande est incorrect
      - Bit 4 ERASE_SEQ_ERR:1; // Erreur ds la squence d'effacement commande
      - Bit 5 ADDRESS_ERR  :1; // Erreur d'adresse 
      - Bit 6 PARAM_ERR    :1; // Erreur dans les arguments de la dernire commande
      - Bit 7 B7           :1; // Toujours  0
  Attention : le signal SDC_CS n'est pas mis  "1"  la fin de cette fonction.
              s'il reste des donnes  transmettre (criture ou lecture)
********************************************************************/
SDC_RESPONSE SendSDCCmd(byte cmd, dword address)
{        
 word timeout=8; // Dlai max entre cmde et rponse = NCR = 8 octets
 byte index;
 SDC_RESPONSE response; // Pour mmoriser localement la rponse
 CMD_PACKET CmdPacket;  // Structure de la commande
	    
 SDC_CS=0; // Card Select  
 // Prpation de la commande complte
 CmdPacket.cmd    =sdmmc_cmdtable[cmd].CmdCode; // Code commande
 CmdPacket.address=address;                     // Argument (4 octets) 
 CmdPacket.crc    =sdmmc_cmdtable[cmd].CRC;     // CRC    
 // Transmission de la trame   
 WriteSPI(CmdPacket.cmd);   // Send Command
 WriteSPI(CmdPacket.addr3); // Argument MSB
 WriteSPI(CmdPacket.addr2);
 WriteSPI(CmdPacket.addr1);
 WriteSPI(CmdPacket.addr0); // Argument LSB
 WriteSPI(CmdPacket.crc);   // Send CRC
 // La commande est-elle suivie d'une rponse ?
 if(sdmmc_cmdtable[cmd].responsetype==R1||sdmmc_cmdtable[cmd].responsetype==R1b)
 {// de type R1 (un seul octet) ou R1b (1 octet):
  do
  {// Lecture d'octets jusqu' octet diffrent de 0xFF ou "Timeout"
   response.r1._byte=ReadSPI();
   timeout--; // Le dlai max de la rponse est NCR=4x8=32 priodes CLK
  } while ((response.r1._byte==0xFF)&&(timeout!=0));
 }
 else if(sdmmc_cmdtable[cmd].responsetype==R2)
 {// de type R2 (commande SEND_STATUS):
  ReadSPI(); // Dlai = 1 octet
  response.r2._byte1=ReadSPI(); // Rception des 2 octets "R2"
  response.r2._byte0=ReadSPI();
 }    
 if(sdmmc_cmdtable[cmd].responsetype==R1b)
 {// de type R1b : attendre la fin de l'opration en cours (signal "busy")
  response.r1._byte=0x00;
  for(index=0; (index<0xFF)&&(response.r1._byte==0x00); index++)
  {
   timeout = 0xFFFF; // Timeout = lecture de 65536 octets !	           
   do
   {// Lectures d'octets de la carte jusqu' ce qu'elle soit "ready"
    // (octet lu = 0) ou "timeout"
    response.r1._byte=ReadSPI(); // Lecture d'un octet
    timeout--;
   } while ((response.r1._byte==0x00)&&(timeout!=0)); 
  }
 }       
 WriteSPI(0xFF); // Impos par la norme pour terminer l'opration (5.1.8)
 // Reste-t-il des donnes  transmettre ?
 //   - non : SDC_CS revient  "1"
 //   - oui : SDC_CS est maintenu  "0"
 if(!(sdmmc_cmdtable[cmd].moredataexpected)) 
  {
   Delay(Delai_48uS); // 48S
   SDC_CS=1; // D-slection SD Card
   Delay(Delai_mini); // 16S
  } 
 return(response);        
}

/*******************************************************************
  Function   : SDC_Error CSDRead(BYTE *buffer)
  Description: Lecture du registre CSD (Card Specific Data) :
    - transmission de la commande CMD9 SEND_CSD
    - rception des 128 bits du registre (16 octets). Infos extraites :
        - capacit de la carte
        - longueur (gblBlkLen) et nombre de blocs (gblNumBlks)
  Paramtres : buffer : destination des donnes lues 
  Retour     : nature de l'ventuelle erreur (voir SDC_Error)
********************************************************************/
SDC_Error CSDRead()
{
 int index;
 word timeout=0x2ff;
 SDC_RESPONSE response;
 byte data_token;
 SDC_Error status = sdcValid;
 byte cmd=SEND_CSD;
 dword address=0x00;
 CMD_PACKET CmdPacket;
	    
 SDC_CS=0; //Card Select (CS="0")   
 // Prparation de la commande SEND_CSD (CMD9)
 CmdPacket.cmd    = sdmmc_cmdtable[cmd].CmdCode; // Code
 CmdPacket.address= address;                     // Argument
 CmdPacket.crc    = sdmmc_cmdtable[cmd].CRC;     // CRC
 WriteSPI(CmdPacket.cmd);   // Transmission Code commande
 WriteSPI(CmdPacket.addr3); // Argument MSB
 WriteSPI(CmdPacket.addr2);
 WriteSPI(CmdPacket.addr1);
 WriteSPI(CmdPacket.addr0); // Argument LSB
 WriteSPI(CmdPacket.crc);   // Transmission CRC en dernier
 // Attente et rception de la rponse de type R1
 //timeout=8; // *** A tester *** Dlai max entre cmde et rponse = NCR = 8 octets
 do
 {
  response.r1._byte=ReadSPI();
  timeout--;
 } while ((response.r1._byte==0xFF) && (timeout!=0));
 // S'assurer que la carte est prte
 if (response.r1._byte != 0x00) status=sdcCardBadCmd;
 else
 {
  index = 0x2FF;                                      
  // Attendre le jeton de dpart et le bloc de donnes   
  do
  {// Attendre le premier octet diffrent de 0xFF
   data_token=ReadSPI();
   index--;
  } while ((data_token==SDC_FLOATING_BUS) && (index!=0));    
  // Le premier octet doit tre le jeton DATA_START_TOKEN
  if((index == 0) || (data_token != DATA_START_TOKEN))
     status = sdcCardTimeout;
  else
  {// Oui : rception des 16 octets suivants (128 bits du registre CSD
   for(index=CSD_SIZE-1; index>=0; index--) gblCSDReg._byte[index]=ReadSPI();  
  }
 // Le transfert se termine avec les 16 bits du CRC ( **** TEST ****)
 WriteSPI(0xFF);WriteSPI(0xFF); // Lecture simple, sans prise en compte
 WriteSPI(0xFF); // Impos par la norme pour terminer l'opration (5.1.8)
 }
 Delay(5*Delai_mini); // 5*16S
 SDC_CS=1;          // D-selection carte SD (CS="1")
 Delay(Delai_mini); // 16S
 return(status);
}

/*******************************************************************
  Function   : byte IsWriteProtected()
  Description: lecture du switch de protection en criture
  Paramtres : Aucun
  Retour     : TRUE si protg en criture
               FLASE dans le cas contraire
********************************************************************/
byte IsWriteProtected(void)
{
 if (SDC_WP) return TRUE;
 else return FALSE;
}

/*******************************************************************
  Function   : SDC_Error SectorRead(uint32 sector_addr, Secteur* buffer)
  Description: la fonction lit 512 octets  partir de l'adresse "sector_addr"
               et range les octets dans "buffer"
  Paramtres :
   - sector_addr : adresse du secteur; chaque secteur comporte un bloc de 512 
                   octets
   - buffer : destination des donnes lues  (buffer de 512 octets)
  Retour     : voir SDC_Error
********************************************************************/
SDC_Error SectorRead(uint32 sector_addr, Secteur* buffer)
{
 uint index;
 SDC_RESPONSE response; // Variable pour la rponse de la commande
 byte data_token;       // Variable pour le jeton
 SDC_Error status = sdcValid; // Pas d'erreur  priori
 #ifdef STATUSLED
  STRLED=1; // Allumer la Led indicatrice de lecture
 #endif
 // Envoi de la commande de lecture d'un bloc de 512 octets
 // Le dcalage de 9 bits correspond  la taille des secteurs (2^9=512)
 response=SendSDCCmd(READ_SINGLE_BLOCK,(sector_addr<<9));
 // Vrifier que la commande a bien t accepte
 if(response.r1._byte!=0x00) status=sdcCardBadCmd;
 else
 {// Octet de rponse = 0x00 : commande accepte -> lecture du bloc
  index = 0x2FF; // Variable "index" utilise ici comme compteur de "Timeout"
  // Attente du jeton "Start Block" et du bloc de donnes
  do
  {// Le premier octet reu diffrent de 0xFF est le jeton
   data_token=ReadSPI();
   index--; // Dcompteur Timeout
  } while ((data_token==SDC_FLOATING_BUS) && (index!=0));
  // Ce jeton doit tre DATA_START_TOKEN
  if((index==0)||(data_token!=DATA_START_TOKEN)) status=sdcCardTimeout;
  else
  {// Le jeton est correct : les 512 octets de donnes suivent
   for(index=0; index<SDC_SECTOR_SIZE; index++) buffer->octets[index]=ReadSPI();
   // Le transfert se termine avec les 16 bits du CRC
   WriteSPI(0xFF);WriteSPI(0xFF); // Lecture simple, sans prise en compte
   //status = mmcCardCRCError;
  }
  WriteSPI(0xFF); // Le 5.1.8 du document "ProdManualSDCardv1.9.pdf" impose 
                  // la production de 8 impulsions d'horloge  la fin des changes 
                  // avant un arrt prolong de l'horloge.
 }
 // Dlai supplmentaire pour s'assurer de l'arrt de l'horloge
 Delay(3*Delai_mini); // 3*16S
 SDC_CS=1; // D-selection de la carte SD
 Delay(Delai_mini); // 16S
 #ifdef STATUSLED
  STRLED = 0; // Eteindre la Led indicatrice de lecture
 #endif 
 return(status);
}

/*******************************************************************
  Function   : SDC_Error SectorWrite(DWORD sector_addr, Secteur* buffer)
  Description: la fonction crit les 512 octets de "buffer"  partir de 
               l'adresse "sector_addr"
  Paramtres :
   - sector_addr : adresse du secteur; chaque secteur comporte un bloc de 512 octets
   - buffer : pointeur vers la source des donnes  crire (tableau de 512 octets)
  Retour     : voir SDC_Error
********************************************************************/
SDC_Error SectorWrite(uint32 sector_addr, Secteur* buffer)
{
 word         index;
 byte         data_response;       
 SDC_RESPONSE response;      // Variable pour la rponse de la commande
 SDC_Error    status=sdcValid; 	
 #ifdef STATUSLED
  STWLED=1; // Allumer la Led indicatrice d'criture
 #endif	
 // Envoi de la commande d'criture d'un bloc de 512 octets
 // Le dcalage de 9 bits correspond  la taille des blocs (2^9=512)
 response=SendSDCCmd(WRITE_SINGLE_BLOCK,(sector_addr << 9));
 // Vrifier que la commande a bien t accepte
 if(response.r1._byte!=0x00) status=sdcCardBadCmd;    
 else
 {// Octet de rponse = 0x00 : commande accepte -> criture du bloc
  WriteSPI(DATA_START_TOKEN); // On commence par transmettre le jeton START
  // Puis les 512 octets
  for(index=0;index<SDC_SECTOR_SIZE;index++) WriteSPI(buffer->octets[index]);
  // Puis transmission des 16 bits CRC:
  WriteSPI(0xFF); WriteSPI(0xFF); // Non calcul
  // Lecture immdiate de l'octet de rponse
  data_response=ReadSPI();
  if ((data_response & 0x0F)!=DATA_ACCEPTED) status=sdcCardDataRejected;
  else
  {// Rponse OK : attente fin de l'criture effective en mmoire flash
   index=0; // Timeout
   do
   {// Lectures successives jusqu' rception d'un octet de rponse non nul ou Timeout
    data_response=ReadSPI();
    index++;
   } while ((data_response==0x00)&&(index!=0));
   // Test Timeout (65536x8 priodes d'horloge !)         
   if (index==0) status=sdcCardTimeout;
  }
  WriteSPI(0xFF); // Impos par la norme pour terminer l'opration (5.1.8)
 }
 Delay(3*Delai_mini); // 3*16S pour attendre la transmission du dernier octet
 SDC_CS=1;            // D-selection de la carte SD
 Delay(Delai_mini);   // 16S
 #ifdef STATUSLED
  STWLED=0; // Eteindre la Led indicatrice d'criture
 #endif
 return(status);
}

/*******************************************************************
  Function   : SDC_Error MediaInitialize()
  Description: Initialisation de la carte SD et des variables associes
  Paramtre  : indicateurs sur la carte :
    isSDMMC : type MMC
    isWP    : protge en criture
  Retour     : nature de l'erreur
    sdcValid                Indicateur de carte valide
    sdcCardInitCommFailure  Aucune communication n'a pu s'tablir avec la carte
    sdcCardNotInitFailure   La carte n'est pas entre dans la phase "initialisation"
    sdcCardInitTimeout      "Timeout" pdt l'initialisation
    sdcCardTypeInvalid      Le type de carte n'a pas pu tre tabli
    sdcCardBadCmd           La carte n'a pas reconnu la commande
    sdcCardTimeout          "Timeout" pdt une lecture, criture ou effacement   
    sdcCardCRCError         Erreur CRC pdt une lecture 
    sdcCardDataRejected     Le "CRC" des donnes ne correspond pas
    sdcEraseTimedOut        "Timeout" pdt effacement
  Note : SDC_CS reste  "0" au retour en l'absence d'erreur
********************************************************************/
SDC_Error MediaInitialize(void)
{
 uint16       timeout;
 SDC_Error    status=sdcValid, CSDstatus=sdcValid;
 SDC_RESPONSE response; 
 // Raz indicateurs  priori
 Flags.isSDMMC = Flags.isWP = 0;
 SDC_CS=1; // D-selection carte SD
 // Mise sous tension de la carte SD
 SDC_POWER=0;
 // Attendre la stabilisation de la tension d'alimentation
 Delay(Delai_40ms);
 SDC_CS_DIR=OUTPUT; // Port CS_SD en sortie
 // A la mise sous tension, la carte SD est en mode "drain ouvert" et n'accepte
 // pas des horloges suprieures  400kHz
 InitSPI(0); // Coupleur SPI en mode "master" 8 bits; horloge = Fcy/64 = 250kHz
 // La carte SD ncessite 80 priodes d'horloge pour tre oprationnelle
 for(timeout=0; timeout<10; timeout++) WriteSPI(0xFF);// 10 octets x 8 bits
 Delay(Delai_500us); // Attendre arrt horloge (transmission du dernier octet)
 SDC_CS=0;   // Card Select
 Delay(Delai_500us);
 // Commande CMD0=GO_IDLE_STATE : reset "software" de la carte
 response=SendSDCCmd(GO_IDLE_STATE,0x0);// Transm. et rception de la rponse "R1"
 if(response.r1._byte==SDC_BAD_RESPONSE)
 {
  status=sdcCardInitCommFailure; // La carte est non connecte ou a une dfaillance
  goto InitError; // Fin prmature
 }
 // La carte est-elle prte (ready) ?
 if (response.r1._byte != 0x01) //0x01 = No Err & Busy Initializing
 {
  status=sdcCardNotInitFailure; // La carte n'est pas prte
  goto InitError; // Fin prmature
 }
 // En mode "Idle", la cmde CMD1 doit tre rpte jusqu' ce que la carte soit prte
 timeout=0xFFF; // Timeout = 4096x8 priodes d'horloge !
 do
 {
  response=SendSDCCmd(SEND_OP_COND,0x0); // Commande CMD1 (initialisation)
  timeout--; // Le processus d'initialisation de la carte prend de 50mS  500mS
 } while (response.r1._byte!=0x00 && timeout!=0);       
 // Ragir si la carte n'est pas initialise
 if(timeout==0)
 {
  status=sdcCardInitTimeout; // Initialisation dfaillante aprs Timeout !
  goto InitError; // Fin prmature
 }                        
 else  
 {// La carte SD est prte : 
  CSDstatus=CSDRead(); // Lecture du registre CSD de la carte
  if(!CSDstatus)       // Aucune anomalie ?
   {//Oui : on peut monter l'horloge
    InitSPI(1);        // Mode "master" 8 bits pour cartes SD; horloge = Fcy/16 = 1MHz
   }
  else 
   {//Non : la carte est dfaillante
    status=sdcCardTypeInvalid;
    goto InitError; // Fin prmature
   } 
 }
 // Arrt CRC7 si possible; commande invalide pour certaines cartes (CMD59)
 response=SendSDCCmd(CRC_ON_OFF,0x0);
 // Affectation de la longueur des blocs  512
 SendSDCCmd(SET_BLOCKLEN,BLOCKLEN_512);
 // Affectation de l'indicateur de protection d'criture
 if (IsWriteProtected()) Flags.isWP=TRUE;
 // Lecture rpte d'un secteur jusqu' l'absence d'erreur. Certaines cartes "San"
 // n'acceptent pas cette fonction d'initialisation pour diffrentes raisons
 for (timeout=0xFF;
	  timeout>0&&SectorRead(0x0,&msd_buffer)!=sdcValid;
	  timeout--) {}
 // Lecture du secteur dfaillante ?
 if(timeout==0)  
 {// Oui :      
  status = sdcCardNotInitFailure; // La carte est dfaillante
  goto InitError; // Fin prmature
 }
 return(status); // Retour en l'absence d'erreur
    
InitError:
 SDC_CS=1; // D-slection de la carte 
 return(status);                            
}

/*******************************************************************
  Function   : void SocketInitialize(void)
  Description: Initialisation des ports utiliss par la SD Card
  Paramtres : Aucun
  Retour     : Aucun
********************************************************************/
void SocketInitialize(void)
{    
 // Couper l'alimentation de la carte mmoire
 SDC_POWER=1;         // Connecteur SD Card non aliment au reset
 SDC_POWER_DIR=OUTPUT;// Port SDC_POWER en sortie
 SDC_CD_DIR=INPUT;    // Port "Card Detect" en entre
 SDC_CS=1;            // CS de la SD Card  "1"
 SDC_CS_DIR=INPUT;    // Port SDC_CS en entre si SD Card non alimente
 SDC_WP_DIR=INPUT;    // Port "Write Protect" en entre
 //InitSPI(0); // Mode "master" 8 bits pour cartes SD; horloge=Fcy/64=250kHz
 #ifdef STATUSLED
  STRTRIS=OUTPUT;
  STWTRIS=OUTPUT;
  STRLED=0; // Led indicatrice "lectures" teinte
  STWLED=0; // Led indicatrice "critures" teinte
 #endif
}// end SocketInitialize

/*******************************************************************
  Function   : void ShutdownMedia(void)
  Description: Coupure alimentation de la carte SD
  Paramtres : Aucun
  Retour     : Aucun
********************************************************************/
void ShutdownMedia(void)
{
 SDC_CS=1;         // D-slection de la carte                              
 SDC_CS_DIR=INPUT; // Port SDC_CS en entre si SD Card non alimente
 CloseSPI();       // Ports SPI en HiZ
 SDC_POWER=1;      // Coupure alimentation        
}

/*********************************************************************
  Function    : void SDC_CD_AR(void)
  Description : Lecture de l'tat actuel du contact CD de la SD Card
                avec fonction "anti-rebonds".
  Traitement : - lance par le programme d'interruption "_T1Interrupt" (10Hz)
               - lecture tat actuel
               - si tat stable pdt 1 seconde -> nouvel tat stable dtect
                 et plac ds SDC_CD_OK
               - SDC_CD_New positionn si chgt d'tat stable
********************************************************************/
void SDC_CD_AR(void)
{
 unsigned int SDC_CD_actuel;
 SDC_CD_actuel=!SDC_CD; // 1 : prsence, 0 : absence de carte 
 // Dtection d'un changement d'tat
 if (SDC_CD_actuel!=Flags.SDC_CD_old) 
  // Etat diffrent entre 2 lectures
  Cpt_AR_SDC_CD=10; // R-initialisation du compteur anti-rebonds (10 secondes)
 else
 {// Mme tat entre 2 lectures
  if (Cpt_AR_SDC_CD!=0)
  {// Dtection rebonds en cours
   Cpt_AR_SDC_CD--; // Dcomptage "anti-rebonds"
   if (Cpt_AR_SDC_CD==0)
   {// Mme tat pdt 10 lectures = 1s
    if (Flags.SDC_CD_OK!=SDC_CD_actuel) // Chgt d'tat stable ?
    {
     Flags.SDC_CD_OK=SDC_CD_actuel; // Nouvel tat stable
     Flags.SDC_CD_New=1;            // Chgt d'tat stable
    }
   }
  }
 }
 Flags.SDC_CD_old=SDC_CD_actuel;
}

/******************************************************************************
*                              GESTION DE LA FAT                              *
*******************************************************************************/

/*********************************************************************
  Function:    void Reset_FAT(void)
  Description: Initialisation de la structure BPB pour les tests     
               (infos prises sur SD Card 1G en FAT16)
********************************************************************/
void Reset_FAT(void)
{
 BPB_StartSector=133; 
 msd_buffer.BPB_FAT.BS_jmpBoot[0]=0xEB; 
 msd_buffer.BPB_FAT.BS_jmpBoot[1]=0; 
 msd_buffer.BPB_FAT.BS_jmpBoot[2]=0x90;
 memcpy(msd_buffer.BPB_FAT.BS_OEMName,"CREMMEL.",8);
 msd_buffer.BPB_FAT.BPB_BytsPerSec[0]=0; msd_buffer.BPB_FAT.BPB_BytsPerSec[1]=0x02;
 msd_buffer.BPB_FAT.BPB_SecPerClus=32;
 msd_buffer.BPB_FAT.BPB_RsvdSecCnt=6; 
 msd_buffer.BPB_FAT.BPB_NumFATs=2;
 msd_buffer.BPB_FAT.BPB_RootEntCnt[0]=0; msd_buffer.BPB_FAT.BPB_RootEntCnt[1]=2;
 msd_buffer.BPB_FAT.BPB_TotSec16[0]=0; msd_buffer.BPB_FAT.BPB_TotSec16[1]=0;
 msd_buffer.BPB_FAT.BPB_Media=0xF8;
 msd_buffer.BPB_FAT.BPB_FATSz16=237;
 msd_buffer.BPB_FAT.BPB_SecPerTrk=63;
 msd_buffer.BPB_FAT.BPB_NumHeads=32;
 msd_buffer.BPB_FAT.BPB_HiddSec=133;
 msd_buffer.BPB_FAT.BPB_TotSec32=1935227;
 msd_buffer.BPB_FAT.BS_DrvNum=0x80;
 msd_buffer.BPB_FAT.BS_Reserved1=0;
 msd_buffer.BPB_FAT.BS_BootSig=0x29;
 msd_buffer.BPB_FAT.BS_VolID[0]=0; msd_buffer.BPB_FAT.BS_VolID[1]=1;
 msd_buffer.BPB_FAT.BS_VolID[2]=2; msd_buffer.BPB_FAT.BS_VolID[3]=3;
 memcpy(msd_buffer.BPB_FAT.BS_VolLab,"SD_CARD    ",11);
 memcpy(msd_buffer.BPB_FAT.BS_FilSysType,"FAT16   ",8);
 msd_buffer.BPB_FAT.SignSector0=0xAA55;
}

/*********************************************************************
  Function:    void SampleFileData(void)
  Description: Affectation de qq strucures de fichiers ds la FAT   
               Utilis pour des tests                
********************************************************************/
void SampleFileData(void)
{
 int i;
 strcpy(msd_buffer.FATFilesInfos[0].DIR_Name,"SD CARD    ");
 msd_buffer.FATFilesInfos[0].DIR_Attr=0x08;
 msd_buffer.FATFilesInfos[0].DIR_NTRes=0;
 msd_buffer.FATFilesInfos[0].DIR_CrtTimeTenth=0;
 msd_buffer.FATFilesInfos[0].DIR_CrtTime[0]=0;
 msd_buffer.FATFilesInfos[0].DIR_CrtTime[1]=0;
 msd_buffer.FATFilesInfos[0].DIR_CrtDate[0]=0;
 msd_buffer.FATFilesInfos[0].DIR_CrtDate[1]=0;
 msd_buffer.FATFilesInfos[0].DIR_LstAccDate[0]=0;
 msd_buffer.FATFilesInfos[0].DIR_LstAccDate[1]=0;
 msd_buffer.FATFilesInfos[0].DIR_FstClusHI=0;
 msd_buffer.FATFilesInfos[0].DIR_WrtTime[0]=0;
 msd_buffer.FATFilesInfos[0].DIR_WrtTime[1]=0;
 msd_buffer.FATFilesInfos[0].DIR_WrtDate[0]=0;
 msd_buffer.FATFilesInfos[0].DIR_WrtDate[1]=0;
 msd_buffer.FATFilesInfos[0].DIR_FstClusLO=0;
 msd_buffer.FATFilesInfos[0].DIR_FileSize=0;

 strcpy(msd_buffer.FATFilesInfos[1].DIR_Name,"TEST1   TXT");
 msd_buffer.FATFilesInfos[1].DIR_Attr=0x20;
 msd_buffer.FATFilesInfos[1].DIR_NTRes=0;
 msd_buffer.FATFilesInfos[1].DIR_CrtTimeTenth=0;
 msd_buffer.FATFilesInfos[1].DIR_CrtTime[0]=0;
 msd_buffer.FATFilesInfos[1].DIR_CrtTime[1]=0;
 msd_buffer.FATFilesInfos[1].DIR_CrtDate[0]=0;
 msd_buffer.FATFilesInfos[1].DIR_CrtDate[1]=0;
 msd_buffer.FATFilesInfos[1].DIR_LstAccDate[0]=0;
 msd_buffer.FATFilesInfos[1].DIR_LstAccDate[1]=0;
 msd_buffer.FATFilesInfos[1].DIR_FstClusHI=0;
 msd_buffer.FATFilesInfos[1].DIR_WrtTime[0]=0;
 msd_buffer.FATFilesInfos[1].DIR_WrtTime[1]=0;
 msd_buffer.FATFilesInfos[1].DIR_WrtDate[0]=0;
 msd_buffer.FATFilesInfos[1].DIR_WrtDate[1]=0;
 msd_buffer.FATFilesInfos[1].DIR_FstClusLO=2;
 msd_buffer.FATFilesInfos[1].DIR_FileSize=0x00000046;

 msd_buffer.FATFilesInfos[2].DIR_Name[0]=FAT_FILE_DELETED;
 msd_buffer.FATFilesInfos[2].DIR_Attr=0x20;
 msd_buffer.FATFilesInfos[2].DIR_NTRes=0;
 msd_buffer.FATFilesInfos[2].DIR_CrtTimeTenth=0;
 msd_buffer.FATFilesInfos[2].DIR_CrtTime[0]=0;
 msd_buffer.FATFilesInfos[2].DIR_CrtTime[1]=0;
 msd_buffer.FATFilesInfos[2].DIR_CrtDate[0]=0;
 msd_buffer.FATFilesInfos[2].DIR_CrtDate[1]=0;
 msd_buffer.FATFilesInfos[2].DIR_LstAccDate[0]=0;
 msd_buffer.FATFilesInfos[2].DIR_LstAccDate[1]=0;
 msd_buffer.FATFilesInfos[2].DIR_FstClusHI=0;
 msd_buffer.FATFilesInfos[2].DIR_WrtTime[0]=0;
 msd_buffer.FATFilesInfos[2].DIR_WrtTime[1]=0;
 msd_buffer.FATFilesInfos[2].DIR_WrtDate[0]=0;
 msd_buffer.FATFilesInfos[2].DIR_WrtDate[1]=0;
 msd_buffer.FATFilesInfos[2].DIR_FstClusLO=2;
 msd_buffer.FATFilesInfos[2].DIR_FileSize=0;

 strcpy(msd_buffer.FATFilesInfos[3].DIR_Name,"TEST3   C  ");
 msd_buffer.FATFilesInfos[3].DIR_Attr=0x20;
 msd_buffer.FATFilesInfos[3].DIR_NTRes=0;
 msd_buffer.FATFilesInfos[3].DIR_CrtTimeTenth=0;
 msd_buffer.FATFilesInfos[3].DIR_CrtTime[0]=0;
 msd_buffer.FATFilesInfos[3].DIR_CrtTime[1]=0;
 msd_buffer.FATFilesInfos[3].DIR_CrtDate[0]=0;
 msd_buffer.FATFilesInfos[3].DIR_CrtDate[1]=0;
 msd_buffer.FATFilesInfos[3].DIR_LstAccDate[0]=0;
 msd_buffer.FATFilesInfos[3].DIR_LstAccDate[1]=0;
 msd_buffer.FATFilesInfos[3].DIR_FstClusHI=0;
 msd_buffer.FATFilesInfos[3].DIR_WrtTime[0]=0;
 msd_buffer.FATFilesInfos[3].DIR_WrtTime[1]=0;
 msd_buffer.FATFilesInfos[3].DIR_WrtDate[0]=0;
 msd_buffer.FATFilesInfos[3].DIR_WrtDate[1]=0;
 msd_buffer.FATFilesInfos[3].DIR_FstClusLO=5;
 msd_buffer.FATFilesInfos[3].DIR_FileSize=0x0001658A;

 for (i=4; i<16; i++) msd_buffer.FATFilesInfos[i].DIR_Name[0]=FAT_FILE_DELETED;
}

/*********************************************************************
  Function:    uint32 FAT_findFreeClusters(void)
  Description: Calcul et renvoi du nombre de cluster libres
********************************************************************/
uint32 FAT_findFreeClusters(void)
{
 uint32 freeclusters=0;
 int i,j;
 uint fatentry;
 for (i=0; i< BPB_FATSize; i++) 
 {// Pour chaque secteur de la FAT ...
  SD_CurrSector=FAT_StartSector+i;
  SectorRead(SD_CurrSector,&msd_buffer);
  for (j=0; j<256; j++) 
  {// ... compter le nombre d'entres de cluster  0
   fatentry=msd_buffer.mots_16b[j]; // A modifier pour FAT12 et FAT32
   if (fatentry==0) freeclusters++;
  }
 }
 return freeclusters;
}

/*********************************************************************
  Function:    uint32 FAT_ConvertClusterToSector(uint16 cluster)
  Description: Calcul de l'adresse du 1er secteur de donnes du cluster
********************************************************************/
uint32 FAT_ConvertClusterToSector(uint16 cluster) 
{
 uint32 sector;
 sector=(uint32)cluster-2;    	// Les 2 premires entres ne sont pas valides
 sector*=BPB_SectorsPerCluster; 
 sector+=FAT_FirstDataSector;   // Calcul de l'adresse du 1 secteur de donnes
 return sector;
}

/********************************************************************
  Function:    FAT_GetNextCluster(uint16 cluster)
  Description: Lecture du N du cluster suivant
********************************************************************/
uint16 FAT_GetNextCluster(uint16 cluster) 
{
 uint32 AdrFATbyte;   // Adresse "octet" relative de l'entre dans le bloc FAT
 uint32 AdrFATsector; // Adresse "secteur" relative de l'entre dans le bloc FAT
 uint32 OffsetFATcluster; // Offset de l'entre dans le secteur
 uint16 data;
 // Recherche du secteur ds le bloc FAT o se trouve l'entre
 if (FATType==FAT16) AdrFATbyte=cluster<<1;     // 2 octets par lment en FAT16
 else if (FATType==FAT32) AdrFATbyte=cluster<<2;// 4 octets par lment en FAT32
 // Adresse relative "secteur" ds le bloc FAT
 AdrFATsector=AdrFATbyte/BPB_BytesPerSector;
 // Adresse absolue du secteur contenant l'entre de cluster
 SD_CurrSector=FAT_StartSector+AdrFATsector;
 // Offset du 1 octet de l'lment ds ce secteur
 OffsetFATcluster=AdrFATbyte-(uint32)(AdrFATsector*BPB_BytesPerSector);
 // Lecture du secteur FAT concern ds le buffer
 SectorRead(SD_CurrSector,&msd_buffer);
 // Lecture du N du cluster suivant
 if (FATType==FAT16) data=msd_buffer.mots_16b[(uint16)OffsetFATcluster>>1];
 else if (FATType==FAT32) data=msd_buffer.mots_16b[(uint16)OffsetFATcluster>>2];
 //Attention: data devrait tre code sur 32 bits en FAT32 :data=msd_buffer.mots_32b[]
 return data;
}

/********************************************************************
  Function:    uint16 FAT_findNextUnusedCluster(uint16 cluster)
  Description: Recherche du n de cluster de la prochaine entre de cluster libre
********************************************************************/
uint16 FAT_findNextUnusedCluster(uint16 cluster) 
{
 uint16 i; // Compteur de secteurs ds le bloc FAT
 uint16 j; // Compteur d'lments ds un secteur
 // On commence avec le N de cluster suivant
 cluster+=1;
 i=cluster>>8; // Adresse du secteur ds la FAT (256 entres par secteur)
 for (; i<BPB_FATSize; i++) 
 {// Pour tous les secteurs restant jusqu' la fin de la FAT :
  SD_CurrSector=FAT_StartSector+i; // Adresse absolue du secteur ds le volume
  SectorRead(SD_CurrSector,&msd_buffer);// Transfert du secteur ds le buffer
  j=cluster&0x00FF; // Indice initial de l'lment ds le secteur courant
                    // (tjrs 0 pour les secteurs suivants)
  for (; j<256; j++) 
  {// Rechercher la prochaine entre libre ds le secteur "SD_CurrSector" :
   if (msd_buffer.mots_16b[j]==0) return (i*256+j);// FAT16
  }
  cluster=0; // Pour que les boucles suivantes commencent avec j=0
 }
 return 0;
}

/*********************************************************************
  Function:    void FAT_QuickFormat()
  Description: formatage rapide du volume
               A n'utiliser qu'en urgence !
********************************************************************/
void FAT_QuickFormat() 
{
 uint32 last;
 uint16 j;
 // Dernier secteur des structures de fichier
 last=FAT_RootDirStartSector+FAT_RootDirNumSectors;
 for (j=0; j< 256; j++) msd_buffer.mots_16b[j]=0; // Tous les octets du buffer  0
 for (SD_CurrSector=FAT_StartSector; SD_CurrSector<last; SD_CurrSector++) 
 {// Pour tous les secteurs du tableau des structures de fichiers :
  WTX2_Msg_Nb32bits("Initialisation du secteur ...:", SD_CurrSector);
  // Ecriture du buffer ds le secteur
  SectorWrite(SD_CurrSector,&msd_buffer); 
 }
}

/*********************************************************************
  Function:    int FAT_LoadFileInRoot(uint16 filenumber)
  Description: ouverture d'un fichier ds le rpertoire racine
  Paramtre  : indice de la structure de fichier ds le bloc des structures
  Retour     : TRUE              : le fichier existe et est "ouvert"
               FAT_END_OF_DIR    : le fichier n'existe pas et le n est le dernier
               FAT_FILE_DELETED  : le fichier n'existe pas (il a t effac)
               FAT_LONG_FILENAME : le fichier existe mais a un nom long (non gr)
********************************************************************/
int FAT_LoadFileInRoot(uint filenumber) 
{
 uint OffsetSector;
 char c;
 // 16 structures de fichiers par secteur (une structure = 32 octets)
 OffsetSector=filenumber>>4;         // Partie entire de (filenumber/16)
 // Indice de la structure d'entre de fichier ds le secteur
 File_DirEntryNum=filenumber&0x000F; // Reste de (filenumber/16)=indice ds le secteur
 SD_CurrSector=FAT_RootDirStartSector+OffsetSector; // Adresse absolue du secteur
 // Adresse courante du secteur de la structure de fichier
 File_DirSectorNum=SD_CurrSector; 
 // Lecture du secteur de la structure de fichier recherche
 WTX2_Msg_Nb32bits(" Lecture du secteur N ",SD_CurrSector);
 SectorRead(SD_CurrSector,&msd_buffer);
 c=msd_buffer.FATFilesInfos[File_DirEntryNum].DIR_Name[0];
 if (c==0x0) 
 {// Structure de fichier libre et c'est la dernire
  WriteLnTXD2_String(" Structure de fichier libre et c'est la derniere.");
  WriteLnTXD2();
  return FAT_END_OF_DIR;
 } 
 if (c==(char)FAT_FILE_DELETED)
 {// Structure de fichier libre
  WriteLnTXD2_String(" Structure de fichier libre.");
  WriteLnTXD2();
  return c;
 }
 // Le fichier existe : lecture des champs de la structure de fichier
 memcpy(File_Name,msd_buffer.FATFilesInfos[File_DirEntryNum].DIR_Name,11); // Nom court
 File_Attr=msd_buffer.FATFilesInfos[File_DirEntryNum].DIR_Attr; // Aucun attribut
 if (File_Attr == 0x0F) // Nom de fichier long, non pris en charge
       return FAT_LONG_FILENAME;
 File_StartCluster=msd_buffer.FATFilesInfos[File_DirEntryNum].DIR_FstClusLO;
 File_CurrCluster=File_StartCluster; // N du 1 cluster du fichier
 // Premier secteur de donne du 1 cluster
 File_StartSector=FAT_ConvertClusterToSector(File_StartCluster);
 File_CurrSector=File_StartSector;   // Secteur de donnes courant
 // Taille du fichier en octets
 File_Size=msd_buffer.FATFilesInfos[File_DirEntryNum].DIR_FileSize;
 File_CurrSectorNumInCluster=0;      // N du secteur ds le cluster courant
 File_CurrByte=0;                    // Compteur d'octets lus
 return TRUE;
}

/*********************************************************************
  Function:    int FAT_LoadFileWithName(char *filename)
  Description: Ouvrir un fichier ds le rpertoire racine avec son nom
  Paramtre  : Nom court du fichier avec son extension, mais sans le "."
  Retour     : Renvoi 
                 - du n du fichier en cas de succs
                 - de -1 en cas d'erreur
********************************************************************/
int FAT_loadFileWithName(char *filename) 
{
 uint16 filenum;
 byte ret;
 filenum=0; // N initial de la structure de fichier
 WriteTXD2_String("Recherche du n de fichier correspondant  : ");
 WriteLnTXD2_String(filename);
 WriteLnTXD2_String("********************************************");
 while (1) 
 {// Recherche du N de fichier correspondant au nom recherch
  WTX2_Msg_Nb8bits(" Test avec le fichier n : ",filenum);
  // Affectations de certaines variables "File_...", notamment "File_Name"
  ret=FAT_LoadFileInRoot(filenum); 
  if (ret==FAT_END_OF_DIR) 
  {
   WriteLnTXD2_String(" <EOD>");
   WriteLnTXD2();
   break;
  }
  if ((ret!=FAT_FILE_DELETED)&&(ret!=FAT_LONG_FILENAME)) 
  {
   if (memcmp(File_Name,filename,11)==0) // Comparaison du nom de fichier court 
   {
    WriteLnTXD2_String(" Fichier trouve !");
    WriteLnTXD2();
    return filenum;
   }
   else 
   {
    WriteTXD2_String(" Le nom ne correspond pas. ");
    WriteLnTXD2_String(File_Name);
    WriteLnTXD2();
   }
  }
  filenum++;
 }
 return -1;
}

/*******************************************************************
  Function:    int FAT_ReadNextSectorIntoBuff(void)
  Description: Lecture du secteur suivant du cluster courant ds le buffer
  Retour     : Renvoi de 
                  - TRUE si l'opration s'est bien passe
                  - FALSE ds le cas contraire
  Le fichier doit tre "ouvert" : affectations des variables "File_..."
********************************************************************/
int FAT_ReadNextSectorIntoBuff(void) 
{
 // Secteur suivant ds le cluster suivant ?
 if (File_CurrSectorNumInCluster==BPB_SectorsPerCluster)
 {// Oui : recherche du N du cluster suivant :
  WriteLnTXD2(); WTX2_Msg_Nb16bits("Ancien cluster N :",File_CurrCluster);
  File_CurrCluster=FAT_GetNextCluster(File_CurrCluster);
  WTX2_Msg_Nb16bits("Nouveau cluster N :",File_CurrCluster);
  if (File_CurrCluster==0||File_CurrCluster>0xFFF7) return FALSE;
  File_CurrSectorNumInCluster=0; // Modulo "BPB_SectorsPerCluster"
 }
 // Premier secteur du cluster courant
 File_CurrSector=(uint32)FAT_ConvertClusterToSector(File_CurrCluster);
 // Ajot de l'offset du secteur ds le cluster
 File_CurrSector+=File_CurrSectorNumInCluster;
 SD_CurrSector=File_CurrSector; // Nouveau secteur courant
 WriteLnTXD2();
 WTX2_Msg_Nb32bits("Lecture du secteur N : ",SD_CurrSector);
 WTX2_Msg_Nb8bits("Offset en nb de secteurs ds le cluster: ",File_CurrSectorNumInCluster);
 SectorRead(SD_CurrSector,&msd_buffer); // Lecture de ce secteur ds le buffer
 File_CurrSectorNumInCluster++; // Secteur suivant ds le cluster courant
 return TRUE;
}

/*******************************************************************
  Function:    int FAT_WriteBuffIntoNextSector(void)
  Description: Ecriture du secteur tampon dans le secteur suivant du fichier ouvert
               (soit dans le secteur suivant du cluster courant)
               Mise  jour des 2 clusters concerns
               Mise  jour de la taille du fichier
  Retour     : TRUE ou FALSE suivant l'opration effectue
  Le fichier doit tre "ouvert" : affectations des variables "File_..."
********************************************************************/
int FAT_WriteBuffIntoNextSector(void) 
{
 uint16 oldcluster;
 oldcluster=File_CurrCluster; // Sauvegarde du N du cluster courant
 // Ecriture ds le premier secteur du premier cluster d'un fichier vide ?
 if (File_StartCluster==0)
 {// Oui : recherche du prochain N de cluster libre :
  File_StartCluster=FAT_findNextUnusedCluster(2); // Premier cluster
  File_CurrCluster=File_StartCluster; // Cluster courant
  WriteLnTXD2(); WTX2_Msg_Nb16bits(" Premier N de cluster du fichier : ",File_StartCluster);
  File_CurrSector=(FAT_ConvertClusterToSector(File_CurrCluster))+File_CurrSectorNumInCluster;
  WTX2_Msg_Nb32bits(" Premier secteur de donnees : ",File_CurrSector);
  // Lecture du secteur de la structure de fichier du fichier courant
  SD_CurrSector=File_DirSectorNum;
  WTX2_Msg_Nb32bits(" Lecture du secteur N (structures de fichier) : ",SD_CurrSector);
  // Lecture du secteur courant ds le buffer
  SectorRead(SD_CurrSector,&msd_buffer); 
  // Affectation du N du 1 cluster
  msd_buffer.FATFilesInfos[File_DirEntryNum].DIR_FstClusLO=File_StartCluster; 
  WTX2_Msg_Nb32bits(" Ecriture du cluster initial ds le secteur N (structures de fichier) : ",SD_CurrSector);
  SectorWrite(SD_CurrSector,&msd_buffer); // Ecriture du buffer modifi ds le secteur
  // Prochaine entre de cluster  EOC
  SD_CurrSector=FAT_StartSector+(File_CurrCluster/256); // Adresse ds la FAT16
  WTX2_Msg_Nb32bits(" Lecture du secteur N (FAT) : ",SD_CurrSector);
  SectorRead(SD_CurrSector,&msd_buffer); // Lecture du secteur de la FAT
  msd_buffer.mots_16b[File_CurrCluster&0xff]=0xFFFF; // N cluster suivant = EOC
  WTX2_Msg_Nb32bits(" Ecriture ds ce secteur de EOC ds le cluster N : ",File_CurrCluster);
  SectorWrite(SD_CurrSector,&msd_buffer); // Ecriture du buffer modifi ds le secteur
  File_Size=0; // Taille initiale du fichier
 }

 // Ecriture ds le premier secteur du cluster suivant ?
 if (File_CurrSectorNumInCluster==BPB_SectorsPerCluster) 
 {// Oui : recherche du N du cluster suivant :
  File_CurrSectorNumInCluster=0; // Modulo "BPB_SectorsPerCluster"
  File_CurrCluster=FAT_findNextUnusedCluster(File_CurrCluster); // Cluster libre suivant
  File_CurrSector=(FAT_ConvertClusterToSector(File_CurrCluster));
  WriteLnTXD2(); WTX2_Msg_Nb16bits(" Nouveau N de cluster aloue : ",File_CurrCluster);
  if (File_CurrCluster==0||File_CurrCluster>0xFFF7) return FALSE;
  // Affectation du N du cluster suivant ds la FAT
  SD_CurrSector=FAT_StartSector+(oldcluster/256); // Adresse ds la FAT16
  WTX2_Msg_Nb32bits(" Lecture du secteur N (FAT) : ",SD_CurrSector);
  SectorRead(SD_CurrSector,&msd_buffer); // Lecture du secteur de la FAT
  msd_buffer.mots_16b[oldcluster&0xff]=File_CurrCluster; // N cluster suivant
  WTX2_Msg_Nb32bits(" Ecriture du N du cluster suivant ds le secteur N : ",SD_CurrSector);
  SectorWrite(SD_CurrSector,&msd_buffer); // Ecriture du buffer modifi ds le secteur
  // Prochaine entre de cluster  EOC
  SD_CurrSector=FAT_StartSector+(File_CurrCluster/256); // Adresse ds la FAT16
  WTX2_Msg_Nb32bits(" Lecture du secteur N (FAT) : ",SD_CurrSector);
  SectorRead(SD_CurrSector,&msd_buffer); // Lecture du secteur de la FAT
  msd_buffer.mots_16b[File_CurrCluster&0xff]=0xFFFF; // N cluster suivant = EOC
  WTX2_Msg_Nb32bits(" Ecriture ds ce secteur de EOC ds le cluster N : ",File_CurrCluster);
  SectorWrite(SD_CurrSector,&msd_buffer); // Ecriture du buffer modifi ds le secteur
 }

 // Ecriture des 512 octets du buffer d'criture ds le secteur courant
 WTX2_Msg_Nb32bits(" Lecture des 512 octets du secteur tampon : ",SecteurTampon);
 SectorRead(SecteurTampon,&msd_buffer); // Lecture du tampon d'criture
 SD_CurrSector=File_CurrSector; // Nouveau secteur courant
 WTX2_Msg_Nb32bits(" Ecriture des 512 octets du tampon ds le secteur N : ",SD_CurrSector);
 SectorWrite(SD_CurrSector,&msd_buffer); // Ecriture du buffer ds le secteur courant

 // Mise  jour de la taille du fichier :
 // Lecture du secteur de la structure de fichier du fichier courant
 SD_CurrSector=File_DirSectorNum;
 WTX2_Msg_Nb32bits(" Lecture du secteur N (structures de fichier) : ",SD_CurrSector);
 // Lecture du secteur courant ds le buffer
 SectorRead(SD_CurrSector,&msd_buffer); 
 // Nouvelle taille du fichier
 File_Size+=512; // 512 octets par secteur
 msd_buffer.FATFilesInfos[File_DirEntryNum].DIR_FileSize=File_Size; 
 WTX2_Msg_Nb32bits(" Mise a jour de la taille du fichier : ",File_Size);
 SectorWrite(SD_CurrSector,&msd_buffer); // Ecriture du buffer modifi ds le secteur

 File_CurrSectorNumInCluster++; // Secteur suivant ds le cluster courant
 File_CurrSector++;
 return TRUE;
}

/*********************************************************************
  Function:    void FAT_ReadFile(char *name)
  Description: Lecture d'un fichier, sortie sur UART2
  Paramtre  : Nom complet du fichier (8+3 caractres)
********************************************************************/
void FAT_ReadFile(char *name) 
{
 uint16 idx;
 if (FAT_loadFileWithName(name)==0xFFFF) return; // "Ouvrir" le fichier
 WriteLnTXD2();
 WriteTXD2_String("Lecture du fichier :");
 WriteLnTXD2_String(name);
 WriteLnTXD2_String("********************");
 while (File_CurrByte<File_Size) // Dernier octet du fichier lu ? 
 {// Non : lecture du secteur suivant 
  if (FAT_ReadNextSectorIntoBuff())
  {
   idx=0; // Indice ds le secteur de l'octet de donne
   while ((File_CurrByte<File_Size)&&(idx<512))
   {// Pour chaque octet du buffer ou jusqu'au dernier octet du fichier :
    WriteTXD2(msd_buffer.octets[idx++]); // Transmission sur UART2
    File_CurrByte++; // Compteur d'octets
   }
  }
 }
}

/*********************************************************************
  Function:    void FAT_PrintCurrFileInfo()
  Description: utilis par ListDir
********************************************************************/
void FAT_PrintCurrFileInfo() 
{
 WriteTXD2_String(" Nom et taille :");
 WriteTXD2_String(File_Name);
 WTX2_Msg_Nb32bits(",",File_Size);
 WTX2_Msg_Nb8bits(" Attributs : ",File_Attr);
 WTX2_Msg_Nb16bits(" N du 1 cluster (File_StartCluster):",File_StartCluster);
 WTX2_Msg_Nb32bits(" Premier secteur de donnees :",FAT_ConvertClusterToSector(File_StartCluster));
 if (File_Attr & FAT_IS_DIR) 
 {
  WriteLnTXD2_String(" <DIR>");
  WriteLnTXD2();
 }
 else WriteLnTXD2();
}

/*********************************************************************
  Function:    void FAT_ListDir(void)
  Description: Transmission sur TX2 des fichiers du rpertoire racine
********************************************************************/
void FAT_ListDir(void) 
{
 int filenum;
 char ret;
 uint32 freespace;
 freespace=FAT_findFreeClusters()*BPB_SectorsPerCluster*BPB_BytesPerSector;
 WriteLnTXD2();
 WTX2_Msg_Nb32bits("Espace libre en octets : ",freespace);
 WriteLnTXD2();
 filenum=0;
 while (1) 
 {
  WTX2_Msg_Nb16bits("Fichier N : ",filenum);
  ret=FAT_LoadFileInRoot(filenum); // "Ouvrir" le fichier
  if (ret==FAT_END_OF_DIR) 
  {
   WriteLnTXD2_String("<EOD>");
   return;
  }
  if ((ret!=(char)FAT_FILE_DELETED)&&(ret!=(char)FAT_LONG_FILENAME)) 
     FAT_PrintCurrFileInfo();
  if (ret==(char)FAT_LONG_FILENAME) 
  {
   WriteLnTXD2_String(" Nom de fichier long non pris en charge");
   WriteLnTXD2();
  }
  filenum++;
 }
}

/*********************************************************************
  Function:    int FAT_CreateFile(char *name, uint32 len)
  Description: Cration d'un fichier
  Paramtre  : Nom complet du fichier (8+3 caractres)
  Retour     : TRUE ou FALSE suivant l'opration
********************************************************************/
int FAT_CreateFile(char *name) 
{
 // first get a root dir entry
 char c;
 int i;
 // Recherche d'une entre de fichier libre ds la FAT
 WriteLnTXD2();
 WriteLnTXD2_String("Creation d'un fichier vide :");
 WriteLnTXD2_String("**************************");
 for (i=0;i< BPB_RootDirEntryCount;i++) 
 {// Pour chaque entre de fichier ou rpertoire de 32 octets :
  // Calcul du secteur de la structure courante
  SD_CurrSector=(uint32)FAT_RootDirStartSector+i/16; // 16 entres par secteur
  File_DirEntryNum=i%16; // Numro de l'entre ds le secteur (0  15) 
  File_DirSectorNum=SD_CurrSector;               // Secteur courant
  WTX2_Msg_Nb32bits("Lecture du secteur N : ",SD_CurrSector);
  SectorRead(SD_CurrSector,&msd_buffer); // Lecture de ce secteur ds le buffer
  WTX2_Msg_Nb16bits("Test de l'entree de fichier N : ",i);
  c=msd_buffer.FATFilesInfos[File_DirEntryNum].DIR_Name[0]; // 1 octet de la structure
  if (c==0x0||c==(char)FAT_FILE_DELETED) // Entre de fichier libre ?
  {// Oui :
   if (c == 0x0) 
   {// Libre et c'est la dernire dans l'tat actuel des fichiers
    WriteLnTXD2_String("EOD");
    if (File_DirEntryNum!=15) // Prochaine entre ds le secteur suivant ? 
    {// Non : prochaine entre = EOD
     msd_buffer.FATFilesInfos[File_DirEntryNum+1].DIR_Name[0]=0;
    } 
    else if ((i+1)!=BPB_RootDirEntryCount) 
    {// Oui et il y a encore de la place :
     SD_CurrSector++; // Secteur suivant
     WTX2_Msg_Nb32bits("Lecture du secteur N : ",SD_CurrSector);
     SectorRead(SD_CurrSector,&msd_buffer); // Lecture de ce secteur ds le buffer
     // Entre de fichier suivante libre et c'est la dernire
     msd_buffer.FATFilesInfos[File_DirEntryNum].DIR_Name[0]=0; 
     WTX2_Msg_Nb32bits("Ecriture ds le secteur N : ",SD_CurrSector);
     SectorWrite(SD_CurrSector,&msd_buffer); // Ecriture du buffer modifi ds le secteur
     SD_CurrSector--; // Retour au secteur de l'entre de fichier en cours
     SectorRead(SD_CurrSector,&msd_buffer); // Lecture de ce secteur ds le buffer
    } 
    else 
    {// Plus de place pour une nouvelle structure de fichier !
     WriteLnTXD2_String("La derniere entree !!!!");
     return FALSE;
    }
   } 
   else 
   {// c==FAT_FILE_DELETED : la structure est libre et ce n'est pas la dernire
    WriteLnTXD2_String("Creation de fichier ds une entre effacee.");
   }
   // Affectation du champ "DIR_Name"
   memcpy(msd_buffer.FATFilesInfos[File_DirEntryNum].DIR_Name,name,11);
   // Affectation du champ "attribut"
   msd_buffer.FATFilesInfos[File_DirEntryNum].DIR_Attr=0;
   // Affectation du champ "DIR_FstClusLO" (16 bits) avec 0 (longueur 0)
   msd_buffer.FATFilesInfos[File_DirEntryNum].DIR_FstClusHI=0;
   msd_buffer.FATFilesInfos[File_DirEntryNum].DIR_FstClusLO=0;
   // Affectation du champ "DIR_FileSize" (32 bits) : 0
   msd_buffer.FATFilesInfos[File_DirEntryNum].DIR_FileSize=0;
   // Autres informations :
   msd_buffer.FATFilesInfos[File_DirEntryNum].DIR_NTRes=0;
   msd_buffer.FATFilesInfos[File_DirEntryNum].DIR_CrtTimeTenth=0;
   msd_buffer.FATFilesInfos[File_DirEntryNum].DIR_CrtTime[0]=0;
   msd_buffer.FATFilesInfos[File_DirEntryNum].DIR_CrtTime[1]=0;
   msd_buffer.FATFilesInfos[File_DirEntryNum].DIR_CrtDate[0]=0;
   msd_buffer.FATFilesInfos[File_DirEntryNum].DIR_CrtDate[1]=0;
   msd_buffer.FATFilesInfos[File_DirEntryNum].DIR_LstAccDate[0]=0;
   msd_buffer.FATFilesInfos[File_DirEntryNum].DIR_LstAccDate[1]=0;
   msd_buffer.FATFilesInfos[File_DirEntryNum].DIR_WrtTime[0]=0;
   msd_buffer.FATFilesInfos[File_DirEntryNum].DIR_WrtTime[1]=0;
   msd_buffer.FATFilesInfos[File_DirEntryNum].DIR_WrtDate[0]=0;
   msd_buffer.FATFilesInfos[File_DirEntryNum].DIR_WrtDate[1]=0;
   WTX2_Msg_Nb32bits("Ecriture dans le secteur : ",SD_CurrSector);

   File_Attr=0; // Aucun attribut
   File_StartCluster=0;
   File_CurrCluster=File_StartCluster;
   File_Size=0;
   File_CurrSectorNumInCluster=0;

   TEST0=1; // Pour dclencher l'oscillo ou l'analyseur logique
   SectorWrite(SD_CurrSector,&msd_buffer); // Ecriture du buffer modifi ds le secteur
   TEST0=0; // Pour dtecter la fin de l'criture d'un secteur
   return TRUE;
  }// Fin du bloc "if (c==0x0||c==(char)FAT_FILE_DELETED)"
 }// Fin du bloc "for (i=0; i< BPB_RootDirEntryCount;  i++)"
 return FALSE;
}

/*********************************************************************
  Function:    void FAT_DeleteFile(char *name)
  Description: Effacer le fichier 
  Paramtre  : Nom complet du fichier  effacer (8+3 caractres)
********************************************************************/
void FAT_DeleteFile(char *name) 
{
 uint16 cluster, newcluster;
 WriteLnTXD2();
 WriteTXD2_String("Effacer le fichier ");
 WriteLnTXD2_String(name);
 WriteLnTXD2_String("******************");
 // Ouverture du fichier : affectation des variables "File_..."
 if (FAT_loadFileWithName(name)==0xFFFF) return;
 FAT_PrintCurrFileInfo(); // Transmettre les principales infos sur TXD2
 WriteLnTXD2_String("Effacer la chaine de cluster :");
 newcluster=File_StartCluster; // Premier cluster du fichier
 WTX2_Msg_Nb16bits(" On commence par le cluster N : ",newcluster);
 cluster=newcluster;
 SD_CurrSector=FAT_StartSector+(cluster/256); // Adresse du secteur (FAT16)
 SectorRead(SD_CurrSector,&msd_buffer);// Transfert du secteur ds le buffer
 while ((newcluster<=0xFFF8)&&(newcluster!=0x0000)) 
 {// Pour chaque cluster valide du fichier :
  cluster=newcluster;
  newcluster=msd_buffer.mots_16b[cluster&0xFF];// N du cluster suivant (FAT16)
  WTX2_Msg_Nb16bits(" On poursuit avec le cluster N : ",newcluster);
  msd_buffer.mots_16b[cluster&0xFF]=0x0; // "Effacer" le cluster courant
  if ((newcluster>=0xFFF8)||(newcluster==0x0000)) {break;}
  if ((cluster/256)!=(newcluster/256)) 
  {// Cluster suivant dans un autre secteur (FAT16) :
   WTX2_Msg_Nb32bits(" Ecriture ds le secteur N : ",SD_CurrSector);
   SectorWrite(SD_CurrSector,&msd_buffer);// Ecriture du buffer ds le secteur courant
   SD_CurrSector=FAT_StartSector+(newcluster/256);//Premier secteur du cluster suivant
   // Transfert du secteur ds le buffer pour la boucle suivante
   SectorRead(SD_CurrSector,&msd_buffer);
  }
 }
 WTX2_Msg_Nb32bits(" Ecriture ds le secteur N : ",SD_CurrSector);
 SectorWrite(SD_CurrSector,&msd_buffer);// Ecriture du buffer ds le dernier secteur
 // "Effacer" l'entre de la structure de fichier
 SD_CurrSector=File_DirSectorNum; // Secteur de la structure de fichier  effacer
 WTX2_Msg_Nb32bits(" Effacer l'entree de la structure de fichier ds le secteur : ",SD_CurrSector);
 SectorRead(SD_CurrSector,&msd_buffer);// Transfert du secteur ds le buffer
 // Affectation du 1 octet : structure libre
 WriteLnTXD2_String("Liberer la structure d'entree de fichier.");
 msd_buffer.FATFilesInfos[File_DirEntryNum].DIR_Name[0]=FAT_FILE_DELETED;
 WTX2_Msg_Nb32bits(" Ecriture ds le secteur N : ",SD_CurrSector);
 SectorWrite(SD_CurrSector,&msd_buffer);// Ecriture du buffer ds le dernier secteur
 return;
}

/*********************************************************************
  Function:    void FAT_Init(void)
  Description: Affectation des variables lie  la FAT de la SD Card     
********************************************************************/
int FAT_Init(void)
{
 WriteLnTXD2();
 WriteLnTXD2_String("Nouvelle SD Card :");
 WriteLnTXD2_String("******************");
 SD_CurrSector=0;
 if (SDC_Status!=sdcValid) return FALSE;
 // Lecture du secteur 0
 if (SectorRead(0,&msd_buffer)!=sdcValid) return FALSE;
 BPB_StartSector=msd_buffer.mots_16b[454/2];
 WTX2_Msg_Nb16bits("Secteur de boot (BPB_StartSector)= ",BPB_StartSector);

 // Lecture du secteur de boot (infos BPB de la FAT)
 if (SectorRead(BPB_StartSector,&msd_buffer)!=sdcValid) return FALSE;
 // Affectations des paramtres et variables de la FAT
 BPB_BytesPerSector=     msd_buffer.BPB_FAT.BPB_BytsPerSec[0]
                    +256*msd_buffer.BPB_FAT.BPB_BytsPerSec[1];
 WTX2_Msg_Nb16bits("Nombre d'octets par secteur (BPB_BytesPerSector)= "
                   ,BPB_BytesPerSector);

 BPB_SectorsPerCluster=msd_buffer.BPB_FAT.BPB_SecPerClus;
 WTX2_Msg_Nb16bits("Nombre de secteurs par cluster (BPB_SectorsPerCluster)= "
                   ,BPB_SectorsPerCluster);

 BPB_ReservedSectorCount=msd_buffer.BPB_FAT.BPB_RsvdSecCnt;
 WTX2_Msg_Nb16bits("Nombre de secteurs reserves (BPB_ReservedSectorCount)= "
                   ,BPB_ReservedSectorCount);

 // On utilise un secteur rserv comme tampon d'criture
 SecteurTampon=BPB_StartSector+1; 
 WTX2_Msg_Nb32bits("Secteur reserve pour le buffer d'criture (SecteurTampon):"
                   ,SecteurTampon);

 BPB_NumFATs=msd_buffer.BPB_FAT.BPB_NumFATs;
 WTX2_Msg_Nb16bits("Nombre de FATs (BPB_NumFATs)= ",BPB_NumFATs);

 BPB_RootDirEntryCount=     msd_buffer.BPB_FAT.BPB_RootEntCnt[0]
                       +256*msd_buffer.BPB_FAT.BPB_RootEntCnt[1];
 WTX2_Msg_Nb16bits("Nombre d'entrees de repertoire (BPB_RootDirEntryCount)= "
                   ,BPB_RootDirEntryCount);

 BPB_TotSec=(uint32)msd_buffer.BPB_FAT.BPB_TotSec16[0]
            +256*(uint32)msd_buffer.BPB_FAT.BPB_TotSec16[1];
 if (BPB_TotSec==0) BPB_TotSec=msd_buffer.BPB_FAT.BPB_TotSec32;
 WTX2_Msg_Nb32bits("Nombre total de secteurs (BPB_TotSec)= ",BPB_TotSec);

 if (msd_buffer.BPB_FAT.BPB_FATSz16!=0) BPB_FATSize=msd_buffer.BPB_FAT.BPB_FATSz16;
                         else BPB_FATSize=msd_buffer.BPB_FAT.BPB_FATSz32;
 WTX2_Msg_Nb32bits("Taille de la FAT en secteurs (BPB_FATSize)= ",BPB_FATSize);

 FAT_StartSector=BPB_StartSector+BPB_ReservedSectorCount;
 WTX2_Msg_Nb16bits("1 secteur du bloc FAT (FAT_StartSector) = ",FAT_StartSector);

 FAT_RootDirNumSectors=((BPB_RootDirEntryCount*32)+(BPB_BytesPerSector-1))/BPB_BytesPerSector;
 WTX2_Msg_Nb16bits("Nombre de secteurs du repertoire racine (FAT_RootDirNumSectors)= "
                   ,FAT_RootDirNumSectors);

 BPB_DataSec=BPB_TotSec-(BPB_ReservedSectorCount+(BPB_NumFATs*BPB_FATSize)+FAT_RootDirNumSectors);
 WTX2_Msg_Nb32bits("Nombre de secteurs dans la zone de donnees (BPB_DataSec)= "
                   ,BPB_DataSec);

 BPB_CountofClusters=BPB_DataSec/BPB_SectorsPerCluster;
 WTX2_Msg_Nb16bits("Nombre de cluster (BPB_CountofClusters)= ",BPB_CountofClusters);

 // Type de FAT
 if (BPB_CountofClusters<4085) 
  {
   FATType=FAT12;
   WriteLnTXD2_String("FAT12");
  } 
  else if (BPB_CountofClusters<65525) 
   {
    FATType=FAT16;
    WriteLnTXD2_String("FAT16");
   }
   else 
    {
     FATType=FAT32;
     WriteLnTXD2_String("FAT32");
    }

 FAT_RootDirStartSector=FAT_StartSector+BPB_NumFATs*BPB_FATSize;
 WTX2_Msg_Nb16bits("Premier secteur du repertoire racine (FAT_RootDirStartSector)= ",
                    FAT_RootDirStartSector);

 FAT_FirstDataSector=BPB_StartSector+BPB_ReservedSectorCount+(BPB_NumFATs*BPB_FATSize)+FAT_RootDirNumSectors;
 //FAT_FirstDataSector=FAT_RootDirStartSector+FAT_RootDirNumSectors;
 WTX2_Msg_Nb16bits("Premier secteur de donnee (FAT_FirstDataSector)= "
                   ,FAT_FirstDataSector);

 return TRUE;
}

                        /*****************************
                        *    Programme principal     *
                        ******************************/
int main (void)
{
 int i,j;
 Reset_FAT();
 U1MODE=0;
 U2MODE=0;       // Inhiber l'UART 2  cause du bootloader
 TEST0_Dir=0; TEST1_Dir=0; // TEST0 et TEST1 en sortie
 ADPCFG=0xFFFF;  // Broches ADC configures en port E/S
 // Initialisation UART1 & UART2
 InitUART1(); // GPS : 4800 bauds, 8 bits, pas de parit, 1 bit stop
              // Sur ports "auxiliaires" U1ATX et U1ARX
 InitUART2(); // PC : 38400 bauds, 8 bits, pas de parit, 1 bit stop
 // Initialisation Timer 1
 InitTMR1();  // Initialisation Timer 1 pour rythmer les transmissions vers PC
 // Initialisation Timer 3
 InitTMR3();  // Initialisation Timer 3 pour mesure vitesse et dlais softs
 SocketInitialize(); // Initialisation des ports utiliss par la SD Card
 // Raz indicateurs interruptions UART1
 IFS0bits.U1TXIF=0; // Clear interrupt flag TX
 IFS0bits.U1RXIF=0; // Clear interrupt flag RX
 // Raz indicateurs interruptions UART2
 IFS1bits.U2TXIF=0; // Clear interrupt flag TX
 IFS1bits.U2RXIF=0; // Clear interrupt flag RX
 // Validations interruptions UART1
 IEC0bits.U1TXIE=1; // Validation interruption TX
 IEC0bits.U1RXIE=1; // Validation interruption RX
 // Validations interruptions UART2
 IEC1bits.U2TXIE=1; // Validation interruption TX
 IEC1bits.U2RXIE=1; // Validation interruption RX
 // FAT_Init();
 // FAT_ListDir();
 // for (i=0;i<10;i++) FAT_LoadFileInCWD(i);
 // FAT_ConvertClusterToSector(2);
 // FAT_ReadFile("TEST1   TXT");
 // SampleFileData();
 // FAT_CreateFile("12345678TXT");
 // for (i=0; i<40; i++)
 // {
  // Initialisation du buffer
  // for (j=0; j<512; j++) msd_buffer.octets[j]='A'+i;
  // Ecriture du buffer ds le secteur tampon
  // SectorWrite(SecteurTampon,&msd_buffer);
  // Ecriture du secteur tampon dans le secteur suivant du fichier ouvert
  // FAT_WriteBuffIntoNextSector();
 // }
 while (1)
  {
   if (Flags.SDC_CD_New) // Chgt d'tat stable du contact CD de la SD Card ?
   {
    Flags.SDC_CD_New=0;  // Acquitement du chgt d'tat
    if (Flags.SDC_CD_OK) // Insertion d'une SD Card ?
    {
     // Initialisation SD Card
     SDC_Status=MediaInitialize();
     if (SDC_Status!=sdcValid) ShutdownMedia();
     else 
     {
      FAT_Init();
      //FAT_ListDir();
      //for (i=0;i<10;i++) FAT_LoadFileInCWD(i);
      //FAT_loadFileWithName("TEXTE2  TXT");
      //FAT_ReadFile("TEXTE3  TXT");
      //FAT_CreateFile("12345678TXT"); // Cration du fichier "12345678.TXT"
      FAT_DeleteFile("12345678TXT");
      /*WriteLnTXD2();
      WriteLnTXD2_String("Affectation du fichier ouvert :");
      WriteLnTXD2_String("*******************************");
      for (i=0; i<40; i++)
      {
       // Initialisation du buffer
       for (j=0; j<512; j++) msd_buffer.octets[j]='A'+i;
       // Ecriture du buffer ds le secteur tampon
       WriteLnTXD2_String(" Ecriture de 512 octets ds le secteur tampon.");
       SectorWrite(SecteurTampon,&msd_buffer);
       // Ecriture du secteur tampon dans le secteur suivant du fichier ouvert
       FAT_WriteBuffIntoNextSector();
       WriteLnTXD2();
      }*/
     }
    }
    else ShutdownMedia();
   }
   Receive_NMEA_RXD1(); // Rception des trames NMEA du GPS sur U1RX
   Detect_Cmd_RXD2(); // Dtection et excution ventuelle d'une commande sur U2RX   
  }
}
