     /********************************************************************
      ***                      Gestion SD Card                         ***
      ***               Programme principal de tests                   ***
      ***                      dsPIC30F4011                            ***
      ***                                                              ***
      ***  Fichier Test_SD_Card.c                                      ***
      ***  Auteur : CREMMEL Marcel                                     ***
      ***  Date   : 30/09/2007                                         ***
      ********************************************************************/

#include <p30f4011.h>
#include <stdlib.h>
#include <string.h>
#include "Hardware.h" // Pour TEST0 et TEST1
#include "Typedefs_FAT.h"
#include "Timer1.h"   // Pour InitTMR1()
#include "Timer3.h"   // Pour InitTMR3()
#include "UART2.h"    // Pour sortie des messages sur TX2 avec printf
#include "SD_Card.h"
#include "FAT.h"

#define STATUSLED

/*------------------------------------------------------------------------------
                                  Variables                              
------------------------------------------------------------------------------*/

/*----------
   SD Card
------------*/
// Indicateurs de fontionnement
struct 
{
 unsigned SDC_CD_OK : 1; // Dtection prsence carte (avec anti-rebonds)
 unsigned SDC_CD_old: 1; // Etat prcdent du contact CD
 unsigned SDC_CD_New: 1; // Nouvel tat stable du contact CD
 unsigned isSDMMC   : 1; // Indicateur de carte SDMMC 
 unsigned isWP      : 1; // Indicateur de protection en criture
 unsigned unused	: 11;
}FlagsSD;
// Couche physique
SDC_Error SDC_Status=sdcNoCard; // Etat de la carte
// Dclaration multitypes du buffer pour un secteur
Secteur *msd_buffer;  // Buffer de lecture et d'criture d'un secteur
// Registre CSD
CSD gblCSDReg;        // Registre CSD : Card Specific Data (128 bits ou 16 octets)

/*-------------------
   Autres variables
---------------------*/
unsigned char Buffer2_Cmd[5];   // Buffer pour la rception des commandes

/*------------------------------------------------------------------------------
             Dclarations des fonctions en mmoire flash
------------------------------------------------------------------------------*/

/*----------------------------------------------------------------------
  Function:        void Detect_Cmd_RXD2(void)
  Description:     Dtection et excution d'une ventuelle commande reue
                   Les commandes ont toujours une longueur de 5 octets :
                             Synchro, code, Dh, Dl, Synchro
                      - code : identifie la commande
                      - Dh, Dl : paramtre de la commande
------------------------------------------------------------------------*/
void Detect_Cmd_RXD2(void)
{
 int i;
 char c;
 if (!ReadRXD2(&c)) return;
 for (i=1; i<5; i++) Buffer2_Cmd[i-1]=Buffer2_Cmd[i];
 Buffer2_Cmd[4]=c;
 if ((Buffer2_Cmd[0]==Synchro) && (Buffer2_Cmd[4]==Synchro))
 {
  switch(Buffer2_Cmd[1]) // Code de la commande
  {
   case 'R' : // Lecture d'un secteur SD Card
   {
    if (SectorRead(Buffer2_Cmd[2]*256+Buffer2_Cmd[3],msd_buffer)==sdcValid)
       for(i=0; i<SDC_SECTOR_SIZE; i++) WriteTXD2(msd_buffer->octets[i]);
    break;
   }
   case 'W' : // Ecriture d'un secteur SD Card
   {
    break;
   }
   case 'A' : // Renvoi de C_SIZE
   {
	WriteTXD2('Z'); WriteTXD2('A');
	WriteIntTXD2(gblCSDReg.C_SIZE_H*4+gblCSDReg.C_SIZE_L);
	WriteTXD2('Z'); 
    break;
   }
   case 'B' : // Renvoi de C_SIZE_MULT et READ_BL_LEN
   {
	WriteTXD2('Z'); WriteTXD2('B');
	WriteTXD2(gblCSDReg.C_SIZE_MULT_H*2+gblCSDReg.C_SIZE_MULT_L);
	WriteTXD2(gblCSDReg.READ_BL_LEN);
	WriteTXD2('Z'); 
    break;
   }
   case 'C' : // Renvoi de TAAC et TRAN_SPEED
   {
	WriteTXD2('Z'); WriteTXD2('C');
	WriteTXD2(gblCSDReg.TAAC);
	WriteTXD2(gblCSDReg.TRAN_SPEED);
	WriteTXD2('Z'); 
    break;
   } 
   case 'D' : // Renvoi de R2W_FACTOR
   {
	WriteTXD2('Z'); WriteTXD2('D');
	WriteTXD2(gblCSDReg.R2W_FACTOR);
	WriteTXD2(0);
	WriteTXD2('Z'); 
    break;
   } 
   default : break;
  }
 }
}

/*------------------------------------------------------------------------------
                             Fonction "main"
------------------------------------------------------------------------------*/
int main(void)
{
 U1MODE=0;
 U2MODE=0;     // Inhiber l'UART 2  cause du bootloader
 TEST0_Dir=0; TEST1_Dir=0; // TEST0 et TEST1 en sortie
 ADPCFG=0xFFFF;// Broches ADC configures en port E/S
 // Initialisation UART2
 InitUART2(); // PC : 38400 bauds, 8 bits, pas de parit, 1 bit stop
 // Initialisation Timer 1
 InitTMR1();  // Initialisation Timer 1 pour rythmer les transmissions vers PC
 // Initialisation Timer 3
 InitTMR3();  // Initialisation Timer 3 pour mesure vitesse et dlais softs
 // Initialisation des ports utiliss par la SD Card
 SocketInitialize(); 
 // Raz indicateurs interruptions UART2
 IFS1bits.U2TXIF=0; // Clear interrupt flag TX
 IFS1bits.U2RXIF=0; // Clear interrupt flag RX
 // Validations interruptions UART2
 IEC1bits.U2TXIE=1; // Validation interruption TX
 IEC1bits.U2RXIE=1; // Validation interruption RX
 while (1)
 {
  if (FlagsSD.SDC_CD_New) // Chgt d'tat stable du contact CD de la SD Card ?
  {
   FlagsSD.SDC_CD_New=0;  // Acquitement du chgt d'tat
   if (FlagsSD.SDC_CD_OK) // Insertion d'une SD Card ?
   {// Oui :
    // Allocation mmoire pour les 512 octets du buffer de secteur
    msd_buffer=(Secteur *)malloc(512);
    if (msd_buffer==NULL)
      WriteLnTXD2_String("Erreur  l'allocation de la mmoire du buffer");
    else
    {
     // Initialisation SD Card
     SDC_Status=MediaInitialize();
     if (SDC_Status!=sdcValid) ShutdownMedia();
     else if (FAT_Init()==FALSE) ShutdownMedia();
    }
   }
   else ShutdownMedia();
  }
  Detect_Cmd_RXD2(); // Dtection et excution ventuelle d'une commande sur U2RX   
 }
}
