     /********************************************************************
      ***                      Projet VAE                              ***
      ***               Fonctions lies au Timer 1                     ***
      ***                      dsPIC30F4011                            ***
      ***                                                              ***
      ***  Fichier "source" : Timer1.c                                 ***
      ***  Auteur : CREMMEL Marcel                                     ***
      ***  Date   : 30/09/2007                                         ***
      ********************************************************************/

#include <p30f4011.h>
#include "Typedefs.h"
#include "Timer1.h"
#include "Hardware.h" // Pour TEST0
#include "SD_Card.h"  // Pour SDC_CD_AR()

#define NbDtaTxd 6    // Nombre de paramtres transmis sur TXD2

/*------------------------------------------------------------------------------
                                  Variables                              
------------------------------------------------------------------------------*/
uint Compt_T1_TXD; // Compteur d'interruption Timer 1 pour transmission datas

/*------------------------------------------------------------------------------
             Dclarations des fonctions en mmoire flash
------------------------------------------------------------------------------*/

/*--------------------------------------------------------          
  Function:        void InitTMR1(void)
  Description:     Initialisation du Timer 1 (Timer A) pour provoquer
                   une interruption  rythme rgulier.
  Note : le rythme est choisi pour obtenir une frquence de transmission
         de l'ensemble des paramtres sur TXD de 10Hz
-----------------------------------------------------------*/
void InitTMR1(void)
{
 T1CON=0x0010;	 // Fonctionnement continu, horloge=Fcy/8=2MHz
 TMR1=0;         // Raz du compteur pour dcaler la 1 interruption
 PR1=(FCY/8)/(10*NbDtaTxd);	// Frquence = 10Hz x NbDtaTxd
 T1CONbits.TON=1;// Validation Timer 1 
 IFS0bits.T1IF=0;// Raz indicateur interruption Timer 1 
 IEC0bits.T1IE=1;// Validation interruption Timer 1 (RTI)
}

/*--------------------------------------------------------          
  Function   : void _ISR _T1Interrupt (void)
  Descrition : Frquence = 10Hz x NbDtaTxd
  Traitement : 
    - Anti-rebonds sur SDC_SD (contact de dtection carte SD) pour 
      affecter l'indicateur SDC_CD_OK
-----------------------------------------------------------*/
void _ISR _T1Interrupt(void)
{
 //TEST0=1;
 IFS0bits.T1IF=0; // Acquitement interruption
 Compt_T1_TXD++;
 if (Compt_T1_TXD==NbDtaTxd) 
  {// Rythme d'excution = 10Hz
   TEST1=!TEST1;   // Pour contrle avec un oscilloscope (TEST1=port RE4)
   Compt_T1_TXD=0; // Compt_T1_TXD modulo NbDtaTxd
   SDC_CD_AR();    // Affectation des indicateurs SDC_CD_OK et SDC_CD_new
  }
 //TEST0=0; 
}

