     /********************************************************************
      ***                      Projet VAE                              ***
      ***               Fonctions lies au Timer 3                     ***
      ***                      dsPIC30F4011                            ***
      ***                                                              ***
      ***  Fichier "source" : Timer3.c                                 ***
      ***  Auteur : CREMMEL Marcel                                     ***
      ***  Date   : 30/09/2007                                         ***
      ********************************************************************/

#include <p30f4011.h>
#include "Timer3.h"

/*------------------------------------------------------------------------------
                                  Variables                              
------------------------------------------------------------------------------*/
uint TMR3_MSB;     // Poids fort de TMR3

/*------------------------------------------------------------------------------
             Dclarations des fonctions en mmoire flash
------------------------------------------------------------------------------*/

/*--------------------------------------------------------          
  Function   : void InitTMR3(void)
  Description: Initialisation du Timer 3 (Timer C)                       
               Horloge = Fcy/256 = 62,5kHz
               Utilis pour : - mesurer la vitesse du moteur        
                              - produire des dlais logiciels
-----------------------------------------------------------*/
void InitTMR3(void)
{
 T3CON=0x0030; // Horloge interne Fcy/256 = 62,5kHz
 TMR3=0;       // Raz du compteur pour retarder la 1 demande d'interruption
 PR3=0xFFFF;   // Comptage de 0x0000  0xFFFF au reset
 T3CONbits.TON=1; // Timer 3 "On"
 IEC0bits.T3IE=1; // Validation interruption Timer 3
}

/*********************************************************************
  Function   : void _ISR _T3Interrupt (void)
  Descrition : Activ au dpassement de TMR3
               Traitement : 
                 - Incrmentation TMR3_MSB au dpassement de TMR3
********************************************************************/
void _ISR _T3Interrupt (void)
{
 IFS0bits.T3IF=0; // Acquitement interruption
 // Incrmentation du compteur 32 bits
 TMR3_MSB++;     
 TMR3=0;
}

/*--------------------------------------------------------          
  Function   : void Delay(uint16 Delai)
  Description: Dlai "logiciel"                       
               Utilise le Timer3
  Paramtre  : Delai : dlai (rsolution 1/62500 s)
               Dlai maximum : 68719 s !               
-----------------------------------------------------------*/
void Delay(uint16 Delai)
{
 uint32 Temps_debut,Temps_actuel;
 IEC0bits.T3IE=0; // Inhibition interruption Timer 3 pdt lecture TMR3
 Temps_debut=TMR3+((unsigned long)TMR3_MSB<<16);//Lecture de TMR3 sur 32 bits
 IEC0bits.T3IE=1; // Validation interruption Timer 3
 do
 {
  IEC0bits.T3IE=0; // Inhibition interruption Timer 3 pdt lecture TMR3
  Temps_actuel=TMR3+((unsigned long)TMR3_MSB<<16);//Lecture de TMR3 sur 32 bits
  IEC0bits.T3IE=1; // Validation interruption Timer 3
 } while ((Temps_actuel-Temps_debut)<=Delai);
}
