     /********************************************************************
      ***                      Projet VAE                              ***
      ***          Dfinitions de types pour gestion FAT               ***
      ***                      dsPIC30F4011                            ***
      ***                                                              ***
      ***  Auteur : CREMMEL Marcel                                     ***
      ***  Date : 16/07/2007                                           ***
      ********************************************************************/

              /***************************************************
              *     Dfinitions de types pour gestion FAT        *
              ****************************************************/
                      
#ifndef FAT_Type_Defs
#define FAT_Type_Defs

#include "Typedefs.h"
#include "Typedefs_SD.h"

/*************************************
  Structure BPB (BIOS Parameter block)
**************************************/
typedef union
{
 struct
 {// Structure dtaille FAT12 et FAT16
  char BS_jmpBoot    [3]; // Instruction de saut vers le code de dmarrage
  char BS_OEMName    [8]; // Indicateur du systme qui a format le volume
  char BPB_BytsPerSec[2]; // Nb d'octets par secteur
  char BPB_SecPerClus   ; // Nb de secteurs par unit d'allocation (cluster)
  uint BPB_RsvdSecCnt   ; // Nb de secteurs rservs ds la zne rserve
  char BPB_NumFATs      ; // Nb de structures de donnes FAT sur le volume
  char BPB_RootEntCnt[2]; // Nb d'entres de rpertoires de 32 octets (racine)
  char BPB_TotSec16[2]  ; // Nb total de secteurs. Si 0, voir BPB_TotSec32
  char BPB_Media        ; // Code du mdia utilis (0xF0 pour amovibles)
  uint BPB_FATSz16      ; // Nb de secteurs occups par une FAT
  uint BPB_SecPerTrk    ; // Nb de secteurs par pistes (disquettes ou HD)
  uint BPB_NumHeads     ; // Nb de "ttes" de lecture (disquettes ou HD)
  uint32 BPB_HiddSec    ; // Nb de sect prcdant la partition FAT (0 si mdias non part.)
  uint32 BPB_TotSec32   ; // Nb total de secteurs sur le volume
  char BS_DrvNum        ; // N de lecteur de l'interruption 0x13 (MSDOS)
  char BS_Reserved1     ; // Utilis par Windows NT ( 0 au formatage)
  char BS_BootSig       ; // Signature de boot tendue (0x29). Indique la prsence
                          // des 3 champs suivants
  char BS_VolID      [4]; // N d'identification du volume
  char BS_VolLab    [11]; // Nom du volume
  char BS_FilSysType [8]; // Normalement 'FATxx '
  char Not_used    [448]; // Bloc d'octets non utiliss
  uint SignSector0      ; // Signature du secteur 0 (0x55AA)
 };
 struct
 {// Structure dtaille FAT32 
  char BS_jmpBoot    [3]; // Instruction de saut vers le code de dmarrage
  char BS_OEMName    [8]; // Indicateur du systme qui a format le volume
  char BPB_BytsPerSec[2]; // Nb d'octets par secteur
  char BPB_SecPerClus   ; // Nb de secteurs par unit d'allocation (cluster)
  uint BPB_RsvdSecCnt   ; // Nb de secteurs rservs ds la zne rserve
  char BPB_NumFATs      ; // Nb de structures de donnes FAT sur le volume
  char BPB_RootEntCnt[2]; // Non utilis en FAT32; tjrs  0
  char BPB_TotSec16[2]  ; // Nb total de secteurs. Si 0, voir BPB_TotSec32
  char BPB_Media        ; // Code du mdia utilis (0xF0 pour amovibles)
  uint BPB_FATSz16      ; // Non utilis en FAT32; tjrs  0 (voir BPB_FATSz32)
  uint BPB_SecPerTrk    ; // Nb de secteurs par pistes (disquettes ou HD)
  uint BPB_NumHeads     ; // Nb de "ttes" de lecture (disquettes ou HD)
  uint32 BPB_HiddSec    ; // Nb de sect prcdant la partition FAT (0 si mdias non part.)
  uint32 BPB_TotSec32   ; // Nb total de secteurs sur le volume
  uint32 BPB_FATSz32    ; // Nb de secteurs occups par 1 FAT
  uint BPB_ExtFlags     ; // Bits 0-3 : n de FAT active (mode mirroring)
                          //      4-6 : rservs
                          //        7 : 0=FAT reproduite sur toutes les FATs
                          //            1=une seule FAT est active
                          //      8-15: rservs
  uint BPB_FSVer        ; // N de version (majeur=MSB, mineur=LSB)
  uint32 BPB_RootClus   ; // N du "cluster" du rpertoire racine (gnral. 2)
  uint BPB_FSInfo       ; // N du secteur contenant la structure FSINFO
  uint BPB_BkBootSec    ; // Si diffrent de 0, indique n secteur zne rserve
                          // contenant une copie du secteur de boot (souvent 6)
  char BPB_Reserved  [12];// Rserv pour une future volution (tout  0)
  char BS32_DrvNum       ;// N de lecteur de l'interruption 0x13 (MSDOS)
  char BS32_Reserved1    ;// Utilis par Windows NT ( 0 au formatage)
  char BS32_BootSig      ;// Signature de boot tendue (0x29). Indique la prsence
                          // des 3 champs suivants
  uint32 BS_VolID        ;// N d'identification du volume
  char BS_VolLab     [11];// Nom du volume
  char BS_FilSysType  [8];// Normalement 'FAT32 '
  char Not_used     [419];// Bloc d'octets non utiliss
  uint SignSector0       ;// Signature du secteur 0 (0x55AA)
 };
} BPB_Sector;

/**************************
  Structure FSInfo (FAT32)
***************************/
typedef union
{
 struct
 {// 1 secteur = 512 octets
  char octets[512];
 };
 struct
 {// Structure dtaille
  char FSI_LeadSig   [4]; // Valeur 0x41615252. Indentifie le secteur FSInfo
  char FSI_Reserved1 [480];// Rserv pour des extensions futures (format  0)
  char FSI_StructSig [4]; // Valeur 0x61417272. Une autre signature
  char FSI_Free_Count[4]; // Dernier dcompte de cluster libres
  char FSI_Nxt_Free  [4]; // N cluster  partir duquel le pilote FAT recherche des
                          // cluster libres
  char FSI_Reserved2 [12];// Rserv pour de futures extensions
  char FSI_TrailSig  [4]; // Valeur 0xAA550000. Signature de fin structure FSInfo
 };
} tFSInfos;

/**************************
  Structure FATFileInfos
***************************/
// 32 octets par structure, 16 structures par secteur (512 octets)
typedef struct
{
 char DIR_Name        [11]; // Nom "court" du fichier (8+3 caractres)
 char DIR_Attr            ; // Attributs du fichier 
 char DIR_NTRes           ; // Rserv pour Windows NT 
 char DIR_CrtTimeTenth    ; // Compteur de 1/10S : 0  199 
 char DIR_CrtTime     [2] ; // Heure de cration du fichier
 char DIR_CrtDate     [2] ; // Date de cration du fichier
 char DIR_LstAccDate  [2] ; // Date de dernier accs
 uint16 DIR_FstClusHI     ; // Mot de poids fort du N du 1 cluster du fichier
                            // Tjrs  0 en FAT12/16
 char DIR_WrtTime     [2] ; // Heure de la dernire criture
 char DIR_WrtDate     [2] ; // Date de la dernire criture
 uint16 DIR_FstClusLO     ; // Mot de poids faible du N du 1 cluster du fichier
 uint32 DIR_FileSize      ; // Taille du fichier en octets
} tFATFileInfos;

/**********************************************************
  Secteur : dfinition multitypes du buffer pour un secteur
***********************************************************/
typedef union
{
 struct
 {// 1 secteur = 512 octets
  byte octets[512];
 };
 struct
 {// 1 secteur = 256 mots de 16 bits
  uint16 mots_16b[256];
 };
 struct
 {// 1 secteur = 128 mots de 32 bits
  uint32 mots_32b[128];
 };
 BPB_Sector BPB_FAT;              // Structure BPB du secteur "boot"
 tFATFileInfos FATFilesInfos[16]; // 16 structures d'entres de fichier
} Secteur;

/*************
  Type de FAT
**************/
typedef enum
{
  FAT12 = 0,
  FAT16 = 1,
  FAT32 = 2,
}FAT_type; // Type de FAT de la SD Card

#endif // FAT_Type_Defs
