     /********************************************************************
      ***                      Projet VAE                              ***
      ***          Dfinitions de types pour gestion cartes SD         ***
      ***                      dsPIC30F3011                            ***
      ***                                                              ***
      ***  Auteur : CREMMEL Marcel                                     ***
      ***  Date : 23/06/2007                                           ***
      ********************************************************************/

              /***************************************************
              *    Dfinitions de types pour cartes mmoire SD   *
              ****************************************************/

#ifndef SD_Type_Defs
#define SD_Type_Defs

#include "Typedefs.h"     // De Microchip : BYTE, WORD, DWORD, POINTER
                          // Perso : uint, uint16 et uint32

/******************************
  SDC_Error : types d'erreurs
*******************************/
typedef enum 
{
 sdcValid=0,              // Indicateur de carte valide
 sdcCardInitCommFailure,  // Aucune communication n'a pu s'tablir avec la carte
 sdcCardNotInitFailure,   // La carte n'est pas entre dans la phase "initialisation"
 sdcCardInitTimeout,      // "Timeout" pdt l'initialisation
 sdcCardTypeInvalid,      // Le type de carte n'a pas pu tre tabli
 sdcCardBadCmd,           // La carte n'a pas reconnu la commande
 sdcCardTimeout,          // "Timeout" pdt une lecture, criture ou effacement   
 sdcCardCRCError,         // Erreur CRC pdt une lecture 
 sdcCardDataRejected,     // Le "CRC" des donnes ne correspond pas
 sdcEraseTimedOut,        // "Timeout" pdt effacement
 sdcNoCard,               // Indique l'absence de carte
 sdcNoMemory              // Mmoire "heap" insuffisante
} SDC_Error;

/******************************
  Registre CSD : structures
*******************************/
typedef union
{
 struct
 {// 16 octets = 128 bits
  byte _byte[16];
 };
 struct
 {// Structure dtaille                 Nom              Bits R/W Indice Valeur Code
  unsigned NOT_USED           :1; //                       1    R   0-0           1b
  unsigned CRC                :7; // CRC                   7   R/W  7-1           CRC7
  unsigned ECC                :2; //                       2   R/W  9-8           00b
  unsigned FILE_FORMAT        :2; // File format           2   R/W 11-10       HD 00b
  unsigned TMP_WRITE_PROTECT  :1; // Temporary write prot. 1   R/W 12-12          0b
  unsigned PERM_WRITE_PROTECT :1; // Permanent write prot. 1   R/W 13-13          0b
  unsigned COPY               :1; // Copy flag (OTP)       1   R/W 14-14          1b
  unsigned FILE_FORMAT_GRP    :1; // File Format Group     1   R/W 15-15       0  0b
  unsigned RESERVED_1         :5; //                       5    R  20-16          00h
  unsigned WRITE_BL_PARTIAL   :1; // Partial blk W allowed 1    R  21-21      No  0b
  unsigned WRITE_BL_LEN       :4; // Max write dta blk lth 4    R  25-22     512  9h
  unsigned R2W_FACTOR         :3; // Write speed factor    3    R  28-26     x16  100b
  unsigned DEFAULT_ECC        :2; // Rserv MMC           2    R  30-29          00b
  unsigned WP_GRP_ENABLE      :1; // Write protect grp EN  1    R  31-31     Yes  1b
  unsigned WP_GRP_SIZE        :7; // Write protect grp siz 7    R  38-32     128  7Fh
  unsigned SECTOR_SIZE        :7; // Erase sector size     7    R  45-39      32  1Fh  
  unsigned ERASE_BLK_EN       :1; // Erase single block EN 1    R  46-46     Yes  1b
  unsigned C_SIZE_MULT_L      :1; // Device size multipli  1    R  47-47     
  unsigned C_SIZE_MULT_H      :2; // Device size multipli  2    R  49-48      64  100b
                                  //                                          32  011b
                                  //                                          16  010b
  unsigned VDD_W_CURR_MAX     :3; // Max RD ICC  Vddmax   3    R  52-50     80mA 110b
  unsigned VDD_W_CUR_MIN      :3; // Max WR ICC  Vddmin   3    R  55-53    100mA 110b
  unsigned VDD_R_CURR_MAX     :3; // Max RD ICC  Vddmax   3    R  58-56     80mA 110b
  unsigned VDD_R_CURR_MIN     :3; // Max RD ICC  Vddmin   3    R  61-59    100mA 111b
  unsigned C_SIZE_L           :2; // Device Size (LSB)     2    R  63-62   
  unsigned C_SIZE_H           :10;// Device Size (MSB)     10   R  73-64    128M  F03h
                                  //                                         64M  EDFh
                                  //                                         32M  74Bh
                                  //                                         16M  383h
                                  //                                          8M  33Fh
  unsigned RESERVED_2         :2; //                       2    R  75-74          00b
  unsigned DSR_IMP            :1; // DSR implemented       1    R  76-76    No    0b
  unsigned READ_BLK_MISALIGN  :1; // Read block misalign   1    R  77-77    No    0b
  unsigned WRITE_BLK_MISALIGN :1; // Write block misalign  1    R  78-78    No    0b
  unsigned READ_BL_PARTIAL    :1; // Partial block allowed 1    R  79-79    Yes   1b
  unsigned READ_BL_LEN        :4; // Max read bloc length  4    R  83-80    512   9h
  unsigned CCC                :12;// Card Command Classes  12   R  95-84    All   1F5h
  unsigned TRAN_SPEED         :8; // Max transfert rate    8    R  103-96   25MHz 32h      
  unsigned NSAC               :8; // Data Read Access Time 8    R  111-104    0   00h
  unsigned TAAC               :8; // Data Read Access Time 8    R  119-112  1,5mS 26h
  unsigned RESERVED_3         :2; //                       2    R                 00b
  unsigned SPEC_VERS          :4; //                       4    R                 0h
  unsigned CSD_STRUCTURE      :2; // Structure CSD         2    R  127-126  V1.0  00b
 };
} CSD;

/************************************
  CID : registre Card IDentification
*************************************/
typedef union
{
 struct
 {// 4 mots de 32 bits = 128 bits
  DWORD _u320;
  DWORD _u321;
  DWORD _u322;
  DWORD _u323;
 };
 struct
 {// 16 octets = 128 bits
  BYTE _byte[16];
 };
 struct
 {// Structure du registre
  unsigned 	NOT_USED           	:1; // LSB
  unsigned 	CRC                	:7; // CRC7    
  unsigned 	MDT                	:8; // Manufacturing Date Code (BCD)
  DWORD 	PSN;    				// Serial Number (PSN)
  unsigned 	PRV                	:8; // Product Revision
  char		PNM[6];    				// Product Name
  WORD 		OID;    				// OEM/Application ID
  unsigned 	MID                	:8; // Manufacture ID                        
 };
} CID;

/******************************
  SDCSTATE : tats de la carte
*******************************/
typedef union _SDCstate
{
 struct
 {
  byte isSDMMC : 1; // Indicateur de carte SDMMC 
  byte isWP    : 1; // Indicateur de protection en criture
 };
 byte  _byte;
} SDCSTATE;

/*******************************************
  sdmmc_cmd : types des commandes utilises
********************************************/
typedef enum
{//                     Index Argument   Reponse Description
 GO_IDLE_STATE,      // CMD0   Aucun        R1   Reset to Idle State
 SEND_OP_COND,       // CMD1   Aucun        R1   Dclenche le processus d'init.
 SEND_CSD,           // CMD9   Aucun        R1   Lecture du registre CSD
 SEND_CID,           // CMD10  Aucun        R1   Lecture du registre CID
 STOP_TRANSMISSION,  // CMD12  Aucun        R1b  Termine une op. R/W blk mult
 SEND_STATUS,        // CMD13  Aucun        R2   Lecture du registre de status
 SET_BLOCKLEN,       // CMD16 [31:0]length  R1   Affecte la longueur des blocs
 READ_SINGLE_BLOCK,  // CMD17 [31:0]adress  R1   Lecture d'un bloc
 READ_MULTI_BLOCK,   // CMD18 [31:0]adress  R1   Lecture continuelle de blocs
 WRITE_SINGLE_BLOCK, // CMD24 [31:0]adress  R1   Ecriture d'un bloc
 WRITE_MULTI_BLOCK,  // CMD25 [31:0]adress  R1   Ecriture continuelle de blocs 
 TAG_SECTOR_START,   // CMD32 [31:0]adress  R1   Adresse dbut  effacer
 TAG_SECTOR_END,     // CMD33 [31:0]adress  R1   Adresse fin  effacer
 UNTAG_SECTOR,       // CMD34 
 TAG_ERASE_GRP_START,// CMD35
 TAG_ERASE_GRP_END,  // CMD36
 UNTAG_ERASE_GRP,    // CMD37
 ERASE,              // CMD38 [31:0]xx      R1b  Effacer les blocs slectionns
 LOCK_UNLOCK,        // CMD49
 SD_APP_OP_COND,     // ACMD41 Aucun        R1   Dclenche l'initialisation
 APP_CMD,            // CMD55 [31:0]stuff   R1   La commande suivante est de type
                     //                          "Application specific"
 READ_OCR,           // CMD58  Aucun        R3   Lecture registre OCR
 CRC_ON_OFF          // CMD59 [31:1]xx      R1   CRC ON ou OFF
                     //       [0:0]CRC option
} sdmmc_cmd;

/************************************
  RESP : types de rponses possibles
*************************************/
typedef enum
{
 R1,   // Rponse standard : longueur 6 octets (48 bits)
 R1b,
 R2,   // Registres CID ou CSD de 136 bits (17 octets)
 R3    // Registre OCR de 48 bits (6 octets) 
} RESP;

/*************************************************
  CMD_PACKET : structures des trames de commandes
**************************************************/
typedef union
{
 struct
 {// 6 octets = 48 bits
  byte field[6];
 };
 struct
 {// Note : l'ordre de transmission n'est pas respect pour viter
  //        un bug du compilateur C30
  byte addr0; // Argument MSB
  byte addr1;
  byte addr2;
  byte addr3; // Argument LSB
  byte cmd;   // Code de la commande
  byte crc;   // CRC
 };
 struct
 {
  dword address;
  byte command;
  unsigned END_BIT:1;
  unsigned CRC7:7;
 };
} CMD_PACKET;

/**************************************
  RESPONSE_1 : natures des rponses R1
  Longueur : 8 bits
***************************************/
typedef union
{
 byte _byte;
 struct
 {
  unsigned IN_IDLE_STATE:1; // La carte est en  mode "Idle" 
                            // et excute la phase d'initialisation
  unsigned ERASE_RESET  :1; // Acquitement de l'excution d'une commande 
                            // d'arrt d'effacement
  unsigned ILLEGAL_CMD  :1; // Commande illgale
  unsigned CRC_ERR      :1; // Le CRC de la dernire commande est incorrect
  unsigned ERASE_SEQ_ERR:1; // Erreur ds la squence d'effacement commande
  unsigned ADDRESS_ERR  :1; // Erreur d'adresse 
  unsigned PARAM_ERR    :1; // Erreur dans les arguments de la dernire commande
  unsigned B7           :1; // Toujours  0
 };
} RESPONSE_1;

/**************************************
  RESPONSE_2 : natures des rponses R2
  Longueur : 16 bits
***************************************/
typedef union
{
 word _word;
 struct
 {
  byte      _byte0;
  byte      _byte1;
 };
 struct
 {
  unsigned IN_IDLE_STATE         :1; // La carte est en  mode "Idle" 
                                     // et excute la phase d'initialisation
  unsigned ERASE_RESET           :1; // Acquitement de l'excution d'une commande 
                                     // d'arrt d'effacement
  unsigned ILLEGAL_CMD           :1; // Commande illgale
  unsigned CRC_ERR               :1; // Le CRC de la dernire commande est incorrect
  unsigned ERASE_SEQ_ERR         :1; // Erreur ds la squence d'effacement commande
  unsigned ADDRESS_ERR           :1; // Erreur d'adresse 
  unsigned PARAM_ERR             :1; // Erreur dans les arguments de la dernire cmd
  unsigned B7                    :1; // Toujours  0
  unsigned CARD_IS_LOCKED        :1; // La carte est verrouille
  unsigned WP_ERASE_SKIP_LK_FAIL :1; // WP Erase Skip / Cmd "Lock/Unlock" dfaillante
  unsigned ERROR                 :1; // Erreur
  unsigned CC_ERROR              :1; // Erreur "Card Controller"
  unsigned CARD_ECC_FAIL         :1; // Erreur ds la mise en oeuvre de "ECC"
  unsigned WP_VIOLATION          :1; // Ecriture ds une zne protge
  unsigned ERASE_PARAM           :1; // Secteurs  effacer invalides
  unsigned OUTRANGE_CSD_OVERWRITE:1; // Dbordement
 };
} RESPONSE_2;

/*************************************
  SDC_RESPONSE : natures des rponses 
**************************************/
typedef union
{
 RESPONSE_1 r1;  
 RESPONSE_2 r2;
} SDC_RESPONSE;

/***************************************
  typSDC_CMD : structure d'une commande  
  Utilis pour dfinir le tableau sdmmc_cmdtable[]
****************************************/
typedef struct
{
    byte    CmdCode;          // Code de la commande
    byte    CRC;              // CRC de la commande (inutile en mode "CRC off")
    RESP    responsetype;     // Nature de la rponse
    byte    moredataexpected; // Vrai si des donnes supplmentaires suivent la cmde
} typSDC_CMD;

#endif // SD_Type_Defs
