     /********************************************************************
      ***                      Projet VAE                              ***
      ***               Fonctions lies  l'UART2                      ***
      ***                      dsPIC30F4011                            ***
      ***                                                              ***
      ***  Fichier "entte" : UART2.h                                  ***
      ***  Auteur : CREMMEL Marcel                                     ***
      ***  Date   : 30/09/2007                                         ***
      ********************************************************************/

#ifndef UART2_Defs
#define UART2_Defs

#include <p30f4011.h>

#define Synchro  'Z'  // Caractre de dtection des trames UART2
#define CR       13   // Retour chariot
#define LF       10   // Saut de ligne

/*------------------------------------------------------------------------------
             Dclarations des fonctions en mmoire flash
------------------------------------------------------------------------------*/

/*---------------------------------------------------------------
  Function   : void InitUART2(void)
  Description: Initialisation de l'UART2
                 - 38400 bauds, 8 bits, pas de parit, 1 bit Stop
                 - interruption sur TX si buffer UART vide
                 - interruption sur RX  chaque caractre
------------------------------------------------------------------*/
void InitUART2(void);

/*-------------------------------------------------------------
  Function   : int ReadRXD2(char *c)
  Description: Lecture d'un octet dans le buffer de rception
                 - renvoi de 0 si aucun caractre reu
                 - renvoi de 1 si caractre reu :
                 - caractre renvoy dans "c"
                 - pointeur circulaire ptrRX2_RDdata incrment
----------------------------------------------------------------*/
int ReadRXD2(char *c);

/*--------------------------------------------------------------
  Function   : void WriteTXD2(char c)
  Description: Ecriture d'un octet dans le buffer d'mission
                 - attendre libration du buffer
                 - envoi du caractre reu
                 - pointeur circulaire ptrTX2_WRdata incrment
-----------------------------------------------------------------*/
void WriteTXD2(char c);

/*---------------------------------------------------------------
  Function   : void WriteIntTXD2(int i)
  Description: Ecriture d'un entier dans le buffer d'mission
  Paramtre  : Entier  transmettre en binaire
-----------------------------------------------------------------*/
void WriteIntTXD2(int i);

/*------------------------------------------------------------------
  Function   : void WriteTXD2_String(char *String)
  Description: Ecriture d'une chane de caractres dans le buffer TX
  Paramtre  : Pointeur de la chane de caractres
--------------------------------------------------------------------*/
void WriteTXD2_String(char *String);

/*-------------------------------------------------------------------
  Function   : void WriteLnTXD2_String(char *String)
  Description: Ecriture d'une chane de caractres dans le buffer TX
               "Retour chariot" (CR+LF) ajout  la fin
  Paramtre  : Pointeur de la chane de caractres
---------------------------------------------------------------------*/
void WriteLnTXD2_String(char *String);

/*---------------------------------------------
  Function   : void WriteLnTXD2(void
  Description: "Retour chariot" (CR+LF) sur TX2
-----------------------------------------------*/
void WriteLnTXD2(void);

/*------------------------------------------------------------------- 
  Fonction    : void WTX2_Msg_Nb8bits(char *String, byte x)
  Description : Transmission d'un message Ascii suivi d'un  nombre 
                cod sur 8 bits       
  Paramtres  : *String : pointeur vers la chane de destination
                x : nombre binaire  convertir et transmettre 
---------------------------------------------------------------------*/
void WTX2_Msg_Nb8bits(char *String, byte x);

/*------------------------------------------------------------------- 
  Fonction    : void WTX2_Msg_Nb16bits(char *String, byte x)
  Description : Transmission d'un message Ascii suivi d'un  nombre 
                cod sur 16 bits       
  Paramtres  : *String : pointeur vers la chane de destination
                x : nombre binaire  convertir et transmettre 
---------------------------------------------------------------------*/
void WTX2_Msg_Nb16bits(char *String,uint x);

/*------------------------------------------------------------------- 
  Fonction    : void WTX2_Msg_Nb32bits(char *String, byte x)
  Description : Transmission d'un message Ascii suivi d'un  nombre 
                cod sur 32 bits
  Paramtres  : *String : pointeur vers la chane de destination
                x : nombre binaire  convertir et transmettre 
---------------------------------------------------------------------*/
void WTX2_Msg_Nb32bits(char *String,uint32 x);

#endif // UART2_Defs
